unit NW2NWCMain;

{ See the Help File for information on what this SW does and does not do.
  This version creates files of NWC format, version 75.

  Status: File & stave headings: fully updated, but not yet checked.
          Element writing : fully updated, but not yet checked.
          Ignoring elements: 3,4,13
  ToDo :  i) Update Rechord to rechord+chord element when more than one note
          ii) Update slurs to above/below.
          iii) Fix chords so no overlaps.
  NB The Pitch byte in NW files increases mod 256 as actual pitch increases.
     The pitch byte is only changed to decreasing as actual pitch increases
     when it is written into WriteFile by WriteNote. It is 0 for B in both
     formats, A is 001, G 002, F 003 etc in the output NWC file,
     but C is 001, D 002, E 003 etc in the input NW file.

  STILL TO FIX:
  i)   Initial stave end line has no format assigned.
  iv)  MP118, line 2, last bar: grace notes of second beat incorrectly coalesced into first beat chord
       MP30B: Contains master repeat instead of local repeat. This has no "(0)" above it.
  vii) MP119 grace note problems in middle of 3rd stave.
  viii) MP133, line 2, bar 2: first crotchet is missing its stem. (Ditto in other similar bars)
        It's OK in MP158, for example.
        *** The chords need to be rationalised so that there are only two lengths of notes:
            One for stems up, and one for stems down.
  ix)  Determine what variable makes notes wider for longer lengths, and implement.
       Currently NWC files have proportional spacing, not minimal spacing.
  xiv) MP38: Absurd stems (a result of loading perhaps?). OK in NW viewer.
  xvi) MP335: Minim grace notes appear as crotchet grace notes. (Same throughout all files)
  xvii) MP411: Bar 1: top beam missing. (Was this a problem with original NW software?)
  xviii)MP335 contains lyrics, in case these need checking.
       They look fine, but may not have been technically lyrics in the original NW file.
  xix) Problem deleting 2nd & subsequent staves when constructing NWC file.
  xx) The 2nd last bar of MP245: the NWC file seems correct for chords and beams
         but both are rendered wrongly in  NW Composer v1.75.


  UNFIXABLE ISSUES
  i)   Only one of a pair of slurs showing in MP302. No working examples in NWC files.
       NWC file seems correct.
  ii)  Repeat bar lines have "(n)" above them, n=#repetitions. Cannot remove this.
  iii) Two+ rests are not possible in chords; replaced by grace notes.
  iv) Changes in the number of staves in a system may not be modified correctly.
  v) There are missing features, including most of the directions:
       Coda, Segno, Fine, ToCoda, DaCapo, DCalCoda, DCalFine, DalSegno, DSalCoda, DSalFine
  vi) Titles, (authors, copyright information,) text, lyrics, fonts,
     instrument patches and volumes may need editing
     as they were not stored well enough or at all in NW files.
  vii) No ghost notes are possible.
  viii) Stem up and stem down chords cannot overlap - AAAAH!

// See https://en.wikipedia.org/wiki/List_of_musical_symbols#Notes_and_rests

Notes have horiz posn of head centre. Crotchet heads are about 12 px wide
and are 10 apart when on opp sides of same stem.
Stave lines are 6px apart, so 3 between successively pitched notes.


FORMAT SUMMARY:

NWC v1.75 Files (uncompressed) for output:

#000   7 chars  Clef
#001  13 chars  Key Sig
#002   5 chars  Barline
#003   5 chars  Flow: "Special ending"
#004  11 chars  Instrument patch
#005   9 chars  Time Sig
#006   8 chars + text  tempo
#007  10 chars  Dynamic
#008  11 chars  Note
#009  11 chars  Rest
#010  13 chars  Chord
#011   6 chars  Sustain Pedal
#012   7 chars  Flow
#013  13 chars + 4*(6chars)  multi-point controller
#014   7 chars  Tempo Var
#015   6 chars  VolVar
#016
#017   6 chars + Text  Performance Style
#018  13 chars  Rest Chord
#083 or other number normally >64: start of staff heading. (Printable char: first of the staff name)


NW File input format:
22 chars for every element.

Ch8 = 0: NewStave, Clef
Ch8 = 1: KeySig
Ch8 = 2: TimeSig
Ch8 = 3: TimeSig, C if no other non-zero params, Cslash if sh9=1
Ch8 = 4: Bar
Ch8 = 5: Note or Rest
Ch8 = 7: Accident
Ch8 = 9: Text
Ch8 =11: Line Segment (no NWC equivalent)
Ch8 =12: Slur Pts
Ch8 =13: Symbols from menu "~" e.g. Accent,Sign,Flow,Volume,Unknown
Ch8 =14: Arpeggio

Current Issue:
  Remove bar lines in non-first staves of system
  Sample2, stave 10 (start system 4) bar 2:


Sample1.NW: some semantic errors which have caused peculiarities in output NWC file:
Fermata are all on top line although half should be on 2nd line.
One note has detatched dot which makes one bar the wrong length.
}


interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ComCtrls, ExtCtrls, Math, Grids, DirOutln, FileCtrl, HelpUnit ;

Const TestStart    = 318 ;  // If TestOmissions is non-zero, then TestOmissions
      TestStave    =   1 ;  // elements after TestStart will be omitted from the
      TestOmissions=   0 ; //106 ;  // output file for the TestStave.
                            // TestStart: 318 OK, 319 not
Type                        // TestOmissions: 106 OK, 105 not
TMainForm = class(TForm)
    GoButton      : TButton;
    CloseButton   : TButton;
    LabelA        : TLabel;
    SourceDirBox  : TDirectoryListBox;
    SourceDriveBox: TDriveComboBox;
    SourceFileBox : TFileListBox;
    WorkingDirBox : TDirectoryListBox;
    WorkingFileBox: TFileListBox;
    Label4        : TLabel;
    ModifiedBox   : TEdit;
    LogFileBox    : TDirectoryListBox;
    LogFileDrive  : TDriveComboBox;
    Label1        : TLabel;
    LogFileListBox: TFileListBox;
    Label5        : TLabel;
    ErrorBox      : TEdit;
    WarningBox    : TEdit;
    Label6        : TLabel;
    HelpButton    : TButton;
    Function  AddSubscript(WFname : String; ListBox: TFileListBox) : String ;
    procedure GoButtonClick    (Sender: TObject);
    procedure CloseButtonClick (Sender: TObject);
    procedure SourceDriveBoxChange(Sender: TObject);
    procedure SourceDirBoxChange(Sender: TObject);
    Procedure ConvertDir(WDNo:Integer) ;
    Procedure ConvertFile(FNo:Integer);
    Procedure ResetAll(Sender: TObject) ;
    procedure WorkingDirBoxChange(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure LogFileDriveChange(Sender: TObject);
    procedure LogFileBoxChange(Sender: TObject);
    Procedure AssignLogFile ;
    Function  OKFileName(FileNo:Integer) : Boolean ;
    Function  MakeUnique(WFname : String; ListBox: TFileListBox) : String ;
    Procedure HelpClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    Procedure ConfigFile(Open: Boolean) ;
End; {TMainForm}

Var MainForm    : TMainForm;

Const LogFname  = 'NWtoNWC_' ;
      MaxErr    =     71 ;
      MaxWarn   =      8 ;
      LineLen   =     22 ; // #chars in line of heading
      MaxStrings=   1000 ;

      NUL       = Chr(0) ;
      Tab       = Chr(9) ;
      Space     = Chr(32);
      OldSuffix = '.nw'  ;
      NewSuffix = '.nwc' ;
      Empty     = ''     ;

      // Strings to convert to semantic element:



Type  ErrorType = 1..MaxErr ;
      WarnType  = 1..MaxWarn;
      LineType  = 1..LineLen;
      LineRA    = Array[LineType] of Byte ;

Var   Errors,
      Warnings  : Integer ;
      ErrorRA   : Array[ErrorType] of Integer ;
      WarnRA    : Array[WarnType]  of Integer ;
      Strings   : Array[0..MaxStrings] of String ;
      NoStrings : Integer ;
      WFname    : String  ;
      Line      : LineRA  ;

    ModifiedFiles,           // Current count of #files that have been modified.
    LenSource   : Integer  ; // Length dir path of location of files to be modified.
    Logged,                  // True iff filename written to log file.
    Modified    : Boolean  ; // True iff current ReadFile has been modified.
    NewFileName,             // Dir Path + Name for NewFile (write modifications into it).
    ReadFileName: String   ; // Full path + name of current ReadFile (original file name).
    NewFile     : File     ; // Modified version of TmpFile.
    ReadFile    : File     ; // Original File to be read
    LogFile     : TextFile ; // List of errors or changes made to files.
    ExeDir      : String   ; // Hopefully the dir containing the executable and help files.



Implementation

{$R *.DFM}

// Procedure ModificationsBeforeWrite ;
// Procedure ModificationsAfterRead(pS : pStave) ;


Procedure TMainForm.HelpClick(Sender: TObject);
// Issue : need to record location of executable and read from that folder.
Var HelpFileName : String ;
Begin
    HelpFileName := ExeDir+'\ReadMe.rtf' ; // 'NW2NWCHelp.rtf' ;
    If FileExists(HelpFileName) then
    Begin
        HelpUnit.HelpForm.OnCreation(HelpFileName) ;
        HelpUnit.HelpForm.Show ;
    End
    else
        MessageDlg('Help file not found.', mtError, [mbCancel], 0)
End ; { HelpClick }


{#############################################################################}
{#############################################################################}
{############################# Utility Procedures ############################}
{#############################################################################}
{#############################################################################}

Procedure Log(S:String) ;
// Adds S to LogFile, and ReadFileName if no updates have occurred previously.
Var LogS, LogFileName : Boolean ;
Begin
    LogS        := True ; // MainForm.UpdateBox.Checked ;
    LogFileName := True ; // MainForm.UpdatedFiles.Checked ;
    Modified    := True ;
    If not Logged and (LogFileName or LogS) then
    Begin
        Logged := True ;
        Writeln(LogFile,'file:\\\'+ReadFileName) ;
    End ;
    If LogS then WriteLn(LogFile,{'   ',}S) ;
End ;

Procedure SimpleLog(S:String) ;
// Adds S to LogFile, and ReadFileName if no updates have occurred previously.
Begin
    // If not Modified then Modified := True ;
    If not Logged then
    Begin
        Logged := True ;
        Writeln(LogFile,'file:\\\'+ReadFileName) ;
    End ;
    WriteLn(LogFile,'   ',S) ;
End ;

Procedure Error(E: ErrorType; S:String) ;
Begin
    // If E<> then
    SimpleLog('Error '+IntToStr(E)+': '+S) ;
    Errors := Errors+1 ;
    MainForm.ErrorBox.Text := IntToStr(Errors) ;
    MainForm.ErrorBox.Update ;
    ErrorRA[E] := ErrorRA[E]+1 ;
End ;

Procedure Warning(W: WarnType; S:String) ;
Begin
    If (W<>30) and (W<>22) and (W<>1) then
    SimpleLog('Warning '+IntToStr(W)+': '+S) ;
    Warnings := Warnings+1 ;
    MainForm.WarningBox.Text := IntToStr(Warnings) ;
    MainForm.WarningBox.Update ;
    WarnRA[W] := WarnRA[W]+1 ;
End ;


{#############################################################################}
{#############################################################################}
{#########################     Enumerated Types     ##########################}
{#############################################################################}
{#############################################################################}


Type BarLine     = (Single,Double, SectionOpen,SectionClose,LocalRepeatOpen,
                    LocalRepeatClose,MasterRepeatOpen,MasterRepeatClose) ;
     StaveType   = (Standard,Upper,Lower,Orchestra) ;
     Colour      = (Zero,One,Two,Three) ;
     PatchType   = (Num1,Num0,General,Roland,MT32) ;
     LabelList   = (NoLabel,First,Top,All) ;
     MeasureList = (None,Plain,Circle,Boxed) ;
     Position    = (Above,Default,Below) ; // For ties and slurs
     Accidentals = (DoubleFlat,Flat,Natural,Sharp,DoubleSharp, NoAcc) ;
     Accents     = (Staccato, Tenuto, Accnt, Mordent, NoAccent) ;
     Volume      = (ppp,pp,p,mp,mf,f,ff,fff) ;
     VolVarType  = (Cresc,Decresc,Dim,Rinf,Sforz,Unfound) ;
     TempVarType = (Breath,Ferm,Accel,Allarg,Rall,Ritard,Ritend,Rubato,Stringendo,UnknownTV) ;
     PedSym      = (pedup,peddn) ;
     FlowType    = (Coda, Segno, Fine, ToCoda, DaCapo, DCalCoda, DCalFine,
                    DalSegno, DSalCoda, DSalFine, NoFlow) ;
     ClefType    = (Treble,Bass,Alto,Tenor) ; // Order as in NW & NWC files.
     NoteType    = (Breve,Semibreve,Minim,Crotchet,Quaver,Semiquaver,Demisemiquaver,Hemidemisemiquaver) ;
     StemType    = (Up,Dn,NoStem,Auto,Unset) ;
     PerfStyle   = (AdLib,Animato,Cantabile,ConBrio,Dolce,Espress,Grazioso,Legato,Maestoso,
                    Marc,MenoMosso,PocoAPoco,PiuMosso,Semplice,Simile,Solo,Sostenuto,
                    SottoVoce,Stacc,Subito,Ten,Tutti,VoltaSubito,NotFound) ;
     StaveElement= (NewStave,Clef,Bar,Note,Rest,ReChord,Chord,TimeSig,KeySig,Text,ALine,Vol,VolVar,PStyle,
                    Pedal,{Tempo,MPC,Ending,Patch,}
                    TempoVar,Flow,Symbol,Accident,SlurPt,Accent,Arpeg,Unknown) ;

Var  PerfTxt     : Array[PerfStyle]  of String ;
     VolVarTxt   : Array[VolVarType] of String ;
     FlowTxt     : Array[FlowType]   of String ;
     PerfMin     : Array[PerfStyle]  of Integer;
     VolVarMin   : Array[VolVarType] of Integer;

Function ByteToBarLine(Ch:Byte) : BarLine ;
Var I   : Integer ;
    Bar : BarLine ;
Begin
    Bar := Single ;
    For I := 1 to Ord(Ch) do Inc(Bar) ;
    ByteToBarLine := Bar
End ;  { ByteToBarLine }

Function ByteToStaveType(Ch:Byte) : StaveType ;
Var I : Integer ;
    G : StaveType ;
Begin
    G := Standard ;
    For I := 1 to Ord(Ch) do Inc(G) ;
    ByteToStaveType := G
End ;  { ByteToStaveType }

Function ByteToColour(Ch:Byte) : Colour ;
Var I : Integer ;
    C : Colour ;
Begin
    C := Zero ;
    For I := 1 to Ord(Ch) do Inc(C) ;
    ByteToColour := C
End ;  { ByteToColour }

Function ByteToPatchType(Ch:Byte) : PatchType ;
Var I : Integer ;
    P : PatchType ;
Begin
    P := Num1 ;
    For I := 1 to Ord(Ch) do Inc(P) ;
    ByteToPatchType := P
End ;  { ByteToPatchType }

Function ByteToLabelList(Ch:Byte) : LabelList ;
Var I : Integer ;
    L : LabelList ;
Begin
    L := NoLabel ;
    For I := 1 to Ord(Ch) do Inc(L) ;
    ByteToLabelList := L
End ;  { ByteToLabelList }

Function ByteToMeasureList(Ch:Byte) : MeasureList ;
Var I : Integer ;
    M : MeasureList ;
Begin
    M := None ;
    For I := 1 to Ord(Ch) do Inc(M) ;
    ByteToMeasureList := M
End ;  { ByteToMeasureList }

Function ByteToAccidentals(Ch:Byte) : Accidentals ;
Var I   : Integer ;
    Acc : Accidentals ;
Begin
    Acc := DoubleFlat ;
    For I := 1 to Ord(Ch) do Inc(Acc) ;
    ByteToAccidentals := Acc
End ;  { ByteToAccidentals }

Function ByteToAccents(Ch:Byte) : Accents ;
Var I : Integer ;
    A : Accents ;
Begin // This needs research.
    If (Ord(Ch) = 0) or (Ord(Ch)>4) then ByteToAccents := NoAccent
    else
    Begin
        A := Staccato ;
        For I := 2 to Ord(Ch) do Inc(A) ;
        ByteToAccents := A
    End
End ;  { ByteToAccents }

Function ByteToVolume(Ch:Byte) : Volume ;
Var I : Integer ;
    V : Volume ;
Begin
    V := ppp ;
    For I := 1 to Ord(Ch) do Inc(V) ;
    ByteToVolume := V
End ;  { ByteToVolume }

Function ByteToTempVarType(Ch:Byte) : TempVarType ;
Var I : Integer ;
    T : TempVarType ;
Begin
    T := Breath ;
    For I := 1 to Ord(Ch) do Inc(T) ;
    ByteToTempVarType := T
End ;  { ByteToTempVarType }

Function ByteToPedSym(Ch:Byte) : PedSym ;
Var I : Integer ;
    S : PedSym ;
Begin
    S := pedup ;
    For I := 1 to Ord(Ch) do Inc(S) ;
    ByteToPedSym := S
End ;  { ByteToPedSym }

Function ByteToFlowType(Ch:Byte) : FlowType ;
Var I : Integer ;
    F : FlowType ;
Begin
    F := Coda ;
    For I := 1 to Ord(Ch) do Inc(F) ;
    ByteToFlowType := F
End ;  { ByteToFlowType }

Function ByteToClef(Ch:Byte) : ClefType ;
// Assumes cleftype order matches NW file character values.
Var I : Integer ;
    Clef : ClefType ;
Begin
    Clef := Treble ;
    For I := 1 to (Ord(Ch) mod 4) do Inc(Clef) ;
    ByteToClef := Clef
End ;  { ByteToClef }

Function ByteToNoteType(Ch:Byte) : NoteType ;
Var I : Integer ;
    N : NoteType ;
    OK: Boolean ;
Begin
    N := Breve ;
    OK:= True  ;
    For I := 1 to Ord(Ch) do
        If N <> Hemidemisemiquaver then Inc(N) else OK := False ;
    ByteToNoteType := N ;
    If not OK then Error(47,'Unknown very short note length '+IntToStr(Ord(Ch))) ;
End ;  { ByteToNoteType }

Function ByteToStr(B:Byte) : String ;
Begin
    ByteToStr := IntToStr(Ord(Chr(B)))
End ;  { ByteToStr }


Function SymToStr(SymNo:Byte) : String ;
Begin
    Case SymNo of
         1 : SymToStr := 'tenor/alto clef' ;
         2 : SymToStr := 'bass clef' ;
         3 : SymToStr := 'treble clef' ;
         6 : SymToStr := 'small bass clef' ;
         7 : SymToStr := 'small treble clef' ;
        21 : SymToStr := 'breve' ;
        22 : SymToStr := 'stemless open notehead' ;
        23 : SymToStr := 'minim' ; // stem up
        24 : SymToStr := 'minim' ; // stem dn
        25 : SymToStr := 'stemless closed notehead' ;
        26 : SymToStr := 'crotchet' ; // stem up
        27 : SymToStr := 'crotchet' ; // stem dn
        28 : SymToStr := 'quaver' ; // stem up
        29 : SymToStr := 'quaver' ; // stem dn
        30 : SymToStr := 'semiquaver' ; // stem up
        31 : SymToStr := 'semiquaver' ; // stem dn
        32 : SymToStr := 'demisemiquaver' ; // stem up
        33 : SymToStr := 'demisemiquaver' ; // stem dn
        34 : SymToStr := 'hemidemisemiquaver' ; // stem up
        35 : SymToStr := 'hemidemisemiquaver' ; // stem dn
        36 : SymToStr := 'slash' ; // as in caesura
        37 : SymToStr := 'dot' ;
        38 : SymToStr := 'double dot' ;
        44 : SymToStr := 'flat' ;
        45 : SymToStr := 'double flat' ;
        46 : SymToStr := 'sharp' ;
        47 : SymToStr := 'double sharp' ;
        48 : SymToStr := 'natural' ;
        49 : SymToStr := 'semibreve rest' ;
        50 : SymToStr := 'breve rest' ;
        51 : SymToStr := 'minim rest' ;
        52 : SymToStr := 'crotchet rest' ;
        53 : SymToStr := 'quaver rest' ;
        54 : SymToStr := 'semiquaver rest' ;
        55 : SymToStr := 'demisemiquaver rest' ;
        56 : SymToStr := 'hemidemisemiquaver rest' ;
        64 : SymToStr := 'Accent' ; // >
        65 : SymToStr := 'Fermata (above)' ;
        66 : SymToStr := 'Fermata (below)' ;
        70 : SymToStr := 'Simile' ; //  ./.
        71 : SymToStr := 'staccato' ; // small square
        73 : SymToStr := 'comma' ; // opening comma
        74 : SymToStr := 'breath' ; // closing comma
        75 : SymToStr := 'Marcato' ; // ^
        76 : SymToStr := 'Accent' ; // >
        78 : SymToStr := 'tenuto' ; // horiz bar _
       115 : SymToStr := 'Coda' ;
       116 : SymToStr := 'Segno' ;
       121 : SymToStr := 'grace flat' ;
       122 : SymToStr := 'grace sharp' ;
       123 : SymToStr := 'grace natural' ;
       124 : SymToStr := 'slash' ; // slowly rising wide slash
       125 : SymToStr := 'tick' ; // Breath mark (tick)
       126 : SymToStr := 'Bow up' ; //  v
       127 : SymToStr := 'Bow Dn' ; //
       128 : SymToStr := 'breath' ; // large comma
       129 : SymToStr := 'mordent' ; // (upper)
       130 : SymToStr := 'mordent' ; // (lower)
       131 : SymToStr := 'stroke' ; // falling \
       132 : SymToStr := 'stroke' ; // rising /, used in slash on note stem
       133 : SymToStr := 'Open string' ; // (small circle)
       134 : SymToStr := 'pedal dn' ;
       135 : SymToStr := 'pedal up' ;
       136 : SymToStr := 'turn'; //  ~
       137 : SymToStr := 'multi-measure rest' ;
       138 : SymToStr := 'pp' ;
       139 : SymToStr := 'p' ;
       140 : SymToStr := 'mp' ;
       141 : SymToStr := 'mf' ;
       142 : SymToStr := 'f' ;
       143 : SymToStr := 'ff' ;
       144 : SymToStr := 'sf' ;
       145 : SymToStr := 'fz' ;
       146 : SymToStr := '8va' ;
       156 : SymToStr := 'semibreve' ;
       else  SymToStr := 'Unknown symbol'
    End
End ; { SymToStr }

Function PitchedSym(SymNo:Byte) : Boolean ;
// If true, then WriteText will position the text properly, else it must be raised above stave.
Begin
    If SymNo in
        [65,66,75,76,126,127,129,130,136]
    then PitchedSym := False else PitchedSym := True ;
End ; { PitchedSym }

{#############################################################################}
{#############################################################################}
{#########################       Record Types       ##########################}
{#############################################################################}
{#############################################################################}

Const NoFonts  =   10 ;
      MaxPitch =  256 ; // Strict upper bound on pitch of highest note.
      MinPitch = -256 ; // Strict upper bound on pitch of lowest note.
      MaxDur   =  256 ; // Strict upper bound on length of longest note.
      Undefined= -256 ; // Useful when determining integers within byte limits.

Type
{    GenRec  = Record
                 Name: String ;
                 Grp : Group ;
                 EndBar: Barline ;
               End ;

     VisRec  = Record
                  Upper, Lower, Lines : Integer ;
                  Layer : Boolean ;
                  Style : Group ;
                  Color : Colour ;
               End ;

     MidRec  = Record
                  Vol, Pan, Trans, Channel : Integer ;
                  Muted : Boolean ;
                  Playback : String ;
               End ;

     InsRec  = Record
                  Name : String ;
                  Patch: PatchType ;
                  Select : Boolean ;
                  Contr,ThirtyTwo : Integer ;
               End ;

     Staff   = Record
                 General    : GenRec;
                 Visual     : VisRec;
                 Midi       : MidRec;
                 Instrument : InsRec;
               End ;

     LyricConfig = Record
                     Lines, Upper,Lower, Offset : Integer ;
                     Standard, Bottom : Boolean
                   End;

     FileInfo = Record
                  Title, Author, Copy1, Copy2, Comments : String ;
                End ;

     TempoProp  = Record // Many components
                  End ;


     VolVarRec  = Record  // There are more properties still to put in...
                      VVT  : VolVarType ;
                      Where: Integer;
                      Width: Boolean;
                      Just : Justification;
                  End ;
}
     HPinRec     = Record
                         Start,
                         Cresc : Boolean
                   End ;

     FontRec     = Record
                         Name : String ;
                         Italic,Bold : Boolean ;
                   End ;

     SlurType    = Record
                         Next,Prev,
                         Id : Byte ;
                   End ;

     TimeSigs    = Record
                         Top,Bot : Integer ;
                         IsTop,
                         Common,AllaBreve :Boolean
                   End ;
     KeySigs     = Record
                         IsSharps,
                         IsNatural,
                         IsFlats : Boolean ;
                         Number  : Integer ;
                   End ;

     NoteRec     = Record  // Used for chords, notes, ReChords and rests.
                         NoteSym   : NoteType ;
                         Pitch     : Integer  ; // Changed from byte to range -128..128.
                                                // This differs between NW & NWC files.
                         BeamTo, BeamFrom, TieTo, TieFrom, SlurTo, SlurFrom,
                         // BeamId, NextInChord,
                         NoInChord : Integer ;
                         Stem      : StemType; // I don't know if "Unset" is possible in NW files.
                         BeamOK, TieOK, // for checking completeness of beam/tie
                         Grace, Dotted, DoubleDot, TripletIn, TripletOut,
                         InChord, RtShift, // Only used in writing NWC file
                         WithReChord,
                         // RestCh {set but unused}, Ghost {set but unused},
                         BeamIn, BeamOut, TieIn, TieUp, TieOut, SlurIn, SlurOut,
                         Reverse   : Boolean ;
                         TiePosn,BeamPosn,SlurPosn : Position ;
                         Accidental: Accidentals ;
                         Accent    : Accents ;
                   End ;

     StaveRec    = Record
                     StType   : StaveType;
                     TSig     : TimeSigs ;
                     KSig     : KeySigs  ;
                     Clef     : ClefType ;
                     // First : Boolean  ;
                     NoLyrics : Integer  ;
                     Numbering,
                     TitleInfo,
                     ExtendLast,
                     IncreaseSpacing : Boolean ;
                     StaffSize,
                     From,Start : Integer ;
                     Labels   : LabelList ;
                     Numbers  : MeasureList ;
                     FinalBar : BarLine ;
                   End ;

     TxtType     = Record
                     // Position,    // Vertical position relative to centre of its stave.
                     StringIndex:Integer;
                     FontNo   : Byte ;  // Font no from NW file.
                     Italic,
                     Small    : Boolean ;
                   End ;

     Element     = Record // Variant; types given under "insert" in Noteworthy
                      StaveNo,
                      ElementNo,
                      VertPosn,  // Vertical position; adjusted to include negative values.
                                 // Eventually adjusted to be relative to centre of its stave.
                      HorizPosn : Integer ;
                                 // Horizontal position; adjusted to include negative values.
                      Case ElType : StaveElement of
                         Clef    : (Cl:ClefType) ;
                         Bar     : (BL:BarLine)  ;
                         Chord,
                         ReChord,
                         Rest,
                         Note    : (No:NoteRec)  ;
                         TimeSig : (TS:TimeSigs) ;
                         KeySig  : (KS:KeySigs)  ;
                         Text    : (Txt:TxtType) ;  // This is an index to an array of Strings
                         Accident: (Acd:Accidentals;Pitch:Integer) ;
                         Accent  : (Ant:Accents) ;
                         Vol     : (V:Volume)    ;
                         VolVar  : (VV:VolVarType) ;
                         PStyle  : (PF:PerfStyle);
                         Pedal   : (Pd:PedSym)   ;
                         // Tempo: (Tm:TempoProp);
                         TempoVar: (TmpVar:TempVarType) ;
                         SlurPt  : (SPt:SlurType);
                         Flow    : (FL:FlowType) ;
                         NewStave: () ;
                         Symbol  : (Sym:Byte{; Pitch:Integer})  ;
                 End;  // Element record


     OptionRec = Record
                 End ;

     MarginRec = Record
                   Top, Inside, Outside, Bottom: Real ;
                   Mirror, Inches              : Boolean;
                 End ;

     pElementList = ^ElementList ;
     ElementList= Record
                    Next : pElementList ;
                    El   : Element ;
                  End ;

     pStave     = ^Stave ;
     Stave      = Record
                    Next       : pStave ;
                    StaveNo,
                    NoElements,
                    VertPosn   : Integer ;  // Absolute vertical posn, as read from NW file
                    FirstInSystem,
                    LastInSystem : Boolean ;
                    Elements   : pElementList ;
                    Stv        : StaveRec ;
                  End ;

     Page    =  Record
                  NWVersionNo,
                  NoStaves: Integer  ;
                  Staves  : pStave   ;
                  Options : StaveRec ;
                  Margins : MarginRec;
                  Title   : TxtType  ;
                  Fonts   : Array[1..NoFonts] of FontRec;
                End;

     Var Music : Page ;
         Head,
         Foot  : String ;


Procedure InitialiseNoteRec(Var Note : NoteRec) ;
Begin
    With Note do
    Begin
        NoteSym  := Crotchet ;
        Dotted   := False ;
        DoubleDot:= False ;
        Pitch    := 0     ;
        Stem     := Up    ;
        BeamOK   := True  ;
        TieOK    := True  ;
        Reverse  := False ;
        BeamIn   := False ;
        BeamOut  := False ;
        BeamTo   := 0     ;
        BeamFrom := 0     ;
        SlurIn   := False ;
        SlurOut  := False ;
        SlurTo   := 0     ;
        SlurFrom := 0     ;
        TieIn    := False ;
        TieOut   := False ;
        TieUp    := False ;
        TieTo    := 0     ;
        TieFrom  := 0     ;
        Grace    := False ;
        TripletIn:= False ;
        TripletOut:=False ;
        RtShift  := False ;
        InChord  := False ;
        NoInChord:= 0     ;
        WithReChord:=False;
        RtShift  := False ;
        TiePosn  := Default ;
        BeamPosn := Default ;
        SlurPosn := Default ;
        Accidental:=NoAcc   ;
        Accent   := NoAccent;
    End ;
End ; { InitialiseNoteRec }


Function NewElement(ET: StaveElement; Next: pElementList) : pElementList ;
Var pE : pElementList ;
Begin
    New(pE) ;
    pE.Next := Next ;
    With pE.El do
    Begin
        StaveNo   := 0 ;
        ElementNo := 0 ;
        VertPosn  := 0 ;
        HorizPosn := 0 ;
        ElType    := ET;
    End ;
    If ET in [Chord, ReChord, Rest, Note] then InitialiseNoteRec(pE.El.No) ;
    NewElement := pE
End;


{#############################################################################}
{#############################################################################}
{#########################        Debugging         ##########################}
{#############################################################################}
{#############################################################################}


Procedure TestDeletions ;
// Removes TestOmissions elements from TestStave, starting at TestStart.
Var pS     : pStave ;
    pE, pF : pElementList ;
    I, NoEls, LastChord, First : Integer ;
Begin
    If TestOmissions > 0 then
    Begin
        pS := Music.Staves ;
        I := 1 ;
        While (pS<>nil) and (I<TestStave) do
        Begin pS := pS.Next ; Inc(I) End ;
        If I<>TestStave then
        MessageDlg('No such stave to test',mtInformation,[mbOK],0)
        else
        Begin
            NoEls := 0 ;
            pE := pS.Elements ;
            While (pE<>nil) and (I<TestStart) do
               Begin  pE := pE.Next ; Inc(NoEls) End ;
            I := 1 ;
            LastChord := 0 ;
            pE := pS.Elements ;
            While (pE<>nil) and (I<TestStart) do
            Begin
                pE := pE.Next ;
                Inc(I);
                If (pE<>Nil) and (pE.El.ElType in [ReChord,Chord]) then
                     LastChord := I+pE.El.No.NoInChord
            End ;
            While (LastChord>I) and (pE<>Nil) do
                Begin pE := pE.Next ; Inc(I) End ;
            If pE=Nil then
                MessageDlg('No elements being removed in test.',mtInformation,[mbOK],0)
            else
            Begin
                First := I ;
                pF := pE ;
                I  := -1 ;
                LastChord := 0 ;
                While (pF<>Nil) and (I<TestOmissions) do
                Begin
                    pF := pF.Next ; Inc(I) ;
                    If (pF<>Nil) and (pF.El.ElType in [ReChord,Chord]) then
                        LastChord := I+pF.El.No.NoInChord
                End ;
                While (LastChord>=I) and (pF<>Nil) do
                    Begin pF := pF.Next ; Inc(I) End ;
                pE.Next := pF ;
                If pF = Nil then
                Log('Removed all remaining elements from stave '+IntToStr(TestStave)+
                    ' of '+IntToStr(NoEls)+' elements, after element '+IntToStr(First))
                else
                Log('Removed '+IntToStr(I)+' elements from stave '+IntToStr(TestStave)+
                    ' of '+IntToStr(NoEls)+' elements, starting at element '+IntToStr(First)) ;
            End
        End
    End
End ; { TestDeletions }

Function ChSt(Ch:Byte): String ;
Var N : Integer ;
Begin
       N := Ord(Ch) ;
       If N < 10 then ChSt := '#00'+IntToStr(N) else
       If N <100 then ChSt := '#0' +IntToStr(N) else
                      ChSt := '#'  +IntToStr(N) ;
End ; { ChSt }

Function Line2Str(L : LineRA) : String ;
Var I : Integer;
    S : String ;
Begin
    S := '' ;
    For I := 1 to LineLen do
    Begin
        S := S+ChSt(L[I]) ;
        If I mod 5 = 0 then S := S+'.'
    End ;
    Line2Str := S ;
End ; { Line2Str }

Function El2String(El : StaveElement) : String ;
//      StaveElement = (Clef,Bar,Note,Rest,TimeSig,KeySig,Text,Vol,VolVar,
//                     Tempo,TempoVar,Flow,Pedal,NewSystem,NewStave,Unknown) ;
Var S : String ;
Begin
    Case El of
      Clef    : S := 'Clef' ;
      Bar     : S := 'Bar Line' ;
      Note    : S := 'Note' ;
      Rest    : S := 'Rest' ;
      Chord   : S := 'Chord' ;
      ReChord : S := 'Rest-chord' ;
      TimeSig : S := 'Time Signature' ;
      KeySig  : S := 'Key Signature' ;
      Text    : S := 'Text' ;
      NewStave: S := 'New Stave' ;
      Unknown : S := 'Unknown' ;
      Accident: S := 'Accidental' ;
      TempoVar: S := 'Tempo Var' ;
      Accent  : S := 'Accent' ;
      SlurPt  : S := 'Slur Point' ;
      Flow    : S := 'Flow Control' ;
      Symbol  : S := 'symbol' ;
      Pedal   : S := 'Ped Up/dn' ;
      VolVar  : S := 'Dynamic variation' ;
      Vol     : S := 'Volume' ;
    else
      // Sign Tempo ...
      S := 'Oops! ' ;
    End ;
    El2String := S
End ; { El2String }


Function PitchToChar(Pitch:Integer):Char ;
// Needs re-testing.
Begin
    If Pitch < 0 then Pitch := Pitch+192 ; // A multiple of 8 and 12 just in case!
    PitchToChar := Chr((Pitch+1) mod 7 + Ord('A'))
End ; { PitchToChar }

Function AccToStr(A : Accents): String ;
Begin
    Case A of
      Staccato : AccToStr := 'Staccato';
      Tenuto   : AccToStr := 'Tenuto'  ;
      Accnt    : AccToStr := 'Accent'  ;
      Mordent  : AccToStr := 'Mordent' ;
      NoAccent : AccToStr := 'unknown' ;
    End
End; { AccToStr }

Function NoteSym2String(N:NoteType) : String ;
Var S : String ;
Begin
        Case N of
        Breve     : S := 'Breve' ;
        Semibreve : S := 'Semibreve' ;
        Minim     : S := 'Minim' ;
        Crotchet  : S := 'Crotchet' ;
        Quaver    : S := 'Quaver' ;
        Semiquaver: S := 'Semiquaver' ;
        Demisemiquaver    : S := 'Demisemiquaver' ;
        Hemidemisemiquaver: S := 'Hemidemisemiquaver' ;
        else
           S := 'Unknown note symbol' ;
        End ;
    NoteSym2String := S
End; { NoteSym2String }

Procedure WriteSummary(S : String; NoStaves:Integer) ;
Var pS : pStave ;
    pE : pElementList ;
    T  : String ;
    BarNo,
    I : Integer ;
Begin
    Log(S) ;
    pS := Music.Staves ;
    I := 1 ;
    Log('No of Staves = '+IntToStr(NoStaves)) ;
    While (pS<>nil) and (I<=NoStaves) do
    Begin
        Log('Start of Stave') ;
        pE := pS.Elements ;
        BarNo := 1 ;
        While pE <> Nil do
        Begin
            T := El2String(pE.El.ElType) ;
            If pE.El.ElType in [Rest,Chord,ReChord,Note] then
            Begin
               If pE.El.ElType = Chord then
               T := T+' of '+IntToStr(pE.El.No.NoInChord)+' notes' ;
               T := T+Space+NoteSym2String(pE.El.No.NoteSym) ;
               T := T+Space+PitchToChar(pE.El.No.Pitch)
            End
            else
            If pE.El.ElType = Bar then
            Begin
                Inc(BarNo); T := T+Space+IntToStr(BarNo)
            End ;
            Log(T) ;
            pE := pE.Next
        End ;
        Log('End of Stave') ;
        pS := pS.Next ;
        Inc(I)
    End;
    Log('End of score.')
End ; { WriteSummary }



Function HasCycleProblem : Boolean ;
// For debugging purposes. This is currently unused.
// Checks no stave has an elementlist longer than MaxElementListLength,
// and no more than MaxStaveListLength staves.
Const MaxElementListLength = 2000 ;
      MaxStaveListLength   =   40 ;
Var pS : pStave ;
    pE : pElementList ;
    Staves, Elements : Integer ;
    OK : Boolean ;
Begin
    OK := True ;
    pS := Music.Staves ;
    Staves := 0 ;
    While (pS<>nil) and (Staves<MaxStaveListLength) and OK do
    Begin
        Inc(Staves) ;
        pE := pS.Elements ;
        Elements := 0 ;
        While (pE<>nil) and (Elements<MaxElementListLength) do
        Begin
            Inc(Elements) ;
            pE := pE.Next
        End ;
        If Elements>=MaxElementListLength then OK := False ;
        pS := pS.Next ;
    End ;
    If Staves>=MaxStaveListLength then OK := False ;
    HasCycleProblem := not OK
End ; { HasCycleProblem }

Function Pointer(pStart: pElementList; StaveNo,ElNo:Integer) : pElementList ;
// Finds pointer to element number ElNo in stave StaveNo if one occurs at or after pStart.
// Starts looking from beginning oflist if pSart=Nil.
// Returns Nil if not found or pStart is Nil or StaveNo=0 or ElNo=0.
Var pE, pEl : pElementList ;
Begin
    If  (pStart=Nil) or (StaveNo=0) or (ElNo=0) then pEl := Nil
    else
    Begin
       pEl := Nil ;
       pE  := pStart ;
       While pE <> Nil do
       Begin
          If (pE.El.StaveNo=StaveNo) and (pE.El.ElementNo=ElNo) then
          Begin
            pEl := pE ; pE := Nil
          End
          else pE := pE.Next
       End
    End ;
    Pointer := pEl
End ;  { Pointer }

Procedure PreDeleteElement(pS: pStave; Start:Integer; FileName:String) ;
Var pE,pPred, pFound : pElementList ;
    Posn     : Integer ;
Begin
  Posn:= Pos(FileName,ReadFileName) ;
  If (Posn>0) and (Posn+Length(FileName)=Length(ReadFileName)+1)  then
  Begin
      pE:= pS.Elements ;
      pPred := Nil ;
      pFound := Nil ;
      While pE<>Nil do
      Begin
          If pE.El.ElementNo = Start then
          Begin pFound := pE; pE := Nil End
          else
          Begin pPred := pE ; pE := pE.Next End
      End ;
      If pFound=Nil then
         Log('Not found element '+IntToStr(Start)+' in stave '+IntToStr(pS.StaveNo)+' requested for deletion')
      else
      Begin
         If pPred=Nil then pS.Elements := pFound.Next
         else pPred.Next := pFound.Next ;
         Dec(pS.NoElements) ;
         Log('Deleted element '+IntToStr(Start)+' of stave '+IntToStr(pS.StaveNo))
      End
  End
End ; { PreDeleteElement }

Procedure DeleteElement(Stave,Start:Integer; FileName:String) ;
Var pE,pPred, pFound : pElementList ;
    Posn     : Integer ;
Begin
  Posn:= Pos(FileName,ReadFileName) ;
  If (Posn>0) and (Posn+Length(FileName)=Length(ReadFileName)+1)  then
  Begin
      pE:= Music.Staves.Elements ;
      pPred := Nil ;
      pFound := Nil ;
      While pE<>Nil do
      Begin
          If (pE.El.StaveNo = Stave) and (pE.El.ElementNo = Start) then
          Begin pFound := pE; pE := Nil End
          else
          Begin pPred := pE ; pE := pE.Next End
      End ;
      If pFound=Nil then
         Log('Not found element '+IntToStr(Start)+' in stave '+IntToStr(Stave)+' requested for deletion')
      else
      Begin
         If pPred=Nil then Music.Staves.Elements := pFound.Next
         else pPred.Next := pFound.Next ;
         Dec(Music.Staves.NoElements) ;
         Log('Deleted element '+IntToStr(Start)+' of stave '+IntToStr(Stave))
      End
  End
End ; { DeleteElement }


Procedure DeleteElements(StartStave,StartNo,FinishStave,FinishNo:Integer; FileName:String) ;
// MP375 fix: removes all elements from StartNo in StartStave to FinishNo in FinishStave.
// #StartNo stays, #FinishNo is removed.
// Removes to end if no element numbered (FinishSatve,FinishNo) after (StartStave,StartNo).
// Don't use (0,0) to finish.
Var pE,pJoin,pNext,pEnd : pElementList ;
    Posn, Deleted  : Integer ;

Begin
  Posn:= Pos(FileName,ReadFileName) ;
  If (Posn>0) and (Posn+Length(FileName)=Length(ReadFileName)+1)  then
  Begin
    // Find StartNo & FinishNo:
    pJoin := Pointer(Music.Staves.Elements,StartStave,StartNo) ;
    pEnd  := Pointer(pJoin,FinishStave,FinishNo) ;

// If pEnd=Nil then Log('pEnd is nil') else Log('pEnd is not nil.') ;

    // Delete to FinishNo:
    pE := pJoin ;
    Deleted := 0 ;
    While (pE<>Nil) and (pE<>pEnd) do
    If (pE.Next<>Nil) then
    Begin // Delete pE.Next:
        pNext := pE.Next ;
        Inc(Deleted) ;
        If pNext.El.ElType = Chord then Deleted := Deleted - pNext.El.No.NoInChord ;
        pE := pNext;
    End
    else pE := Nil ;
    // Update:
    If pJoin <> Nil then
       If pEnd=Nil then pJoin.Next := Nil else pJoin.Next := pEnd.Next ;
    Music.Staves.NoElements := Music.Staves.NoElements - Deleted ;
    // Report:
    If pJoin = Nil then
    Log('Not found start element '+IntToStr(StartNo )+' of stave '+IntToStr(StartStave)
                    +' to start deleting elements')
    else
    If pEnd=Nil then
    Log('Deleted elements '+IntToStr(StartNo )+' of stave '+IntToStr(StartStave)+' to end')
    else
    Log('Deleted elements '+IntToStr(StartNo )+' of stave '+IntToStr(StartStave)
                    +' to '+IntToStr(FinishNo)+' of stave '+IntToStr(FinishStave)) ;
  End
End ; { DeleteElements }

Procedure RemoveGrace(StartStave,StartNo,FinishStave,FinishNo:Integer; FileName:String) ;
// MP375 fix: removes grace note property for all elements from StartNo to FinishNo in StaveNo.
// #StartNo grace property stays, #FinishNo grace property is removed.
// Removes property to end if no element numbered FinishNo after StartNo.
Var pE, pStart, pEnd : pElementList ;
    Posn             : Integer ;
Begin
  Posn:= Pos(FileName,ReadFileName) ;
  If (Posn>0) and (Posn+Length(FileName)=Length(ReadFileName)+1)  then
  Begin
    // Find StartNo & FinishNo:
    pStart := Pointer(Music.Staves.Elements,StartStave,StartNo) ;
    pEnd   := Pointer(pStart,FinishStave,FinishNo) ;
    // Change Grace note property:
    pE := pStart ;
    While pE<>Nil do
    Begin
        If (pE.El.ElType in [Chord,Rest,ReChord,Note])then
           pE.El.No.Grace  := False ;
        If pE=pEnd then pE := Nil else pE := pE.Next
    End ;
    // Report
    If pStart = Nil then
        Log('Failed to find start element '+IntToStr(StartNo)+' of stave '
               +IntToStr(StartStave)+' for removing grace note property')
    else
    If pEnd=Nil then
    Log('Removed grace note property for elements '+IntToStr(StartNo )+' of stave '+IntToStr(StartStave)+' to end')
    else
    Log('Removed grace note property for elements '+IntToStr(StartNo)+' of stave '+IntToStr(StartStave)
         +' to '+IntToStr(FinishNo)+' of stave '+IntToStr(FinishStave)) ;
  End
End ; { RemoveGrace }

Procedure RemoveBeams(StartStave,StartNo,FinishStave,FinishNo:Integer; FileName:String) ;
// MP375 fix: removes beams from StartNo to FinishNo in StaveNo.
// Beams in #StartNo stay, that in #FinishNo is removed.
// Removes beams to end if no element numbered FinishNo after StartNo.
Var pE, pStart, pEnd : pElementList ;
    Posn       : Integer ;
Begin
  Posn:= Pos(FileName,ReadFileName) ;
  If (Posn>0) and (Posn+Length(FileName)=Length(ReadFileName)+1)  then
  Begin
    // Find StartNo & FinishNo:
    pStart := Pointer(Music.Staves.Elements,StartStave,StartNo) ;
    pEnd   := Pointer(pStart,FinishStave,FinishNo) ;
    // Change Beam property:
    pE := pStart ;
    While pE <> Nil do
    Begin
        If (pE.El.ElType in [Chord,Rest,ReChord,Note])then
        Begin
            pE.El.No.BeamIn  := False ;
            pE.El.No.BeamOut := False ;
        End ;
        If pE=pEnd then pE := Nil else pE := pE.Next
    End ;
    // Report
    If pStart = Nil then
        Log('Failed to find start element '+IntToStr(StartNo)+' of stave '
               +IntToStr(StartStave)+' for removing beams')
    else
    If pEnd=Nil then
    Log('Removed beams from elements '+IntToStr(StartNo )+' of stave '+IntToStr(StartStave)+' to end')
    else
    Log('Removed beams from elements '+IntToStr(StartNo)+' of stave '+IntToStr(StartStave)
        +' to '+IntToStr(FinishNo)+' of stave '+IntToStr(FinishStave)) ;
  End
End ; { RemoveBeams }


Procedure SmallestStave ;
Begin
    With Music.Staves^ do
    Begin
       NoElements:= 2 ;
       Elements  := Nil ;
    End
End ; { SmallestStave }

Procedure SwapElements(StaveNo,N1,N2 : Integer; FileName:String) ;
// Swaps positions of elements N1 and N2 in stave StaveNo
Var pE, pPred, pN1, pN2, pPred1, pPred2, pT : pElementList ;
    Posn   : Integer ;
Begin
  Posn:= Pos(FileName,ReadFileName) ;
  If (Posn>0) and (Posn+Length(FileName)=Length(ReadFileName)+1)  then
  Begin
    Log('Swapping elements '+IntToStr(N1)+' and '+IntToStr(N2)+' of stave '+IntToStr(StaveNo)) ;
    If N1<>N2 then // else nothing to do.
    Begin
      pE    := Music.Staves.Elements ;
      pPred := Nil ;
      pPred1:= Nil ;
      pPred2:= Nil ;
      pN1   := Nil ;
      pN2   := Nil ;
      // Locate the two elements:
      While pE <> Nil do
      Begin
         If (pE.El.StaveNo=StaveNo) then
         Begin
            If pE.El.ElementNo=N1 then
            Begin
                pPred1 := pPred;
                pN1    := pE   ;
            End;
            If pE.El.ElementNo=N2 then
            Begin
                pPred2 := pPred;
                pN2    := pE   ;
            End
         End ;
         pPred := pE ;
         pE := pE.Next
      End ;

      If (pN1<>Nil) and (pN2<>Nil) then // The elements exist
      Begin
          // Swap the two elements:
          If (pPred1<>pN2) and (pPred2<>pN1) then // All 4 pointers are distinct
          Begin
              If pPred1<>Nil then pPred1.Next := pN2 else Music.Staves.Elements := pN2 ;
              If pPred2<>Nil then pPred2.Next := pN1 else Music.Staves.Elements := pN1 ;
              pT       := pN1.Next ;
              pN1.Next := pN2.Next ;
              pN2.Next := pT ;
          End
          else
          If pPred1=pN2 then      // N1 is the element after N2
          Begin
              If pPred2<>Nil then pPred2.Next := pN1 else Music.Staves.Elements := pN1 ;
              pN2.Next := pN1.Next ;
              pN1.Next := pN2 ;
          End
          else
          If pPred2=pN1 then      // N2 is the element after N1
          Begin
              If pPred1<>Nil then pPred1.Next := pN2 else Music.Staves.Elements := pN2 ;
              pN1.Next := pN2.Next ;
              pN2.Next := pN1 ;
          End ;
      End
    End
  End
End ; { SwapElements }

Procedure ExtendChord(StaveNo,StartNo,NoNotes:Integer; FileName:String) ;
// Extends chord with first note at StartNo so it has NoNotes notes.
// (It may also decrease the size of the chord.)
// This allows for some inadequacies in the MakeChord procedure.
Var pE, pPred, pChord : pElementList ;
    CurrentNo, Posn, Count   : Integer ;
Begin
  Posn:= Pos(FileName,ReadFileName) ;
  If (Posn>0) and (Posn+Length(FileName)=Length(ReadFileName)+1)  then
  Begin
    pPred := Nil ;
    pE    := Music.Staves.Elements ;
    While (pE<>Nil) and ((pE.El.StaveNo<>StaveNo)or(pE.El.ElementNo<>StartNo)) do
    Begin pPred := pE ; pE := pE.Next End ;
    pChord := pPred ;
    If pChord.El.ElType<>Chord then                    // Problem for MP118
       Log('Not found extendable chord at element '+IntToStr(StartNo)
            +' in stave '+IntToStr(StaveNo)+' of '+FileName)
    else
    Begin
        Log('Extending chord at '+IntToStr(StartNo)+' in stave '+IntToStr(StaveNo)+' to '
          +IntToStr(NoNotes)+' elements in '+FileName) ;
        CurrentNo := pChord.El.No.NoInChord ;
        pChord.El.No.NoInChord := NoNotes ;
        pE := pChord.Next ;
        Count := 0 ;
        If NoNotes<CurrentNo then
        Begin // Change InChord to false for some notes.
            While Count < CurrentNo do
            Begin
                If pE.El.ElType in [Rest,Note] then
                Begin
                    Inc(Count) ;
                    If Count > NoNotes then pE.El.No.InChord := False
                End ;
                If pE=Nil then Count := CurrentNo else pE := pE.Next
            End ;
        End
        else
        If NoNotes>CurrentNo then
        Begin // Change InChord to true for some notes.
            While Count < NoNotes do
            Begin
                If pE.El.ElType in [Rest,Note] then
                Begin
                    Inc(Count) ;
                    If Count > CurrentNo then pE.El.No.InChord := True
                End ;
                If pE=Nil then Count := NoNotes else pE := pE.Next
            End ;
        End
    End ;
  End ;
End ; { ExtendChord }


Procedure InsertRest(pS:pStave; ElNo:Integer; NT:NoteType; FileName:String) ;
// Fix For MP162. Inserts "rest" to make chord with given element.
// But needs to be grace note as Noteworthy won't accept rests in chords.
// Unused at the moment.
Var pE, pRest : pElementList ;
    Posn, P   : Integer ;
Begin
  Posn:= Pos(FileName,ReadFileName) ;
  If (Posn>0) and (Posn+Length(FileName)=Length(ReadFileName)+1)  then
  Begin
     pE := pS.Elements ;
     While (pE<>Nil) and (pE.El.ElementNo<>ElNo) do pE := pE.Next ;
     If (pE<>Nil) and (pE.El.ElType=Note) then
     Begin
        Log('Added rest at element '+IntToStr(ElNo)+' of stave '+IntToStr(pS.StaveNo)+
          ' in '+FileName) ;
        pRest := NewElement(Rest,pE.Next) ;
        With pRest.El do
        Begin
            StaveNo   := pS.StaveNo   ;
            ElementNo := pE.El.ElementNo ;
            VertPosn  := pE.El.VertPosn ;
            HorizPosn := pE.El.HorizPosn ;
            ElType    := Rest ;
        End ;

        P := pE.El.No.Pitch ;
        If P > 128 then P := 4 else P := 251 ;

        With pRest.El.No do
        Begin
            NoteSym:= NT ;
            Pitch  := P  ;
            Stem   := pE.El.No.Stem ;
        End ;
        pRest.Next := pE.Next ;
        pE.Next    := pRest
     End
     else
        Log('Not found element '+IntToStr(ElNo)+' of stave '+IntToStr(pS.StaveNo)+' in '+FileName) ;
  End ;
End ; { InsertRest }


Procedure LogElementTypes ;
Var pE : pElementList ;
Begin
    Log('Element list after processing the structure:') ;
    pE := Music.Staves.Elements ;
    While pE <> Nil do
    Begin
        Log('Element '+IntToStr(pE.El.ElementNo)+', type '+El2String(pE.El.ElType));
        pE := pE.Next ;
    End
End ; { LogElementTypes }


Procedure ModificationsBeforeWrite ;
{ This is called after all other processing to create NWC data structure,
  immediately before the NWC files is created.
  It is intended for altering specific, named files to correct them.
  All alterations are to the top stave of the concatenated systems.
  Procedures to choose from are:
      DeleteElement(Stave,Start:Integer; FileName:String) ;
      DeleteElements(StartStave,StartNo,FinishStave,FinishNo:Integer; FileName:String) ;
      RemoveGrace(StartStave,StartNo,FinishStave,FinishNo:Integer; FileName:String) ;
      RemoveBeams(StartStave,StartNo,FinishStave,FinishNo:Integer; FileName:String) ;
      SwapElements(StaveNo,N1,N2 : Integer; FileName:String) ;
      SmallestStave
}
Begin
      // DeleteElements(1,15,0,0,'MP245.NW') ;  // 15 works, 20 doesn't.
      // LogElementTypes ;

//      ExtendChord(3,11,3,'MP118.NW') ;
//      ExtendChord(3,63,3,'MP118.NW') ;
End ; { ModificationsBeforeWrite }

Procedure ModificationsAfterRead(pS : pStave) ;
// This is called before any other processing of stave pS,
// immediately after reading the stave.
// It is intended for altering specific, named files to correct them.
Begin
    // If pS.StaveNo=1 then InsertRest(pS,17,Crotchet,'MP162.NW') ; // This doesn't work yet.
End ;


{#############################################################################}
{#############################################################################}
{########################       Space Recovery       #########################}
{#############################################################################}
{#############################################################################}

Procedure DisposeOfStave(pS : pStave) ;
Var pEl, pEmpty : pElementList ;
Begin
    pEl := pS.Elements ;
    While pEl<>nil do
    Begin
        pEmpty := pEl ;
        pEl := pEl.next ;
        Dispose(pEmpty) ;
    End ;
    pS.Elements := Nil ;
    pS.NoElements := 0 ;
End; { DisposeOfStave }

Procedure DisposeOfMusic ;
// Recovers space used by data structure after processing a file.
Var pS,pEmpty  : pStave ;

Begin
    pS := Music.Staves ;
    While pS<>nil do
    Begin
        pEmpty := pS ;
        pS := pS.next ;
        DisposeOfStave(pEmpty) ;
        Dispose(pEmpty) ;
    End ;
    Music.NoStaves := 0 ;
    Music.Staves   := Nil ;
End ; { DisposeOfMusic }


{#############################################################################}
{#############################################################################}
{########################      Writing NWC File      #########################}
{#############################################################################}
{#############################################################################}


Procedure WriteNul;
Var Ch : Char    ;
Begin
    Ch := Nul ;
    BlockWrite(NewFile,Ch,1) ;
End ;

Procedure WriteCh(B:Byte);
Var Ch : Char    ;
Begin
    Ch := Chr(B) ;
    BlockWrite(NewFile,Ch,1)
End ;

Procedure WriteNuls(N:Integer);
Var I : Integer ;
Begin
    For I:=1 to N do WriteNul
End ; { WriteNuls }

Procedure WriteString(S : String) ;
Var I, L : Integer ;
Begin
    L := Length(S) ;
    For I := 1 to L do BlockWrite(NewFile,S[I],1) ;
    WriteNul ;
End ; { WriteString }

Procedure WriteChar(Ch:Char);     Begin BlockWrite(NewFile,Ch,1)  End ;

Procedure WritePlacement ;
// This is a procedure with default values for the
// "Expression placement" dialog box in Noteworthy Composer.
Const Default = 0 ;
Var Justification, Alignment, Placement : Integer ;
    PreserveWidth : Boolean ;
Begin
    PreserveWidth := False   ; // Unticked or ticked.
    Justification := Default ; // Left,Centre,Right
    Alignment     := Default ; // best fit, before sigs, after sigs, at next
    Placement     := Ord(PreserveWidth)+2*Justification+8*Alignment ;
    WriteCh(Placement) ;
End ; { WritePlacement }

Procedure WriteVisibility ;
// This is a procedure with default values for the
// "Visibility" dialog box in Noteworthy Composer.
Const Default = 0 ;
Var Show, Colour, Visibility : Integer ;
Begin
    Show       := Default ;     // Default, always, on top, never
    Colour     := Default ;     // Default, red, green, blue
    Visibility := Show+8*Colour ;
    WriteCh(Visibility) ;
End ; { WriteVisibility }

Procedure WriteElHeader(pE : pElementList; ElNo : Integer) ;
// Writes first 5 bytes of element with default visibility and placement.
Var VPosn : Integer   ;
Begin
    VPosn := (256+pE.El.VertPosn) mod 256 ;
    WriteCh(ElNo) ;
    WriteNul ;
    WriteVisibility ;
    WriteCh(VPosn) ;
    WritePlacement ;
End ; { WriteElHeader }

Procedure WriteClef(pE : pElementList) ;
// 7 bytes, e.g. #000#000#000#001#000#000#000 for bass clef.
// This is dependent of the ordering of clefs in the enumerated ClefType.
// 0 for treble clef.
Begin
    WriteCh(0) ;
    WriteNul ;
    WriteVisibility ;
    WriteCh(Ord(pE.El.Cl)) ;
    WriteNuls(3) ;
End ; { WriteClef }

Function AccidentalToInt(A : Accidentals) : Byte ;
// Values required for writing a byte of notes in NWC file.
Begin
    Case A of
        NoAcc      : AccidentalToInt := 5 ;
        Natural    : AccidentalToInt := 2 ;
        Flat       : AccidentalToInt := 1 ;
        Sharp      : AccidentalToInt := 0 ;
        DoubleSharp: AccidentalToInt := 3 ;
        DoubleFlat : AccidentalToInt := 4 ;
    else
                     AccidentalToInt := 5 ;
    End
End ; { AccidentalToInt }

Procedure WriteNote(pE : pElementList) ;
// 11 bytes for note/rest e.g. #008#000#002#000#016#000#000#000#003#013000#000#
// 13 bytes for chord/rest chord. This procedure writes the first 10/12 only.
// 13 bytes for rest-chord
// There seems to be no ghost note available.
// Used for chords, notes and rests.
// Unlike v55, RtShift is not required in v75.
Var Beam, Triplet, Stm, Tie, Slur,
    Dur, Ch5, Ch8, Slr, Pit, Ch11 : Byte ;
    LType : StaveElement ;
    ChosenSlur, ChosenTie : Position ;
    ExtraTie, ExtraSlur, PitProp : Boolean  ;
Begin
    ChosenSlur := Default ; // True if Slur selected to be above or below instead of default.
    ChosenTie  := Default ; // ditto for tie.
    LType := pE.El.ElType ;
    If LType in [Rest,Chord,ReChord,Note] then
    With pE.El.No do
    Begin
        If BeamIn and BeamOut then Beam := 2 else
        If BeamOut            then Beam := 1 else
        If BeamIn             then Beam := 3 else
                                   Beam := 0 ;
        If SlurIn and SlurOut then Slur := 3 else
        If SlurOut            then Slur := 1 else
        If SlurIn             then Slur := 2 else
                                   Slur := 0 ;
        If TieIn and TieOut   then Tie  := 3 else
        If TieOut             then Tie  := 2 else
        If TieIn              then Tie  := 1 else
                                   Tie  := 0 ;
        Case Stem of
           Up:                     Stm  := 1 ;  // This means upper or lower part
           Dn:                     Stm  := 2 ;
           else                    Stm  := 0 ;  // Auto choice of stem.
        End ;

        If TripletIn and TripletOut then Triplet := 2 else
        If TripletOut         then Triplet := 1 else
        If TripletIn          then Triplet := 3 else
                                   Triplet := 0 ;
        If Ord(NoteSym) = 0 then
        Begin // An error
           Dur := 0 ;
           Error(1,'NWC does not support breve notes; breve written as a semibreve') ;
           // ### Consider writing as two semibreves
        End
        else
        Dur := Ord(NoteSym) - 1 ;
        Ch5 := Beam + 4*(Triplet+4*Stm)
                 + 64*Ord(ChosenSlur=Below) ;  //   ok
        Ch8 := Ord(DoubleDot)+                 //   ok
                 2*( Ord(Accent=Staccato)+     //   ok
                 2*( Ord(Dotted)+              //   ok
                 2*( Tie+                      //   ok
                 4*( Ord(Accent=Accnt) )))) ;  //   ok
        If (LType in [Rest,Note]) and InChord then Ch8 := Ch8+64 ;
        If (LType in [Rest,Note,Chord]) and WithReChord then Ch8 := Ch8+128 ;
        ExtraTie :=  (ChosenTie<>Default) and TieOut ;
        ExtraSlur:=  (ChosenSlur<>Default) and SlurOut ;
        Slr := Slur + 4*Ord(Accent=Tenuto)+32*Ord(Grace)
                    + 64*Ord(ExtraTie) + 128*Ord(ExtraSlur) ;
        If LType in [Rest,ReChord] then Pit := 0 else
           If Pitch > 0 then Pit := 256-Pitch
           else              Pit := -Pitch ;
        PitProp := ((Pit < 128) and (((Tie=0))or(ChosenTie=Default))) or
                    ((Tie<>0)and(ChosenTie=Below)) ;
        If LType in [Rest,ReChord] then Ch11 := 0   // Always 0 for rest/rechord
        else
        If LType = Chord then Ch11 := 8*Ord(PitProp) + 5 // No modification for accidentals
        else                                    // Sometimes the value for Chord has the wrong mult of 8.
           Ch11 := {Ord(WithReChord)*Ch9 +} 8*Ord(PitProp)
                    + AccidentalToInt(Accidental) {+ 16*Ord(RtShift)} ; //  RtShift was needed in v55.
               // AccidentalToInt(Accidental) = 5 for normal notes (i.e. no associated accidental)

        // Write element type:
        Case LType of
           Note  : WriteCh(8) ;
           Rest  : WriteCh(9) ;
           Chord : WriteCh(10);
           ReChord:WriteCh(18);
        End ;
        WriteNul ;
        WriteVisibility ;
        WriteCh(Dur) ; // Duration
        WriteNul ;     // OK
        WriteCh(Ch5) ; // OK
        WriteNul ;     // OK
        WriteCh(Ch8) ; // OK
        WriteCh(Slr) ; // Slur + Tenuto + Grace
        WriteCh(Pit) ; // OK
        WriteCh(Ch11); // OK except sometimes for chords
        If pE.El.ElType in [Chord,ReChord] then
        Begin WriteCh(pE.El.No.NoInChord); WriteNul End ;
    End
    else
        Error(2,'Trying to write note when no note present.') ;
End ; { WriteNote }



Procedure WriteText(pE : pElementList; pS : pStave) ;
{  NB: To correct: Where "text" contains ord values above 128, the music font is intended:
                #130 is "A" ... #136 is "G"
                #128 is a flat
   6 bytes+<text>+NUL, e.g. #017#000#000#242#008#001

   See WriteFonts for details of fonts to be used in NWC file. Fonts nos used in NWC file are:
        0 for "Staff Italic". This is used for "da capo", "D.S." etc  12pt bold italic
        1 for "Staff Bold".   This is used for bar numbers.           10pt bold
        2 for "Staff Lyric"   This is used for any lyrics.            12pt italic
        3 for "Page Title Text" This is used for the title.           20pt bold
        4 for "Page Text"                                             12pt plain
        5 for "Page Small Text"                                        9pt plain
}
    Function NWCFont(NWFNo : Integer) : Integer ;
    Begin // See ReadText
        Case NWFNo of
           1 : NWCFont := 5 ; // small plain    9pt    -> Page Small Text
           2 : NWCFont := 4 ; // medium plain  12pt    -> Staff Bold
           3 : NWCFont := 2 ; // medium italic 12pt    -> Staff Lyric
           4 : NWCFont := 0 ; // big italic    15pt    -> Staff Italic
           5 : NWCFont := 3 ; // very large plain 20pt -> Page Title Text
           else NWCFont:= 4 ; // shouldn't occur 12pt  ->
        End ;
    End ; { NWCFont }
Var FontNo, SI : Integer   ;
    S          : String    ;
Begin
    SI := pE.El.Txt.StringIndex ;
    S  := Strings[SI] ;
    FontNo := NWCFont(pE.El.Txt.FontNo) ;
    WriteElHeader(pE,17) ;
    WriteCh(FontNo) ;
    WriteString(S) ;
End ; { WriteText }


Procedure WriteSymbol(pE : pElementList) ;
Var FontNo : Integer ;
    Sym    : Byte    ;
Begin
    Sym := pE.El.Sym ;
    If not PitchedSym(Sym) then pE.El.VertPosn := 244 ;
    FontNo := 6 ; // small italic
    WriteElHeader(pE,17) ;
    WriteCh(FontNo) ;
    WriteString(SymToStr(Sym)) ;
End ; { WriteSymbol }

Procedure WriteTempo(pE : pElementList) ;
// 8chars + Optional text; to complete.
Begin
    WriteElHeader(pE,6) ;
    WriteNuls(3+1) ;
End ; { WriteTempo }

Procedure WriteMPC(pE : pElementList) ;
// 13+4*6 chars; to complete.
Begin
    WriteElHeader(pE,13) ;
    WriteNuls(8+4*6) ;
End ; { WriteMPC }

Procedure WritePatch(pE : pElementList) ;
// 11 chars; to complete.
Begin
    WriteElHeader(pE,4) ;
    WriteNuls(6) ;
End ; { WritePatch }

Procedure WriteEnding(pE : pElementList) ;
Var No : Integer ;
// For endings a,b,c,... use No = 2^(a-1)+2^(b-1)+2^(c-1)...
Begin
    No := 0 ;
    WriteCh(3) ;
    WriteNul ;
    WriteVisibility ;
    WriteCh(No) ;
    WriteNul ;
End ; { WriteEnding }

Procedure WritePerfStyle(pE : pElementList) ;
// e.g. #016#000#000#248#000#003
Var PF    : PerfStyle ;
Begin
    PF := pE.El.PF ;
    If PF <> NotFound then
    Begin
        WriteElHeader(pE,16) ;
        WriteCh(Ord(PF)) ;
    End
End ; { WritePerfStyle }


Procedure WriteVolume(pE : pElementList) ;
// 10 chars, e.g. #007#000#000#018#253#045#000#000#255#255
//                #007#000#000#010#008#099#000#000#255#255
Var OverDefNoteVel, OverDefMIDIvol,
    DefVol, NoteVel, MidiDef, MidiVol,
    Offset                       : Integer ;
    ChangeNoteVel, ChangeMIDIvol : Boolean ;

Begin
    Case Ord(pE.El.V) of
      0 : DefVol := 10 ;
      1 : DefVol := 30 ;
      2 : DefVol := 45 ;
      3 : DefVol := 60 ;
      4 : DefVol := 75 ;
      5 : DefVol := 92 ;
      6 : DefVol :=108 ;
      7 : DefVol :=127 ;
      else DefVol:= 75 ;
    End ;
    MidiDef := -1 ;
    ChangeNoteVel := False ;
    ChangeMIDIvol := False ;
    OverDefNoteVel := DefVol  ; // change to what's in original NW file
    OverDefMIDIvol := MidiDef ; // change to what's in original NW file
    Offset := 32*Ord(ChangeNoteVel) + 64*Ord(ChangeMIDIvol) ;
    If ChangeNoteVel then NoteVel := OverDefNoteVel else NoteVel := DefVol ;
    If ChangeMIDIvol then MidiVol := OverDefMIDIvol else MidiVol := MidiDef ;
    WriteElHeader(pE,7) ;
    WriteCh(Offset+Ord(pE.El.V)) ;
    WriteCh(NoteVel) ;
    WriteNul ;
    WriteCh((MidiVol+256) mod 256) ;
    If MidiVol>=0 then WriteCh(0) else WriteCh(255); // Unsure about this
End ; { WriteVolume }

Procedure WriteVolVar(pE : pElementList) ;
//  e.g. #015#000#000#247#000#004
Var VV    : VolVarType ;
Begin
    VV    := pE.El.VV ;
    If VV<>Unfound then
    Begin
        WriteElHeader(pE,15) ;
        WriteCh(Ord(VV)) ;
    End;
End ; { WriteVolVar }

Procedure WriteBarline(pE : pElementList) ;
// 5 bytes, e.g. #002#000#00n#001
// n is given by ordering enumerated type BarLine to coincide with NWC order.
Var RepNo, BarLineStyle : Byte ;
    Bar : BarLine ;
Begin
    Bar := pE.El.BL ;
    BarLineStyle:= Ord(Bar) ;
    RepNo := 1 ; //  Number of times barline is passed due to repetitions.
                 //  1 may be normal value away from repeat bars and endings and signs.
                 //  May only be critical for end of local repeat.
    If Bar in [LocalRepeatClose,MasterRepeatClose] then RepNo := 2 ;
    WriteCh(2) ;
    WriteNul ;
    WriteVisibility ;
    WriteCh(BarLineStyle) ; // i.e. single,double, repeat open,...
    WriteCh(RepNo) ;
End ; { WriteBarline }


Function LogTwo(N:Integer) : Integer ;
Var Log : Integer ;
Begin
    Log := 0;
    While N>1 do Begin Inc(Log); N := N div 2 End ;
    LogTwo := Log
End ; { LogTwo }

Procedure WriteTimeSig(pE : pElementList) ;
// 9 bytes, e.g.  5 0 0 7 0 2 0 0 0 is 7/4; 2 = log_2(4)
//                5 0 0 4 0 2 0 1 0 is Common Time
Begin
    WriteCh(5) ;
    WriteNul ;
    WriteVisibility ;
    WriteCh(pE.El.TS.Top) ;
    WriteNul ;
    WriteCh(LogTwo(pE.El.TS.Bot)) ;
    WriteNul ;
    If pE.El.TS.Common then WriteCh(1)
    else
    If pE.El.TS.AllaBreve then WriteCh(2)
                       else WriteCh(0) ;
    WriteNul ;
End ; { WriteTimeSig }

Function NoFlats(N:Integer) : Integer ;
Var I, Count : Integer ;
Begin
    Count := 0 ;
    For I := 1 to N do
    Case I of
    1 : Count := Count +  2 ; // B
    2 : Count := Count + 16 ; // E
    3 : Count := Count +  1 ; // A
    4 : Count := Count +  8 ; // D
    5 : Count := Count + 64 ; // G
    6 : Count := Count +  4 ; // C
    7 : Count := Count + 32 ; // F
    End ;
    NoFlats := Count
End ; { NoFlats }

Function NoSharps(N:Integer) : Integer ;
Var I, Count : Integer ;
Begin
    Count := 0 ;
    For I := 1 to N do
    Case I of
    1 : Count := Count + 32 ; // F
    2 : Count := Count +  4 ; // C
    3 : Count := Count + 64 ; // G
    4 : Count := Count +  8 ; // D
    5 : Count := Count +  1 ; // A
    6 : Count := Count + 16 ; // E
    7 : Count := Count +  2 ; // B
    End ;
    NoSharps := Count
End ; { NoSharps }


Procedure WriteKeySig(pE : pElementList; pS: pStave) ;
{ This might count as two stave elements if cancelling naturals are counted separately.
  13 bytes using binary coding for positions of symbols, e.g.
  1  0 19  0  ...         19 in binary is 0010011. Using these positions for GFEDCBA, it means E,B,A.
  1  0  0  0 125 0  0  0  0  0  0  0  is 6 initial sharps, no lines needing pre-adjusting.
  1  0 95  0  0  0  0  0  0  0 125 0  is 6 flats, with 125 cancelling previous 6 sharp pitch changes
  1  0  0  0  0  0  0  0 95  0  0  0  is naturals in 95 position, to cancel previous key sig
  1  0  0  0 36  0  0  0  0  0 36  0  is two sharps, where latter 36 indicates lines that need to retune.
                                         Need not be the same as first 36?
  1  0  0  0 36  0  0  0  0  0  0  0  is two sharps, where lines don't need resetting with naturals
  1  0  0  0  0  0  0  0  0  0 100 0  is three naturals only.  100 = 64+32+4: GFC
  The bytes seem to be grouped as two sets of three: 1 0 <three> 0 0 0 <three> 0.
  The the sets of three are related to flat/natural/sharp, the second three being naturals
  to cancel previous sharps/flats, and the first three being new flats/sharps.
}
Var pPrev, pLastKeySig : pElementList;
Begin
    pPrev := pS.Elements ;
    pLastKeySig := Nil ;
    While (pPrev<>nil) and (pPrev<>pE) do
    Begin
        If pPrev.El.ElType = KeySig then pLastKeySig := pPrev ;
        pPrev := pPrev.Next
    End ;
    WriteCh(1) ;
    WriteNul ;
    WriteVisibility ;
    // Bytes 4..6 Details of new flats/sharps to be written:
    If pE.El.KS.IsFlats  then WriteCh(NoFlats (pE.El.KS.Number)) else WriteNul ;
    WriteNul ; // This byte would record a key sig of naturals!
    If pE.El.KS.IsSharps then WriteCh(NoSharps(pE.El.KS.Number)) else WriteNul ;
    WriteNuls(3) ;  // The middle Nul is sometimes 1, not known why.
    // Bytes 10..12 Details of cancelling naturals to be written:
    If (pLastKeySig<>nil) and pLastKeySig.El.KS.IsFlats
       then WriteCh(NoFlats (pLastKeySig.El.KS.Number)) else WriteNul ;
    WriteNul ;
    If (pLastKeySig<>nil) and pLastKeySig.El.KS.IsSharps
       then WriteCh(NoSharps(pLastKeySig.El.KS.Number)) else WriteNul ;
    WriteNul
End ; { WriteKeySig }


Procedure WriteFlow(pE : pElementList; pS : pStave) ;
// Flow directions, 7 bytes, e.g. #012#000#000#248#009#001#000 segno
//                                #012#000#000#249#009#000#000 coda
//                                #012#000#000#006#001#002#000 fine
//                                #012#000#000#251#000#004#000 DC
Var Flow         : FlowType ;
Begin
    Flow := pE.El.FL ;
    If Flow <> NoFlow then
    Begin
        pE.El.VertPosn := 248 ;
        WriteElHeader(pE,12) ;
        WriteCh(Ord(Flow)) ;
        WriteNul   ;
    End;
End ; { WriteFlow }

Procedure WriteTempoVar(pE : pElementList; pS : pStave) ;
//  7 chars. Example: #014#000#000#004#000#006#003
Var Delay : Integer ;
    TV    : TempVarType ;
Begin
    TV  := pE.El.TmpVar ;
    If TV<>UnknownTV then
    Begin
        If TV = Ferm then pE.El.VertPosn:= pE.El.VertPosn-5 ;  // VPosn is that of note head below it.
        Delay := 0 ;
        WriteElHeader(pE,14) ;
        WriteCh(Ord(TV)) ;
        WriteCh(Delay) ;
    End;
End ; { WriteTempoVar }

Procedure WritePed(pE : pElementList; pS : pStave) ;
// 6 bytes  e.g. #011#000#000#254#000#001 ped dn
//               #011#000#000#006#001#000 ped up
Begin
    WriteElHeader(pE,11) ;
    WriteCh(Ord(pE.El.Pd)) ;
End ; { WritePed }


Procedure WriteElement(pE : pElementList; pS : pStave) ;
Begin
    Case pE.El.ElType of                    // Element type number:
       Clef     : WriteClef    (pE) ;       //   0
       Chord,                               //  10
       ReChord,                             //  18
       Note,                                //   8
       Rest     : WriteNote    (pE) ;       //   9
       Text     : WriteText    (pE,pS) ;    //  17
       Bar      : WriteBarline (pE) ;       //   2
       TimeSig  : WriteTimeSig (pE) ;       //   5
       KeySig   : WriteKeySig  (pE,pS) ;    //   1
       Flow     : WriteFlow    (pE,pS) ;    //  12 - segno
       TempoVar : WriteTempoVar(pE,pS) ;    //  14
       Pedal    : WritePed     (pE,pS) ;    //  11
       Symbol   : WriteSymbol  (pE) ;       //  (17 - changed to text)
       Vol      : WriteVolume  (pE) ;       //   7
       VolVar   : WriteVolVar  (pE) ;       //  15
       PStyle   : WritePerfStyle(pE);       //  16 ?
       // Tempo : WriteTempo   (pE) ;       //   6
       // MPC   : WriteMPC     (pE) ;       //  13
       // Ending: WriteEnding  (pE) ;       //   3
       // Patch : WritePatch   (pE) ;       //   4
       Unknown  : ;
    else
       Error(63,'Failed to write element of type '+El2String(pE.El.ElType)+' into file.')
       // Other cases should have been explicitly dealt with in another way.
    End ;
End ; { WriteElement }

Function NoElements(pS : pStave) : Integer ;
{ Chords complicate the answer to this:
        2 for an empty stave (the end bars),
        Count 1 for chord element (whichever the stem direction), 0 for each of its members,
        1 for a time or key sig.
  Some elements appear only in the top stave of a system, e.g. tempo, but they must be counted
  in all staves of the system. (##### This last detail is yet to be implemented.)
  Every StaveElement which has a Write routine in WriteElement (above) must be counted.
  = (NewStave,Clef,Bar,Note,Rest,ReChord,Chord,TimeSig,KeySig,Text,ALine,Vol,VolVar,
     Pedal,Tempo,TempoVar,Flow,Style,Symbol,Accident,SlurPt,Accent,Unknown) ;
}
Var pE : pElementList ;
    Count : Integer ;
Begin
    pE := pS.Elements ;
    Count := 2 ;
    While pE<>nil do
    Begin
        Case pE.El.ElType of
           Clef, Bar, Note, Rest, Text, TimeSig, KeySig, Pedal, PStyle,
           TempoVar, VolVar, Flow, Symbol, Vol
                    : Inc(Count) ;
           Chord    : // Adjust for number of note/rest elements in chord:
                      Count := Count + 1 - pE.El.No.NoInChord ;
        else  {Rechord,Accent,SlurPt,Accident,Unknown}
               // Contribute 0 as in ReChord or
               // ignored because not implemented - checklist in WriteElement().
        End ;
        pE := pE.Next
    End ;
    NoElements := Count
End ; { NoElements }

Procedure WriteLyrics(pS : pStave) ;
Var LyricSize, I, J : Integer ;
    LyrJ : Byte ;
Begin
    WriteNul ;
    WriteCh(4) ;
    For I := 1 to pS.Stv.NoLyrics do
    Begin
        LyricSize := 0 ; // To be fixed when needed.
        WriteCh(LyricSize mod 256) ;
        WriteCh(LyricSize div 256) ;
        WriteNul ;
        WriteCh(1) ;
        LyrJ := 0 ; // To be fixed when needed.
        For J := 1 to 1024 do If J <= LyricSize then WriteCh(LyrJ) else WriteNul ;
    End ;
End ; { WriteLyrics }

Function MaxVertPos(pS : pStave) : Integer ;
// Vert Posn increases downwards.
Const TextElements = [Text,Vol,VolVar,Pedal,{Tempo,}TempoVar,Symbol,Flow,Pedal,Accent] ;
Var pE  : pElementList ;
    Max, Min, VP : Integer ;
Begin
    Max := -128 ;
    Min := +128 ;
    pE := pS.Elements ;
    While pE <> Nil do
    Begin
        If pE.El.ElType in TextElements then
        Begin
        VP := pE.El.VertPosn ;
        If VP > 128 then VP  := VP-256 ;
        If VP > Max then Max := VP ;
        If VP < Min then Min := VP ;
        End;
        pE := pE.Next
    End;
    Min := -Min ;
    If Min+4 < 12 then Min := 12 else Min := Min+4 ; // +4 for height of a character.
         // Should replace with something dependent on font height.
    MaxVertPos := Min ;
End ; { MaxVertPos }

Procedure WriteStaveHeading(pS : pStave) ;
Const Name = 'Staff-' ;
      Group= 'Standard' ;
Var // EndingBar: Byte    ;
    // Visible  : Boolean ;
    StaffName   : String  ;
    Upper, Lower, ClosingBar, StaveNo, Instrument, VisualStyle,
    Transposition, PartVol, StereoPan, UserFontNo, NoLyrics, ChannelNo,
    Controller0, Controller32, ElementCount  : Integer ;
    // Flute, PianoLeft, PianoRight
    Muted, SelPatchBnk   : Boolean ;
Begin
    // General:
    StaveNo      := pS.StaveNo ;
    StaffName    := Name+IntToStr(StaveNo) ;
    UserFontNo   := 5+StaveNo ;
    Upper        := MaxVertPos(pS) ;
    Lower        := 12 ;
    ClosingBar   := Ord(pS.Stv.FinalBar in [LocalRepeatClose,MasterRepeatClose])  ;
    NoLyrics     := pS.Stv.NoLyrics ;
    ElementCount := NoElements(pS) ;
    Instrument   :=  0 ;
    VisualStyle  := Ord(pS.Stv.StType) ; // 0 for Standard, 1 for upper grand, 2 for lower grand, 3 for orchestral
    Transposition:=  0 ;
    PartVol      :=127 ;
    StereoPan    := 64 ;
    ChannelNo    :=  1 ;
    Muted        := False ;
    SelPatchBnk  := False ;
    Controller0  :=  0 ;
    Controller32 :=  0 ;

    WriteString(StaffName) ;  // OK
    WriteString(Group) ;      // OK
    // 1st "tetrad" (the zero ending previous WriteString is part of the tetrad.)
    // The tetrad is sometimes 0 0 1 0; the '1' may mean not visible.
    // St Header 1:
    WriteCh(ClosingBar);      // 0=section close, 1=master repeat close, 2=single bar, 3=double bar;
    WriteCh(Ord(Muted)) ;
    WriteNul ;                // OK
    // 2nd & 3rd tetrads:
    WriteCh(ChannelNo-1) ;    // OK   ############# Problem for MP115, MP119 StaveNo=0.
    WriteNul ;                // OK
    WriteCh(0) ;              // =1 in Moonlite, both staves. No idea about meaning.
    WriteNul ;                // OK
    WriteCh(Ord(SelPatchBnk)) ;
    WriteNul ;                // OK
    WriteCh(128*(Controller0 mod 2)+Controller32) ;
    WriteCh(Controller0 div 2) ;
    // 4th tetrad: Instrument
    WriteCh(Instrument) ;
    // St Header 2:
    WriteNul ;
    WriteCh(VisualStyle) ;           // see above; affects end bars.
    WriteNul ;
    // 5th tetrad: Stave Size
    WriteCh((256-Upper) mod 256) ;  // OK
    WriteCh(255) ;                  // OK
    WriteCh(Lower) ;                // OK
    WriteNul ;                      // OK
    WriteCh(5) ;
    // 6th tetrad: unknown
    WriteNuls(2) ;                  // Bytes 1, 3 and 4 are sometimes non-nul.
    WriteCh((Transposition+256) mod 256) ;
    WriteNul ;                      // OK
    // St Header 3:
    // 7th tetrad: volume
    WriteCh(PartVol) ;              // OK (This is mod 256, and can be set to "-1".
    WriteNul ;                      //
    WriteCh(StereoPan) ;            // OK (This is mod 256, and can be set to "-1".
    WriteNuls(2) ;                  // Transposition? (This is 255 if next byte is 255)
    // optional tetrad: lyrics
    If NoLyrics>0 then
    Begin
        WriteCh(1) ; WriteNul ; WriteCh(NoLyrics) ; WriteNul ;
        WriteNuls(4) ;
        WriteLyrics(pS) ;
    End
    else WriteNuls(4) ;
    // 9th tetrad: #elements in stave
    WriteNuls(2) ;
    WriteCh(ElementCount mod 256) ; // This is 1 too many for HereIAm.nw
    WriteCh(ElementCount div 256) ;
End ; { WriteStaveHeading }


Procedure WriteStave(pS : pStave) ;
Var pE : pElementList ;
Begin
    WriteStaveHeading(pS) ;
    pE := pS.Elements ;
    While pE<>Nil do
    Begin
        WriteElement(pE,pS) ;
        pE := pE.Next ;
    End
End ; { WriteStave }


{#############################################################################}
{#############################################################################}
{#######################     Write NWC File Heading    #######################}
{#############################################################################}
{#############################################################################}


Procedure WriteFont({Font:String;}Style,Size:Integer) ;
Const Font = 'Times New Roman' ;
// A String (terminated by #000) plus 4 bytes e.g. #001#024#000#000
      Procedure WriteFontStyle(I:Byte); Begin WriteCh(I) End ;
      Procedure WriteFontSize (I:Byte); Begin WriteCh(I) End ;
Begin
    WriteString(Font) ;
    WriteFontStyle(Style) ;
    WriteFontSize(Size) ;
    WriteNuls(2) ;
End ; { WriteFont }

Procedure WriteFonts ;
// Six system fonts, four user fonts. [Six in later versions.]
Begin
    WriteFont(3,12) ; // Staff Italic  BoldItalic Used for "da capo", "D.S.", "cresc." etc
    WriteFont(1,10) ; // Staff Bold    Bold Used for bar numbers, #repeats, etc
    WriteFont(2,12) ; // Staff Lyric   Italic
    WriteFont(1,20) ; // Page Title Text Bold
    WriteFont(0,12) ; // Page Text     Plain
    WriteFont(0, 9) ; // Page Small Text Plain
    WriteFont(0, 8) ; // User 1
    WriteFont(1, 9) ; // User 2        Bold
    WriteFont(2, 9) ; // User 3        Italic
    WriteFont(3, 9) ; // User 4        BoldItalic
    WriteFont(1,12) ; // User 5        Bold
    WriteFont(0,10) ; // User 6        Plain
End ; { WriteFonts }

Procedure WriteMarginTetrads ;
{   e.g. N   N   _ #000  F   2 #000#000#000#001#000
    This all seems to be working correctly as expected. }
Const Margin = '0.50000000' ; // 10 chars; not found the bit which determines cms/ins
Var ExtendLastSystem,
    IncreaseNoteSpacing,
    Version95,                 // Year of version.
    FalseCh,
    Two            : Char ;
    MeasureStyle,
    MeasureStart   : Byte ;
    MirrorMargins  : Boolean ;
    Margins        : String ;
    Zoom           : Integer ; // Size of stave on screen in mm.
Begin
    ExtendLastSystem    := 'N' ;
    IncreaseNoteSpacing := 'Y' ;
    Version95           := Chr(95) ;
    FalseCh             := 'F' ;
    Two                 := '2' ;
    MeasureStyle        :=  0  ; // 0=none,1=plain,2=boxed,3=circled
    MeasureStart        :=  1  ; // possibly 0 for no bar numbers
    Margins             :=  Margin+Space+Margin+Space+Margin+Space+Margin ;
    MirrorMargins       := False ;
    Zoom                :=  8 ;
    // 11 bytes between "comment" and Margin strings:
    // Tetrad 1:
    WriteChar(ExtendLastSystem) ;    // Always Y/N
    WriteChar(IncreaseNoteSpacing) ; // Always Y/N
    WriteChar(Version95) ;
    WriteNul ;
    // Tetrad 2:
    WriteChar(FalseCh) ;             // Always 'F'
    WriteChar(Two) ;                 // Always '2'
    WriteNul ;
    WriteCh(MeasureStyle) ;          // 0 in all my files.
    // Last 3 byte:
    WriteNul ;
    WriteCh(MeasureStart) ;          // 1 in all my files
    WriteNul ;
    // Margin string:
    WriteString(Margins) ;           // Works
    // Final 3 bytes:
    // WriteCh(Ord(MirrorMargins)) ;    // Correct
    WriteNul ;
    WriteCh(Zoom) ;                  // Correct - one more than staff size on screen in points.
    WriteNul ;
End ; { WriteMarginTetrads }

Procedure Write9Tetrads ;
// Mostly zeroes e.g. MP158tri: #000#005#000#001  #000.....#000  #020#000
// Two bytes, then 32 nuls, then two bytes.  Probably all correct.
Var StaffSize, I,
    Visible,       // Binary of this indicates which staves are visible.
    NoStaves  : Integer ;
    Layering  : Boolean ;
Begin
    StaffSize := 24 ; // This may mean 3 pixels per position on stave, as used below.
    Layering  := False ;
    NoStaves  := Music.NoStaves ;
    Visible   := 0 ;
    For I := 1 to NoStaves do Visible := 2*Visible+1 ;
    WriteCh(Visible)   ;   // Meaning may depend on previous parameters.
    WriteNuls(31)      ;
    If Layering then WriteCh(255) else WriteCh(0) ;
    WriteCh(StaffSize) ; // Staff size when printed.
                         // Sometimes a different value in pre-55 versions.
    WriteNul ;
End ; { Write9Tetrads }

Procedure Write6 ;
// Final 6 bytes of file heading.  Info from "File/Page Setup"
// Examples: #255#000#000#000#003#000
//           #001#000#000#000#002#000
Var NoStaves, FirstPageNumber, StaffLabels  : Integer ;
    Title     : Boolean ;
Begin
    FirstPageNumber := 0 ; // 0 for no page numbers, else start number.
    StaffLabels     := 0 ; // 0..3 is acceptable range.
    NoStaves := Music.NoStaves ;
    Title    := True ;
    If Title then WriteCh(255) else WriteCh(0) ;  // Not sure of meaning.
    WriteCh(StaffLabels) ;
    WriteCh(FirstPageNumber) ;
    WriteNul ;
    WriteCh(NoStaves) ;
    WriteNul ;
End ; { Write6 }

Procedure WriteHeading ;
// All the strings can be changed without consequence except for NWA and NWC.
Const NWA     = '[NoteWorthy ArtWare]' ;
      NWC     = '[NoteWorthy Composer]' ;
      Copy1   = 'Copyright  1995 <Song Author Name>' ;
      // see http://homepages.rootsweb.ancestry.com/~george/ansi_ascii_character_chart.html
      // for chars numbered 128 to 255.
      Copy2   = 'All Rights Reserved' ;
      RegUser = '' ; // Acceptable: about half of all cases.
      // #067#079#076#073#078#032#068#032#087#065#076#084#069#082
      RegCode = '' ; // Acceptable: blank is always the value for blank RegUser.
                     // Otherwise unknown function of file content.
      // #065#097#110#119#104
      Author  = '' ;
      NWCVersionNo = 75 ; // Version number of NWC format used in creating NWC file.
                          // This can't be changed without changing the output procedures.
Var   Copy3, Copy4, Comment : String  ;

Begin { WriteHeading }
    Comment := Head ;
    Copy3   := Foot ;
    Copy4   := ''   ;
    If Head<>'' then Log('   NW file header written as "Comment" in File/Info') ;
    If Foot<>'' then Log('   NW file footer written as "Copyright Notice 1" in File/Info') ;
    WriteString(NWA);
    WriteNuls(2)    ;
    WriteString(NWC);
    WriteCh(NWCVersionNo) ; // In my files one of: #020, #050, #055, #075 DO NOT CHANGE!
                            // The file formats are different for other versions.
    WriteCh(1) ;
    WriteCh(2) ;
    WriteNuls(3) ;
    WriteString(RegUser) ;
    WriteString(RegCode) ;
    WriteNuls(6) ;
    WriteCh(0) ; WriteCh(0) ;
    WriteCh(16) ;
    { Unknown variable. May be a count of #bytes around the  WriteNuls(7)
      in WriteStaveHeading. Always one of 4, (6), 8, 9, or 16.
      It's a critical parameter to get right. }
    WriteCh(0) ;
    WriteString(Strings[Music.Title.StringIndex])  ;
    WriteString(Author) ;
    WriteString(Copy3)  ;
    WriteString(Copy4)  ;
    WriteString(Comment);
    WriteMarginTetrads  ;
    Write9Tetrads ;
    WriteFonts ;
    Write6 ;
End ; { WriteHeading }


{#############################################################################}
{#############################################################################}
{##################   Consistency Check of Original File   ###################}
{#############################################################################}
{#############################################################################}


Procedure SwapBeamPtrs(pS : pStave; pE, pF : pElementList) ;
// Pre-condition:  Beam goes from pE to pF
// Post-condition: Beam goes from pF to pE, with related beams updated.
Var pG, pH : pElementList ;
    StNo, pENo, pFNo, pGNo, pHNo : Integer ;
Begin
    StNo := pE.El.StaveNo   ;
    pENo := pE.El.ElementNo ;
    pFNo := pE.El.ElementNo ;

    // Get predecessor pG of pE:
    pGNo := 0  ;
    If pE.El.No.BeamIn then
    Begin
            pGNo := pE.El.No.BeamFrom ;
            pG   := pS.Elements ;
            While (pG<>Nil) and ((pG.El.StaveNo <> StNo) or (pG.El.ElementNo <> pGNo)
                          or (pG.El.ElType<>Note))
            do pG := pG.Next ;
            If pG<>Nil then // Make pG point to pF instead of pE,
               pG.El.No.BeamTo   := pFNo ;
    End ;
    // Make pF point back to pG instead of pE:
    pF.El.No.BeamFrom := pGNo ;
    If pGNo=0 then pF.El.No.BeamIn := False ;

    // Get Successor pH of pF:
    pHNo := 0  ;
    If pF.El.No.BeamOut then
    Begin
            pHNo := pF.El.No.BeamTo ;
            pH   := pS.Elements ;
            While (pH<>Nil) and ((pH.El.StaveNo <> StNo) or (pH.El.ElementNo <> pHNo)
                          or (pH.El.ElType<>Note))
            do pH := pH.Next ;
            If pH<>Nil then // Make pH point back to pE instead of pF,
               pH.El.No.BeamTo   := pENo ;
    End ;
    // Make pE point to pH instead of pF:
    pE.El.No.BeamTo := pHNo ;
    If pHNo=0 then pE.El.No.BeamOut := False ;

    // Change direction of beam between pE and pF:
    pE.El.No.BeamIn  := True ; pE.El.No.BeamFrom := pFNo ;
    pF.El.No.BeamOut := True ; pF.El.No.BeamTo   := pENo ;
End ; { SwapBeamPtrs }

Procedure SwapBeams(pS: pStave; pE : pElementList) ;
// Beams are not listed in order; they are reconfigured to be so.
// Pre-condition pE.El.ElType = Note and
//     EITHER    pE.El.No.BeamOut and pE.El.No.BeamTo  < pE.El.ElementNo
//         OR    pE.El.No.BeamIn  and pE.El.No.BeamFrom> pE.El.ElementNo
Var pF : pElementList ;
    StNo, pENo, pFNo : Integer ;
Begin
    StNo := pE.El.StaveNo   ;
    pENo := pE.El.ElementNo ;
    If pE.El.No.BeamOut and (pE.El.No.BeamTo < pENo) then
    Begin // Get ptr pF to pE.El.No.BeamTo:
        pFNo := pE.El.No.BeamTo ;
        pF := pS.Elements ;
        While (pF<>Nil) and ((pF.El.StaveNo <> StNo) or (pF.El.ElementNo <> pFNo)
                          or (pF.El.ElType<>Note))
        do pF := pF.Next ;
        If pF=Nil then
        Begin
            Error(68,'Missing beam end from element '+IntToStr(pENo)
                      +' to element '+IntToStr(pFNo)
                      +', in stave '+IntToStr(StNo));
            pE.El.No.BeamOut := False ;
        End
        else
        If (not pF.El.No.BeamIn) or (pF.El.No.BeamFrom<>pENo) then
        Begin
            Error(69,'Improperly matched beam ends between elements '
                      + IntToStr(pENo) + ' and ' + IntToStr(pFNo)
                      + ', in stave ' + IntToStr(StNo));
        End
        else
            SwapBeamPtrs(pS, pE, pF)
    End ;

    If pE.El.No.BeamIn and (pE.El.No.BeamFrom> pENo) then
    Begin // Get ptr pF to pE.El.No.BeamFrom:
        pFNo := pE.El.No.BeamFrom ;
        pF := pS.Elements ;
        While (pF<>Nil) and ((pF.El.StaveNo <> StNo) or (pF.El.ElementNo <> pFNo)
                          or (pF.El.ElType<>Note))
        do pF := pF.Next ;
        If pF=Nil then
        Begin
            Error(70,'Missing beam start from element '+IntToStr(pFNo)
                      +' to element '+IntToStr(pENo)
                      +', in stave '+IntToStr(StNo));
            pE.El.No.BeamIn := False ;
        End
        else
        If (not pF.El.No.BeamOut) or (pF.El.No.BeamTo<>pENo) then
        Begin
            Error(71,'Improperly matched beam ends between elements '
                      + IntToStr(pENo) + ' and ' + IntToStr(pFNo)
                      + ', in stave ' + IntToStr(StNo));
        End
        else
            SwapBeamPtrs(pS, pF, pE)
    End;
End ; { SwapBeams }

Procedure BeamCheck(pS : pStave) ;
// Unrigorous checks to see if beam in/outs all match.
// (Editing may have removed one end of a beam.)
// Doesn't allow for beams extending over a stave end.
// Pre-condition: Apply before making chords as that updates beaming.
Var pE, pF : pElementList ;
    pENo   : Integer  ;
Begin
    // 1 Initialise BeamOK and TieOK to false:
    pE := pS.Elements ;
    While pE <> Nil do
    Begin
        If pE.El.ElType = Note then pE.El.No.BeamOK := False;
        pE := pE.Next
    End;
    // 2 Update BeamOK:
    pE := pS.Elements ;
    While pE <> Nil do
    Begin
      If pE.El.ElType = Note then
      With pE.El.No do
      Begin
        pENo := pE.El.ElementNo ;
        If (BeamOut and (BeamTo  <pENo))
        or (BeamIn  and (BeamFrom>pENo)) then
            SwapBeams(pS,pE) // Unusual but possible
        else
        If BeamOut then
        Begin
            pF := pE ;
            While (pF <> Nil) and (pF.El.ElementNo <> BeamTo) do pF := pF.Next ;
            If (pF<>Nil) and (pF.El.ElType=Chord) then
            Repeat pF := pF.Next until (pF = Nil) or (pF.El.ElementNo = BeamTo) ;
            If (pF=Nil) or (pF.El.ElType<>Note) or (not pF.El.No.BeamIn)
                or (pF.El.No.BeamFrom <> pENo) then
            Begin
                 Error(53,'Missing beam end, from element '+IntToStr(pENo)
                  +' to element '+IntToStr(BeamTo)
                  +', in stave '+IntToStr(pE.El.StaveNo));
                 BeamOut := False
            End
            else pF.El.No.BeamOK := True
        End
        else
        If BeamIn and not BeamOK {and (BeamFrom < pE.El.ElementNo)} then
        Begin
            Error(54,'Missing beam start, to element '+IntToStr(pE.El.ElementNo)
                  +' from element '+IntToStr(BeamFrom)
                  +', in stave '+IntToStr(pE.El.StaveNo));
            // Could check forward for beamfrom end in such cases.
            BeamIn := False
        End ;
      End ;
      pE := pE.Next
    End ;
End ; { BeamCheck }

Procedure TieCheck(pS : pStave) ;
// Unrigorous checks to see if tie in/outs all match.
// (Editing may have removed one end of a tie.)
// Doesn't allow for ties extending over stave end.
Var pE, pF : pElementList ;
    pENo   : Integer  ;
Begin
    // 1 Initialise TieOK to false:
    pE := pS.Elements ;
    While pE <> Nil do
    Begin
        If pE.El.ElType = Note then pE.El.No.TieOK  := False ;
        pE := pE.Next
    End;
    // 2 Update TieOK:
    pE := pS.Elements ;
    While pE <> Nil do
    Begin
      If pE.El.ElType = Note then
      With pE.El.No do
      Begin
        pENo := pE.El.ElementNo ;
        If TieOut then
        Begin
            pF := pE ;
            While (pF <> Nil) and (pF.El.ElementNo <> TieTo) do pF := pF.Next ;
            If pF.El.ElType=Chord then
            Repeat pF := pF.Next until (pF = Nil) or (pF.El.ElementNo = TieTo) ;
            If (pF=Nil) or (pF.El.ElType<>Note) or (not pF.El.No.TieIn)
                or (pF.El.No.TieFrom <> pE.El.ElementNo) then
            Begin
                Error(55,'Missing tie end, from element '+IntToStr(pE.El.ElementNo)
                  +' to element '+IntToStr(TieTo)
                  +', in stave '+IntToStr(pE.El.StaveNo));
                TieOut := False
            End
            else pF.El.No.TieOK := True
        End
        else
        If TieIn and not TieOK then
        Begin
            Error(56,'Missing tie start, to element '+IntToStr(pE.El.ElementNo)
                  +' from element '+IntToStr(TieFrom)
                  +', in stave '+IntToStr(pE.El.StaveNo));
            TieIn := False
        End
      End ;
      pE := pE.Next
    End ;
End ; { TieCheck }

{#############################################################################}
{#############################################################################}
{###################   Minor Processing of Single System   ###################}
{#############################################################################}
{#############################################################################}

Function SpecifiedBarLength(Top,Bottom:Integer)    : Integer ; Forward ;
Function GetBarLength(pFirst,pLast : pElementList) : Integer ; Forward ;

Procedure InsertClefs ;
// Inserts a clef element if stave does not start with a clef.
// Also inserts 4/4 time sig if there is no time sig.
// Assumes all systems concatenated into one.
// Only called by ReadNWFile after JoinSystems.
Var pE, pPrev, pClef, pTSig, pFirst, pLast : pElementList ;
    pS     : pStave ;
    TopNo,BotNo,
    BarLen : Integer ;
Begin
  pS := Music.Staves ;
  While pS<>nil do
  Begin
    pE := pS.Elements ;
    // Skip to first musical element:
    pPrev := Nil ;
    While (pE <> nil) and not (pE.El.ElType in
      [Clef,Bar,Note,Rest,TimeSig,KeySig,Accident,SlurPt,Accent,ReChord,Chord]) do
    Begin
        pPrev := pE ; pE := pE.Next
    End ;
    // First music element pE should be a clef:
    If (pE <> nil) and (pE.El.ElType<>Clef) then
    Begin // Clef is missing
        pClef := NewElement(Clef,pE) ;
        If pPrev=Nil then pS.Elements := pClef else pPrev.Next := pClef ;
        With pClef.El do
        Begin
            StaveNo := pS.StaveNo   ;
            Cl      := pS.Stv.Clef  ;
        End ;
        Inc(pS.NoElements) ;
        pE := pClef ;
    End ;

    // Time sig should be next musical element except for clef or key sig:
    While (pE<> nil) and not (pE.El.ElType in
        [Note,Rest,TimeSig,Accident,SlurPt,Accent,ReChord,Chord]) do
        Begin pPrev := pE ; pE := pE.Next End ;

    If (pE <> nil) and (pE.El.ElType <> TimeSig) then
    Begin // TimeSig is missing:
        pTSig := NewElement(TimeSig,pE) ;
        If pPrev=Nil then pS.Elements := pTSig else pPrev.Next := pTSig ;
        Inc(pS.NoElements) ;
        pE := pTSig ;
        pTSig.El.StaveNo := pS.StaveNo ;

        // Check bar length to determine best choice:
        While (pE<>Nil) and (pE.El.ElType<>Bar) do pE := pE.Next ;
        If pE<>Nil then pFirst := pE.Next else pFirst := Nil ;
        pE := pFirst ;
        While (pE<>Nil) and (pE.El.ElType<>Bar) do pE := pE.Next ;
        If pE<>Nil then pLast := pE else pLast := Nil ;
        If (pFirst<>Nil) and (pLast<>Nil) then BarLen := GetBarLength(pFirst,pLast) else BarLen := 0 ;

        // Check some possibilities:
        If (BarLen = SpecifiedBarLength(4,4)) then TopNo := 4 else
        If (BarLen = SpecifiedBarLength(3,4)) then TopNo := 3 else
        If (BarLen = SpecifiedBarLength(6,4)) then TopNo := 6 else
        If (BarLen = SpecifiedBarLength(9,8)) then TopNo := 9 else
                                                   TopNo := 4 ;
        If TopNo=9 then BotNo := 8 else BotNo := 4 ;
        If (BarLen = SpecifiedBarLength(TopNo,BotNo)) then
            Warning(3,'Inserted a time signature '+IntToStr(TopNo)+'/'+IntToStr(BotNo))
        else
            Error(3,'Inserted a time signature '+IntToStr(TopNo)+'/'+IntToStr(BotNo)+
                       ' which doesn''t match bar length') ;
        With pTSig.El.TS do
        Begin
            Top    := TopNo ;
            Bot    := BotNo ;
            IsTop  := True ;
            Common := False;
            AllaBreve:=False ;
        End ;
    End ;
    pS := pS.Next
  End
End ; { InsertClefs }


Procedure RemoveEndBarLine ;
// Removes superfluous final barline. Remove more consecutive bar lines?
// Called only in ReadNWFile.

Var pS : pStave  ;
    pLastBar, pPrev,
    pE : pElementList ;
    Delete : Boolean ;
Begin
    pS := Music.Staves ;
    While pS<>nil do
    Begin
        pE       := pS.Elements ;
        pLastBar := Nil ;
        pPrev    := Nil ;
        While pE<>nil do
        Begin
           If pE.El.ElType = Bar then pLastBar := pE    ;
           pPrev := pE ;
           pE    := pE.Next
        End ;
        pE := pLastBar ;
        // pLastBar = Last barline;
        Delete := (pLastBar<>Nil) ; // and (pLastBar.El.BL in [Single,Double]) ;
        While pE<>Nil do
        Begin
            If not(pE.El.ElType in [Bar,Text,NewStave,Accident,SlurPt,Accent,Chord,Unknown])
            then Delete := False ;
            pE    := pE.Next
        End ;
        If Delete then
        Begin
            pS.Stv.FinalBar := pLastBar.El.BL ;
            pLastBar.El.ElType := Unknown ;
        End
        else pS.Stv.FinalBar := SectionClose ;
        pS := pS.Next
    End ;
End; { RemoveEndBarLine }

Function EqualKeySigs(KS1,KS2:pElementList) : Boolean ;
// Checks KeySig elements for equality. Used in RemoveUnnecessaryElements
Begin
    With KS1.El.KS do
    EqualKeySigs := (IsFlats = KS2.El.KS.IsFlats) and (Number = KS2.El.KS.Number)
End ; { EqualKeySigs }

Function EqualTimeSigs(TS1,TS2:pElementList) : Boolean ;
// Checks TimeSig elements for equality. Used in RemoveUnnecessaryElements
Begin
    With TS1.El.TS do
    EqualTimeSigs := (Top = TS2.El.TS.Top) and (Bot = TS2.El.TS.Bot) and
                     (Common = TS2.El.TS.Common) and (AllaBreve = TS2.El.TS.AllaBreve)
End ; { EqualTimeSigs }

Procedure RemoveUnnecessaryElements ;
// Removes mid-stave elements of type newstave, clef, time sig, key sig
// where these are unnecessary after concatenating systems.
// Called only in ReadNWFile.
Var pS : pStave  ;
    pE, pKeySig, pTimeSig, pClef : pElementList ;
Begin
    pS := Music.Staves ;
    While pS<>nil do
    Begin
        pKeySig  := nil ;
        pTimeSig := nil ;
        pClef    := nil ;
        pE := pS.Elements ;
        While pE<>nil do
        Begin
           Case pE.El.ElType of
                KeySig   : If pKeySig=Nil then pKeySig := pE
                           else if EqualKeySigs(pKeySig,pE) then pE.El.ElType := Unknown
                           else pKeySig := pE ;
                TimeSig  : If pTimeSig=Nil then pTimeSig := pE
                           else if EqualTimeSigs(pTimeSig,pE) then pE.El.ElType := Unknown
                           else pTimeSig := pE ;
                Clef     : If pClef=Nil then pClef := pE
                           else if pClef.El.Cl = pE.El.Cl then pE.El.ElType := Unknown
                           else pClef := pE ;
                NewStave : pE.El.ElType := Unknown ;
                else ;
           End ;
           pE := pE.Next ;
        End ;
        pS := pS.Next
    End ;
End ; { RemoveUnnecessaryElements }


{#############################################################################}
{#############################################################################}
{####################   Joining Staves into One System   #####################}
{#############################################################################}
{#############################################################################}


Function BiggestNoteSym(L:Integer) : NoteType ;
// Input: duration L of a note in hemidemisemiquavers.
// Ouput: name of longest (undotted) note less than L.
// NB Need to allow for L larger than longest note, viz a double dotted semibreve.
// Used only in InitialiseRest
Var NoteLen : Integer  ;
    NoteSym : NoteType ;
Begin
    NoteSym := Hemidemisemiquaver ;
    NoteLen := 1 ;
    While (2*NoteLen<=L) and (Ord(NoteSym)>0) do
    Begin
        NoteLen := 2*NoteLen ;
        NoteSym := Pred(NoteSym) ;
    End ;
    BiggestNoteSym := NoteSym ;
End ; { BiggestNoteSym }

Function BiggestNoteLength(L:Integer) : Byte ;
// Input: duration L of a note
// Ouput: duration in same units as input of longest undotted note less than L.
// NB Need to allow for L larger than longest note, viz a double dotted semibreve.
// Used only in InitialiseRest
Var NoteLen : Integer  ;
    NoteSym : NoteType ;
Begin
    NoteSym := Hemidemisemiquaver ;
    NoteLen := 1 ;
    While (2*NoteLen<=L) and (Ord(NoteSym)>0) do
    Begin
        NoteLen := 2*NoteLen ;
        NoteSym := Pred(NoteSym) ;
    End ;
    BiggestNoteLength := NoteLen ;
End ; { BiggestNoteLength }

Function BiggestNoteDots(L:Integer) : Integer ;
// Input: duration L of a note
// Ouput: number of dots <=2 on longest dotted note less than duration L.
// NB Need to allow for L larger than longest note, viz a double dotted semibreve.
// Used only in InitialiseRest
Var NoteLen : Integer  ;
    NoteSym : NoteType ;
Begin
    NoteSym := Hemidemisemiquaver ;
    NoteLen := 1 ;
    While (2*NoteLen<=L) and (Ord(NoteSym)>0) do
    Begin
        NoteLen := 2*NoteLen ;
        NoteSym := Pred(NoteSym) ;
    End ;
    If (2*NoteLen<=L) then
       BiggestNoteDots := 0
    else
    If (NoteLen mod 4 = 0) and (7*NoteLen <=4*L) then
       BiggestNoteDots := 2
    else
    If (NoteLen mod 2 = 0) and (3*NoteLen <=2*L) then
       BiggestNoteDots := 1
    else
       BiggestNoteDots := 0 ;
End ; { BiggestNoteDots }

Function BiggestNoteLeft(L:Integer) : Integer ;
// Input: duration L of a note.
// Ouput: duration after removing longest dotted note less than L.
// NB Need to allow for L larger than longest note, viz a double dotted semibreve.
// Used only in FillEmptyStave

Var NoteLen : Integer  ;
    NoteSym : NoteType ;
Begin
    NoteSym := Hemidemisemiquaver ;
    NoteLen := 1 ;
    While (2*NoteLen<=L) and (Ord(NoteSym)>0) do
    Begin
        NoteLen := 2*NoteLen ;
        NoteSym := Pred(NoteSym) ;
    End ;
    // NoteLen = longest undotted note no longer than L.
    If (2*NoteLen<=L) then
       BiggestNoteLeft := L - NoteLen
    else
    If (NoteLen mod 4 = 0) and (7*NoteLen <=4*L) then
       BiggestNoteLeft := L - 7*NoteLen div 4   // 2 dots
    else
    If (NoteLen mod 2 = 0) and (3*NoteLen <=2*L) then
       BiggestNoteLeft := L - 3*NoteLen div 2   // 1 dot
    else
       BiggestNoteLeft := L - NoteLen ;
End ; { BiggestNoteLeft }



Function NoteDuration(pE:pElementList):Integer ;
{  Returns length of a note in hemidemisemiquavers and
   sets pE.El.No.Duration to this value if length is non-zero.
   Used only in PartBarLength, GetBarLength, ChordTemplate.
}
Var SoFar  : Integer ;
    NSym   : NoteType;
Begin
    SoFar := 0 ;
    If (pE<>nil) and (pE.El.ElType in [Chord,Rest,Note]) then
    With pE.El.No do
    Begin
           NSym  := Hemidemisemiquaver   ;
           SoFar := 1 ;
           While NSym<>NoteSym do
           Begin
               NSym  := Pred(NSym) ;
               SoFar := 2*SoFar
           End ;
           If DoubleDot then
           Begin
               If (SoFar mod 4)<>0 then Error(48,'Too short a double dotted note') ;
               SoFar := (7*SoFar) div 4 ;
           End
           else
           If Dotted then
           Begin
               If (SoFar mod 2)<>0 then Error(49,'Too short a dotted note') ;
               SoFar := (3*SoFar) div 2 ;
           End ;
    End ;
    NoteDuration := SoFar ;
End ; { NoteDuration }


// These elements need copying vertically down from top into the added staves, but none else:
Const NonNoteElements = [Bar,TimeSig,{Tempo,}TempoVar,Flow,Pedal,Vol,VolVar,PStyle] ;


Function PartBarLength(pE:pElementList) : Byte ;
{ Length returned is in hemidemisemiquavers.
  Pre-condition: Chord elements have already been inserted (MakeChords).
  Elements from pE to next NonNoteElements are counted, including pE itself.
  (Grace notes/grace rests do not contribute.)
  NB 1: A chord element's note has length of shortest note in chord, so presence of
  longer notes in chord does not disrupt the count.
  NB 2: A ReChord rest length is counted but the following Chord is not.
}
Var LengthSoFar,
    NewDuration  : Integer ;
Begin
    LengthSoFar := 0 ;
    While (pE<>nil) do
    Begin
        NewDuration := 0 ;
        If pE.El.ElType in NonNoteElements then pE := nil
        else
        Case pE.El.ElType of
          Rest,
          Note  : If not pE.El.No.InChord and not pE.El.No.Grace then
                  NewDuration := NoteDuration(pE) ;
          Chord : NewDuration := NoteDuration(pE) ;
          ReChord: Begin
                     NewDuration := NoteDuration(pE) ;
                     pE := pE.Next // i.e. skip following Chord element
                   End
          else    // Nothing
        End ;
        LengthSoFar := LengthSoFar + NewDuration ;
        If pE<>nil then pE := pE.Next ;
    End ;
    PartBarLength := LengthSoFar
End ; { PartBarLength }

Procedure InitialiseRest(Var Note : NoteRec; RestLen: Integer) ;
// Sets Note to longest rest of single symbol with length <= RestLen
Var Dots : Integer ;
Begin
    With Note do
    Begin
        NoteSym  := BiggestNoteSym (RestLen) ;
        Dots     := BiggestNoteDots(RestLen) ;
        If Dots>0 then Dotted   := True ;
        If Dots>1 then DoubleDot:= True ;
    End ;
End ; { InitialiseRest }

Procedure FillEmptyStave(pTop,pNew:pStave) ;
{ Fills empty stave pNew with rest bars of same duration as bars in pTop.
  It copies bits of bars, rather than whole bars; the bits are delimited by
  elements in NonNoteElements, all of which are copied. }

Var pElTop, pElNew, pNext : pElementList ;
    RestLen               : Integer      ;
Begin
    pElTop := pTop.Elements ;
    pElNew := pNew.Elements ;
    While pElTop<>Nil do
    Begin
        If not (pElTop.El.ElType in NonNoteElements) then
        Begin
            // Insert rest(s) equal to length in pTop up to next NonNoteElement:
            RestLen := PartBarLength(pElTop) ;
            While RestLen>0 do
            Begin  // Add rest(s) to value of RestLen:
                pNext := NewElement(Rest,Nil) ;
                Inc(pNew.NoElements) ;
                With pNext.El do
                Begin
                    StaveNo   := pNew.StaveNo   ;
                    ElementNo := pNew.NoElements ;
                End ;
                InitialiseRest(pNext.El.No,RestLen) ;
                If pElNew=Nil then pNew.Elements := pNext else
                pElNew.Next := pNext ;
                pElNew      := pNext ;
                RestLen     := BiggestNoteLeft(RestLen)
            End ;
            // Skip to end of part bar:
            Repeat
                pElTop := pElTop.Next
            until (pElTop=nil) or (pElTop.El.ElType in NonNoteElements)
        End ;

        // Copy barlines, time sigs, endings etc from next part bar of pTop:
        While (pElTop<>nil) and (pElTop.El.ElType in NonNoteElements) do
        Begin
            pNext := NewElement(pElTop.El.ElType,Nil) ;
            pNext.El := pElTop.El ;
            If pElNew=Nil then pNew.Elements := pNext else
            pElNew.Next := pNext ;
            pElNew      := pNext ;
            Inc(pNew.NoElements) ;
            pElTop := pElTop.Next ;
        End
    End
End ; { FillEmptyStave }

Function SpecifiedBarLength(Top,Bottom:Integer) : Integer ;
// For time signature Top/Bottom, this returns number of hemidemisemiquavers in bar.
// 1 = value of Bottom in Time Sig for Semibreve.
Var BarLen,
    Lower  : Integer ;
    NT     : NoteType;
Begin
    Lower  := 1 ;
    BarLen := 0 ;
    For NT := Semibreve to Hemidemisemiquaver do
    Begin
        If Bottom = Lower then BarLen := 1 else BarLen := 2*BarLen ;
        Lower := 2*Lower ;
    End ;
    SpecifiedBarLength := Top*BarLen ;
End ; { SpecifiedBarLength }

Function GetBarLength(pFirst,pLast : pElementList) : Integer ;
// pFirst..pLast contains all the musical elements of the bar
// (or other element sequence of interest).
Var LengthSoFar  : Integer ;
    pE           : pElementList ;
    AfterReChord : Boolean ;
Begin
    LengthSoFar := 0 ;
    pE := pFirst ;
    AfterReChord := False ;
    While pE <> Nil do
    Begin
        If (((pE.El.ElType = Chord) and not AfterReChord) or
           ((pE.El.ElType in [Note,Rest]) and not pE.El.No.InChord))
           and not pE.El.No.Grace then
             LengthSoFar := LengthSoFar + NoteDuration(pE) ;
        If pE.El.ElType = Chord then AfterReChord := False ;
        If pE = pLast then pE := Nil else pE := pE.Next
    End ;
    GetBarLength := LengthSoFar
End ; { GetBarLength }


Procedure CheckBarLineAtStaveEnd(pS : pStave; pEnd : pElementList) ;
{ Deletes last barline at or before pE if ordinary barline and
  not followed by any note or rest and preceding bar is not complete.
  Here pEnd is a concatenation point (or last element) of staves.
  Assumes chord elements already created.
}
Var pLastBarLine, pPenultimateBL, pE : pElementList ;
    Found                            : Boolean ;
    Top, Bottom, TSBarLen, EndBarLen : Integer ;
Begin
    // 1. Find bar length (i.e. time signature) and last two barlines before pEnd:
    pE := pS.Elements ;
    pLastBarLine   := Nil ;
    pPenultimateBL := pE  ;
    Top            := 4 ;
    Bottom         := 4 ;
    While (pE<>Nil) do
    Begin
        If pE.El.ElType = Bar then
        Begin pPenultimateBL := pLastBarLine ; pLastBarLine := pE End
        else
        If pE.El.ElType = TimeSig then
        Begin
            If pE.El.TS.Common then Begin Top := 4 ; Bottom := 4 End
            else
            If pE.El.TS.AllaBreve then Begin Top := 2 ; Bottom := 2 End
            else
            Begin Top := pE.El.TS.Top ; Bottom := pE.El.TS.Bot End
        End ;
        If pE = pEnd then pE := nil else pE := pE.Next
    End ;

    // 2. Check if any notes beyond pLastBarLine:
    Found := False ;
    pE := pLastBarLine ;
    While (pE<>nil) and not Found do
    Begin
        If (pE.El.ElType = Note) or (pE.El.ElType = Rest) then Found := True ;
        If pE = pEnd then pE := nil else pE := pE.Next
    End ;

    // 3. Remove the (single) barline if necessary.
    If not Found and (pLastBarLine<>Nil) and (pLastBarLine.El.BL = Single) then
    Begin
        TSBarLen  := SpecifiedBarLength(Top,Bottom) ;
        EndBarLen := GetBarLength(pPenultimateBL,pLastBarLine) ;
        If EndBarLen<TSBarLen then
        Begin // Delete the barline pLastBarLine:
            pE := pPenultimateBL ;
            While (pE<>Nil) and (pE<>pLastBarLine) and (pE.Next<>pLastBarLine) do pE := pE.Next ;
            If (pE<>Nil) then pLastBarLine.El.ElType := Unknown ;
        End ;
    End
End; { CheckBarLineAtStaveEnd }

Function MaxNoStavesInSystem(Var Missing : Boolean; Var pMaxSys: pStave) : Integer ;
// Returns max no of staves in any system,
// with Missing true iff some system doesn't have that number.
// and pMaxSys pointing to start of max system.
Var pS, pFirst  : pStave ;
    NoStaves,
    MaxNoStaves : Integer ;
Begin
    pS := Music.Staves ;
    pMaxSys := Nil ;
    pFirst  := Nil ;
    MaxNoStaves := 0 ;
    Missing := False;
    While pS <> Nil do
    Begin
        If not pS.FirstInSystem then
        Begin
            Error(6,'Failed to detect first stave in system') ;
            // MessageDlg('Failed to detect first stave in system', mtInformation, [mbOk], 0);
        End
        else pFirst := pS ;
        // Count NoStaves in system starting at pFirst:
        NoStaves := 1 ;
        pS := pS.Next ;
        While (pS<>nil) and not pS.FirstInSystem do
        Begin
            pS := pS.Next ;
            Inc(NoStaves) ;
        End ;
        // Update MaxNoStaves and Missing:
        If MaxNoStaves=0 then Begin MaxNoStaves:= NoStaves; pMaxSys := pFirst End else
        If NoStaves>MaxNoStaves then
            Begin Missing := True ; MaxNoStaves := NoStaves; pMaxSys := pFirst End
        else
        If NoStaves<MaxNoStaves then Missing := True ;
    End ;
    MaxNoStavesInSystem := MaxNoStaves ;
End ; { MaxNoStavesInSystem }

Procedure NewStaveRec(Var pNew : pStave) ;
Begin
    New(pNew) ;
    With pNew^ do
    Begin
        Next         := Nil  ;
        StaveNo      := 0    ;
        NoElements   := 0    ;
        VertPosn     := 0    ;
        FirstInSystem:= False;
        LastInSystem := False;
        Elements     := Nil  ;
        With Stv do
        Begin
            StType   := Standard ;
            With TSig do Begin
               Top := 0; Bot := 0; IsTop := False; Common := False; AllaBreve := False
            End ;
            With KSig do Begin
               IsSharps:= False; IsNatural:= False; IsFlats := False; Number := 0
            End ;
            Clef     := Treble ;
            // First := Boolean  ;
            NoLyrics := 0 ;
            Numbering:= False ;
            TitleInfo:= False ;
            ExtendLast:= False ;
            IncreaseSpacing := False ;
            From     := 0  ;
            StaffSize:= 20 ;
            Start    := 0 ;
            Labels   := NoLabel ;
            Numbers  := None ;
        End;
    End;
End ; { NewStaveRec }

Procedure AddEmptyStaves(MaxNoStaves: Integer) ;
// Supplements short systems with extra empty staves
// with barlines to give them all MaxNoStaves staves.
Var pS, pStart, pNew, pPrev : pStave ;
    NoStaves, I    : Integer ;
Begin
    pS := Music.Staves ;
    While pS <> Nil do
    Begin
        // Find start of next system:
        While (pS<>nil) and not pS.FirstInSystem do pS := pS.Next ;
        If pS<>nil then
        Begin
            // Find last stave in system:
            pStart:= pS ;
            pPrev := pS ;
            pS    := pS.Next;
            NoStaves := 1 ;
            While (pS<>nil) and not pS.FirstInSystem do
            Begin
                pPrev := pS ;
                pS := pS.Next ;
                Inc(NoStaves) ;
            End ;
            // pS is nil, or points to start of next system.
            // Insert extra staves, as needed, between pPrev & pS:
            For I := NoStaves+1 to MaxNoStaves do
            Begin // Insert an empty stave:
                NewStaveRec(pNew) ;
                If I = MaxNoStaves then pNew.LastInSystem := True ;
                Inc(Music.NoStaves) ;
                pNew.Next  := pS ;
                pPrev.Next := pNew ;
                pPrev := pNew ;
                // Add empty bars as in pStart to pNew:
                FillEmptyStave(pStart,pNew) ;
            End
        End ;
    End
End ; { AddEmptyStaves }

Function LengthSystemOne : Integer ;
Var pS       : pStave ;
    NoStaves : Integer ;
Begin
    pS := Music.Staves ;
    If pS = Nil then LengthSystemOne := 0
    else
    Begin
        NoStaves := 1 ;
        pS := pS.Next ;
        While (pS<>nil) and not pS.FirstInSystem do
           Begin Inc(NoStaves); pS := pS.Next End ;
        LengthSystemOne := NoStaves ;
    End
End ; { LengthSystemOne }

Procedure TransferStaveProperties(pMaxSys : pStave; L, NoStaves : Integer) ;
// Copies relevant stave properties from that at pMaxSys to the first system.
Var pS, pMax : pStave ;
    Count : Integer ;
Begin
    pS    := Music.Staves ;
    pMax  := pMaxSys      ;
    Count := 0 ;
    While Count<NoStaves do // Count staves have been processed so far
    Begin
        // Need to copy the stave type, key sig and clef:
        pS.Stv.StType := pMax.Stv.StType;
        If Count >= L then
        Begin
            pS.Stv.Clef   := pMax.Stv.Clef  ;
            pS.Stv.KSig   := pMax.Stv.KSig  ;
            // Must also create the Clef and KSig elements in these staves.
        End ;
        pS   := pS.Next ;
        pMax := pMax.Next ;
        Inc(Count) ;
    End;
End ; { TransferStaveProperties }

Procedure RenumberStaves(NoStaves : Integer) ;
Var pS    : pStave ;
    Count : Integer;
Begin
    pS := Music.Staves ;
    Count := 0 ;
    While Count<NoStaves do // Count staves have been processed so far
    Begin
        Inc(Count) ;
        pS.StaveNo := Count ;
        pS := pS.Next
    End;
End ; { RenumberStaves }

Procedure JoinSystems ;
{ Systems (with possibly different numbers of staves) are concatenated
  but extra elements such as clefs and key sigs are still to be removed.
  Barlines may need to be added or removed where staves are joined together.
  To Do: Need to fix upper and lower bounds of concatenated staves
  to max extent of their component staves.
  Pre-condition: chord elements have been created already (so bar length can be found)
}
Var pS, pMaxSys,
    pPrev, pNext : pStave ;
    pE, pEnd     : pElementList ;
    MaxNoStaves,
    I, L         : Integer ;
    Missing      : Boolean ; // True iff staves need adding.
Begin
    // 1.Find max number of staves in any system:
    L := LengthSystemOne ;
    MaxNoStaves := MaxNoStavesInSystem(Missing,pMaxSys) ;
    // 2. Add extra staves to systems which are short:
    If Missing then AddEmptyStaves(MaxNoStaves) ;
    // 3. Copy properties of biggest system to first system:
    If L < MaxNoStaves then TransferStaveProperties(pMaxSys,MaxNoStaves,L) ;
    RenumberStaves(MaxNoStaves) ;

    // 3. Find the start of the second system, if any, and terminate after first system:
    pNext := Music.Staves ;
    pPrev := Music.Staves ;
    I     := 1 ;
    If pNext<>Nil then pNext := pNext.Next ;
    While (pNext<>nil) and not pNext.FirstInSystem do
       Begin pPrev := pNext ; pNext := pNext.Next ;  Inc(I) End ;
    If (pNext<>Nil) and (I <> MaxNoStaves) then
        Error(50,'Unexpected first system length of '+IntToStr(I)+' staves.') ;
    If pPrev<> Nil then pPrev.Next := Nil ; // This ends the list of staves after MaxNoStaves staves.
    Music.NoStaves := MaxNoStaves ;

    // 4. Extend the staves of the first system:
    While pNext<>Nil do // pNext is start of next system to join in.
    Begin
        If not pNext.FirstInSystem then Error(50,'Unexpected loss of start of system') ;
        // pNext points to start of the next system for concatenating to first system:
        // (The first set is indexed by pS, the next set by pNext)
        pS := Music.Staves ;
        For I := 1 to MaxNoStaves do
        If pNext=Nil then
            Error(51,'Unexpected missing stave')
        else
        Begin // concatenate stave to end of Ith stave in first system:
            // Find last element pEnd in pS:
            pE := pS.Elements ;
            pEnd := Nil ;
            While pE<>Nil do Begin pEnd := pE; pE := pE.next End ;
            If pEnd<>Nil then
               pEnd.Next := pNext.Elements
            else
               pS.Elements := pNext.Elements;
            pS.NoElements := pS.NoElements + pNext.NoElements ;
            // Insert or delete barline at pEnd, as necessary:
            CheckBarLineAtStaveEnd(pS,pEnd) ;
            // Move to next stave in first system and next stave in current system:
            pS    := pS.Next ;
            pNext := pNext.Next ;
        End
    End ;
End ; { JoinSystems }



{#############################################################################}
{#############################################################################}
{#####################   Processing of Chord Elements   ######################}
{#############################################################################}
{#############################################################################}


Procedure SetReChord(pReChord : pElementList) ;
Var pE, pChord  : pElementList ;
    Count, Size : Integer      ;
Begin
    pReChord.El.No.NoInChord   := 1 ;
    pReChord.El.ElType         := ReChord ;
    pReChord.El.No.WithReChord := True ;
    // Update chord elements:
    pChord := pReChord.Next ;
    Size := pChord.El.No.NoInChord - 1 ;
    pChord.El.No.NoInChord := Size ; // Due to moving rest from chord set.
    pE := pChord ;
    Count := 0 ;
    While (Count < Size) and (pE<>Nil) do
    Begin
        If pE.El.ElType=Note then
           Begin pE.El.No.WithReChord := True ; Inc(Count) End ;
        pE := pE.Next ;
    End ;
End ; { SetReChord }

Procedure UnsetChord(pChord : pElementList) ;
Var pE  : pElementList ;
    Size, Count : Integer ;
Begin
    Size := pChord.El.No.NoInChord ; // Should raise error if this is not = 1
    pE := pChord ;
    Count := 0 ;
    While (Count < Size) and (pE<>Nil) do
    Begin
        If pE.El.ElType=Note then
        Begin
           pE.El.No.WithReChord := False ;
           pE.El.No.InChord     := False ;
           Inc(Count)
        End ;
        pE := pE.Next ;
    End ;
    pChord.El.ElType := Unknown ;
End ; { UnsetChord }

Procedure CreateReChordElements ;
// Creates all ReChord elements and inserts them before corresponding Chord element.
// Assumes the chord elements contain at least one note with WithReChord true
// if ReChord element needs creating.
// Chords were created in MakeChords. Chord elements with <2 elements are deleted.
// ReChord elements don't contribute to the total NoElements.
Var pE, pF, pPrev, pReChord, pChord, pRest : pElementList ;
    pS : pStave ;
Begin
  pS := Music.Staves ;
  While pS<>nil do
  Begin
    pE := pS.Elements ;
    pPrev := Nil ;
    While pE <> Nil do
    Begin
        If (pE.El.ElType = Chord) and (pE.El.No.WithReChord) then
        Begin  // Find Rest element of chord, turn it into a ReChord element,
               // move to before Chord element and update chord element count:
            pChord := pE ;
            pF     := pChord ;
            While (pF.Next<>Nil) and (pF.Next.El.ElType<>Rest) do pF := pF.Next ;
            pRest    := pF.Next ;
            pF.Next  := pRest.Next ;
            pReChord := pRest ;
            pReChord.Next := pChord ;
            If pPrev = Nil then pS.Elements := pReChord else pPrev.Next := pReChord ;
            SetReChord(pReChord) ; // Updates content of ReChord
            pPrev := pReChord ;
            Dec(pS.NoElements)   // ReChord elements don't count.
        End;
        If (pE.El.ElType = Chord) and (pE.El.No.NoInChord<2) then
        Begin // Remove Chord element and update chord elements
            pChord := pE ;
            UnsetChord(pChord) ;
            pChord.El.ElType := Unknown ;
        End;
        pPrev := pE ;
        pE := pE.Next
    End ;
    pS := pS.Next
  End
End ; { CreateReChordElements }

Function PartSize(pFirst, pLast : pElementList; Stem: StemType) : Integer ;
// Stem = Unset includes all notes with stem not in [Up,Dn]
Var Size    : Integer ;
    pE      : pElementList ;
    StemSet : Set of StemType ;
Begin
    If Stem = Unset then StemSet := [NoStem,Auto,Unset] else StemSet := [Stem] ;
    Size := 0 ;
    pE := pFirst ;
    While pE <> Nil do
    Begin
        If (pE.El.ElType in [Note,Rest]) and (pE.El.No.Stem in StemSet) then Inc(Size) ;
        If pE=pLast then pE := Nil else pE := pE.Next
    End ;
    PartSize := Size
End ; { PartSize }


Procedure RightShiftChordEl(pChord,pLast : pElementList) ;
// Copy RtShift value to pChord if set in member note.
Var pF          : pElementList ;
    FoundRtShft : Boolean ;
Begin
    FoundRtShft := False ;
    pF := pChord.Next ;
    While pF<>Nil do
    Begin
        If (pF.El.ElType = Note) and pF.El.No.RtShift then FoundRtShft := True ;
        If pF = pLast then pF := Nil else pF := pF.Next
    End ;
    pChord.El.No.RtShift := FoundRtShft
End ; { RightShiftChordEl }

Procedure AssignRightShifts(pFirst,pLast : pElementList) ;
// Precndition : Call FixStems first since stem up parts and
// stem dn parts can overlap at the same pitch, but no further.
// Sets Note field RtShift to true for notes in chord if head is shited rightwards.
// Exactly one of two notes on adjacent pitches is affected, and none other.
Var PitchRA: Array[MinPitch..MaxPitch] of Boolean ;      // True iff Chord has element of given pitch
    StemRA : Array[MinPitch..MaxPitch] of StemType ;     // Specifies stem for given pitch
    PtrRA  : Array[MinPitch..MaxPitch] of pElementList ; // Specifies where element of give pitch is
    MidPitch    : Integer ;      // MidPitch = highest stem dn pitch, or lowest stem up.
    Right, Meet : Boolean ;      // Meet = true iff upper and lower parts share a pitch
    pMid        : pElementList ; // pMid = other note of same pitch if Meet.

        Procedure InitialiseRAs ;
        Var I, P : Integer ;
            pF   : pElementList ;
        Begin
            For I := MinPitch to MaxPitch do PitchRA[I] := False ;
            Meet  := False ;
            pF := pFirst ;
            While pF<>Nil do
            Begin
                If pF.El.ElType in [note,rest] then
                Begin
                    P := pF.El.No.Pitch ;
                    If PitchRA[P] then Begin Meet := True; pMid := pF; MidPitch := P End else
                    Begin
                      PitchRA[P] := True ;
                      PtrRA  [P] := pF ;
                      StemRA [P] := pF.El.No.Stem
                    End
                End ;
                If (pF<>pLast) then pF := pF.Next else pF := Nil
            End ;
        End ; { InitialiseRAs }

        Function RightShiftNeeded : Boolean ;
        Var Offset : Boolean ;
            I : Integer ;
        Begin
            Offset := False ;
            For I := MinPitch to MaxPitch-1 do
              If PitchRA[I] and PitchRA[I+1] then Offset := True ;
            If Meet and ((PartSize(pFirst, pLast, Up)>1)or(PartSize(pFirst, pLast, Dn)>1))
              then Offset := True ;
            RightShiftNeeded := Offset
        End ; { RightShiftNeeded }

        Procedure GetStemUpRtShifts ;
        Var I, Prev : Integer ;
            pF   : pElementList ;
        Begin
            If Meet and (pMid.El.No.Stem=Dn) then
            Begin // Swap pMid with element in arrays:
                pF := pMid;
                pMid := PtrRA[MidPitch];
                PtrRA[MidPitch] := pF;
                StemRA[MidPitch]:= pF.El.No.Stem  // = Up, I hope!
            End ;
            Prev := MinPitch ;
            Right:= False ; // Right shift for next head.
            For I := MinPitch to MaxPitch do
            If PitchRA[I] and (StemRA [I] = Up) then
            Begin
                If Prev = I-1 then
                     Begin PtrRA[I].El.No.RtShift := Right; Right := not Right End
                else Begin PtrRA[I].El.No.RtShift := False ; Right := True End ;
                Prev := I ;
            End;
        End ; { GetStemUpRtShifts }

        Procedure GetStemDnRtShifts ;
        Var I, Prev : Integer ;
            pF   : pElementList ;
        Begin
            If Meet and (pMid.El.No.Stem=Up) then
            Begin // Swap pMid with element in arrays:
                pF := pMid;
                pMid := PtrRA[MidPitch];
                PtrRA[MidPitch] := pF;
                StemRA[MidPitch]:= pF.El.No.Stem  // = Dn, I hope!
            End ;
            Prev := MaxPitch ;
            Right:= True ; // Right shift for next head.
            For I := MaxPitch downto MinPitch do
            If PitchRA[I] and (StemRA [I] = Dn) then
            Begin
                If Prev = I+1 then
                     Begin PtrRA[I].El.No.RtShift := Right; Right := not Right End
                else Begin PtrRA[I].El.No.RtShift := True ; Right := False End ;
                Prev := I ;
            End;
        End ; { GetStemDnRtShifts }

Begin
    InitialiseRAs ;
    If RightShiftNeeded then
    Begin
        GetStemUpRtShifts ;
        GetStemDnRtShifts ;
    End;
End ; { AssignRightShifts }


Procedure FixStems(pFirst, pLast, pRest : pElementList{; RestIsShorter : Boolean}) ;
// Pre-condition: chord Contains one rest, pRest.
// Modifies stems in pFirst..pLast so rest and notes have opp stem directions, and all are Up or Dn.

            Function StemDirection(HasStem,IsGrace,IsRest: Boolean) : StemType ;
            // Returns StemType of notes/rests of parameter type, Unset if none found.
            Var pE : pElementList; Stem : StemType ;
            Begin
                pE := pFirst ;    Stem := Unset ;
                While pE<>Nil do
                Begin
                    If pE.El.ElType in [Rest,Note] then
                    Begin
                        If ((pE.El.ElType = Rest)=IsRest)
                         and (((pE.El.No.NoteSym in [breve,semibreve]) or
                              (pE.El.ElType = Rest)) <> HasStem )  // cases with no visible stem
                         and  (pE.El.No.Grace=IsGrace)
                         and   (pE.El.No.Stem in [Up,Dn])
                        then Begin Stem := pE.El.No.Stem ; pE := pLast End ;
                    End ;
                    If pE = pLast then pE := Nil else pE := pE.Next
                End;
                StemDirection := Stem
            End ; { StemDirection }

Var pE : pElementList ;
    NoAbove, NoBelow, Pitch : Integer ;
    Stem, RestStem : StemType ;
Begin
    // Find a note with an up or dn stem. (params: HasStem,IsGrace,IsRest.)
                         Stem := StemDirection(True, False,False) ;
    If Stem = Unset then Stem := StemDirection(False,False,False) ;
    If Stem = Unset then Stem := StemDirection(False,False,True ) ;
    If Stem = Unset then Stem := StemDirection(True, True, False) ;
    If Stem = Unset then Stem := StemDirection(False,True, False) ;
    If Stem = Unset then Stem := StemDirection(False,True, True ) ;
    // If stem direction of pRest is same as rest of chord then
    // decide on stem direction by looking at pitches, (or counts if that doesn't work).
    // NB Rest Pitch in NW files is Vert Posn of symbol, so indicates intended part
    If (pRest.El.No.Stem = Stem) or (Stem = Unset) then
    Begin
        NoAbove := 0 ;
        NoBelow := 0 ;
        Pitch:= pRest.El.No.Pitch ;
        pE   := pFirst ;
        While pE<>Nil do
        Begin // Find next note/rest in chord:
            If pE.El.ElType in [Rest,Note] then
            If pE.El.No.Pitch > Pitch then Inc(NoAbove) else
            If pE.El.No.Pitch < Pitch then Inc(NoBelow) ;
            If pE = pLast then pE := Nil else pE := pE.Next
        End ;
        If NoAbove > NoBelow then
             Begin Stem := Up; RestStem := Dn End
        else Begin Stem := Dn; RestStem := Up End
    End
    else RestStem := pRest.El.No.Stem ;

    // Set stems in chord:
    pE := pFirst ;
    While pE<>Nil do
    Begin
        If pE.El.ElType = Rest then pE.El.No.Stem := RestStem
        else
        If pE.El.ElType = Note then pE.El.No.Stem := Stem ;
        If pE = pLast then pE := Nil else pE := pE.Next ;
    End;
End ; { FixStems }


Procedure ReorderChordEls(pChord, pLast: pElementList) ;
// Assumes pChord is a Chord element with last rest/note element at pLast.
// Moves non-note elements so there are no such elements between chord ends.
// Such elements can occur and cause NW Composer & Viewer to crash.
// Ordering pitches (e.g. from highest to lowest) causes problems with beams
// so notes kept in elementno order.
Var pF, pPrev : pElementList ;
Begin
   pF := pChord ;
   Repeat
       pPrev := pF ;
       pF    := pF.Next ;
       If not (pF.El.ElType in [Note,Rest]) and (pF<>pLast) then
       Begin // Swap pF to element after pLast:
          pPrev.Next:= pF.Next ;
          pF.Next   := pLast.Next ;
          pLast.Next:= pF ;
          pF        := pPrev
       End ;
   until (pF=Nil) or (pF=pLast)
End ; { ReorderChordEls }


Procedure FixOverlappingChordElements(pChord,pLast : pElementList)  ;
{ The procedure modifies stem directions in the chord pFirst..pLast
  so that all are Up or Dn and all stem dn notes are below all stem up notes.
  NWC cannot cope without this property.
  Parts can share a note only as long as it is the same length in both parts.
  Moved elements have their duration modified if necessary (with beaming consequences)
  Notes of equal pitch may have one deleted, with beam, tie and timing consequences.
  NOT all stem directions are set to Up or Dn in this routine.
}
Var Unsets              : Integer ;
    pFirst, pHigh, pLow : pElementList ;

          Function HighestOrLowest(Stem: StemType) : pElementList ;
          // Returns highest pitched stem dn or lowest pitched stem up note.
          // Ignores notes with no fixed stem direction.
          Var pE, pBest : pElementList ;
          Begin
              pBest := Nil ;
              pE := pFirst ;
              While pE <> Nil do
              Begin
                  If (pE.El.ElType = Note) and (pE.El.No.Stem = Stem) then
                  Begin
                      If pBest = Nil then pBest := pE else
                      If (Stem=Dn) and (pE.El.No.Pitch > pBest.El.No.Pitch) then pBest := pE else
                      If (Stem=Up) and (pE.El.No.Pitch < pBest.El.No.Pitch) then pBest := pE
                  End ;
                  If pE=pLast then pE := Nil else pE := pE.Next
              End ;
              HighestOrLowest := pBest
          End ; { HighestOrLowest }

          Procedure FixCrotchets(Stem:StemType) ;
          // Changes crotchets to shorter notes in chords where the difference cannot be seen.
          Var pE          : pElementList ;
              HasCrotchet : Boolean  ;
              MaxSym      : NoteType ;
          Begin
              HasCrotchet := False; MaxSym := Crotchet ;
              pE := pFirst ;
              While pE<>Nil do
              Begin
                  If (pE.El.ElType=Note) and (pE.El.No.Stem = Stem) then
                  Begin
                      If pE.El.No.NoteSym = Crotchet then HasCrotchet := True else
                      If pE.El.No.NoteSym > MaxSym then MaxSym := pE.El.No.NoteSym ;
                  End ;
                  If pE = pLast then pE := Nil else pE := pE.Next
              End ;
              pE := pFirst ;
              If HasCrotchet and (MaxSym<>Crotchet) then
              While pE<>Nil do
              Begin
                  If (pE.El.ElType=Note) and (pE.El.No.Stem = Stem) and
                      (pE.El.No.NoteSym = Crotchet) then pE.El.No.NoteSym := MaxSym ;
                  If pE = pLast then pE := Nil else pE := pE.Next
              End ;
          End ; { FixCrotchets }

          Procedure FixStemless ;
          // Checks for stemless notes, and adds dn stem
          // if one note lower than one with dn stem & similar for up stems.
          Var pE,pF : pElementList ;
          Begin
              If Unsets=0 then pE := Nil else pE := pFirst ;
              While pE<>Nil do
              Begin
                  If (pE.El.ElType=Note) and (pE.El.No.Stem = NoStem) then
                  Begin
                      pF := pFirst ;
                      While pF<>Nil do
                      Begin
                          If (pF.El.ElType=Note) then
                            If ((pF.El.No.Stem = Dn) and (pF.El.No.Pitch = pE.El.No.Pitch+1)) then
                                 pE.El.No.Stem := Dn
                            else
                            If ((pF.El.No.Stem = Up) and (pF.El.No.Pitch = pE.El.No.Pitch-1)) then
                                 pE.El.No.Stem := Up ;
                          If pF = pLast then pF := Nil else pF := pF.Next
                      End
                  End ;
                  If pE = pLast then pE := Nil else pE := pE.Next
              End ;
          End ; { FixStemless }

          Procedure ReduceNoUnsets(pHigh,pLow : pElementList) ;
          // pHigh, pLow are highest stem dn & lowest stem up notes.
          // Sets all stems except unset ones between pHigh and pLow.
          Var pE : pElementList ;
              Unsets, pLowPitch, pHighPitch, pEPitch : Integer ;
          Begin
              Unsets := PartSize(pFirst,pLast,Unset) ;
              If pLow <>Nil then pLowPitch  := pLow .El.No.Pitch else pLowPitch  := MaxPitch ;
              If pHigh<>Nil then pHighPitch := pHigh.El.No.Pitch else pHighPitch := MinPitch ;
              If (Unsets=0) or ((pLow=Nil)and(pHigh=Nil)) then pE := Nil else pE := pFirst ;
              While pE<>Nil do
              Begin
                  If (pE.El.ElType=Note) and not (pE.El.No.Stem in [Up,Dn]) then
                  Begin
                    pEPitch := pE.El.No.Pitch ;
                    If (pEPitch<=pLowPitch) and (pEPitch<=pHighPitch) then pE.El.No.Stem := Dn
                    else
                    If (pEPitch>=pHighPitch) and (pEPitch>=pLowPitch) then pE.El.No.Stem := Up
                  End ;
                  If pE = pLast then pE := Nil else pE := pE.Next
              End ;
          End ; { ReduceNoUnsets }

          Procedure RemoveOverlap ;
          Var Done    : Boolean ;
              Changes : Integer ;
              UpSize, DnSize, UpDuration, DnDuration : Integer ;

              Procedure ChangeLow(Var pLow: pElementList);
              // Changes stem direction of pLow from Up to Dn.
              // UpDuration, DnDuration are note lengths for pLow and pHigh.
              Var LowPitch : Integer ;
              Begin
                  // Change stem of current pLow:
                  pLow.El.No.Stem := Dn ;
                  If UpDuration <> DnDuration then
                  With pLow.El.No do
                  Begin // Set duration of pLow to that of pHigh:
                      NoteSym  := pHigh.El.No.NoteSym ;
                      Dotted   := pHigh.El.No.Dotted ;
                      DoubleDot:= pHigh.El.No.DoubleDot ;
                  End;
                  Inc(DnSize) ; Dec(UpSize) ;
                  // Find new pLow ... & remove other notes at previous low:
                  LowPitch := pLow.El.No.Pitch ;
                  pLow := HighestOrLowest(Up) ;
                  While (pLow <> Nil) and (LowPitch = pLow.El.No.Pitch) do
                  Begin
                      pLow.El.ElType := Unknown ;
                      Dec(UpSize) ;
                      pLow := HighestOrLowest(Up) ;
                  End;
              End ; { ChangeLow }

              Procedure ChangeHigh(Var pHigh: pElementList);
              // Changes stem direction of pHigh from Dn to Up.
              // UpDuration, DnDuration are note lengths for pLow and pHigh.
              Var HighPitch : Integer ;
              Begin
                  // Change stem of current pHigh:
                  pHigh.El.No.Stem := Up ;
                  If UpDuration <> DnDuration then
                  With pHigh.El.No do
                  Begin // Set duration of pHigh to that of pLow:
                      NoteSym  := pLow.El.No.NoteSym ;
                      Dotted   := pLow.El.No.Dotted ;
                      DoubleDot:= pLow.El.No.DoubleDot ;
                  End;
                  Inc(UpSize) ; Dec(DnSize) ;
                  // Find new pHigh ... & remove other notes at previous high:
                  HighPitch := pHigh.El.No.Pitch ;
                  pHigh := HighestOrLowest(Dn) ;
                  While (pHigh <> Nil) and (HighPitch = pHigh.El.No.Pitch) do
                  Begin
                      pHigh.El.ElType := Unknown ;
                      Dec(DnSize) ;
                      pHigh := HighestOrLowest(Dn) ;
                  End;
              End ; { ChangeHigh }

          Begin
              UpSize := PartSize(pFirst,pLast,Up) ;
              DnSize := PartSize(pFirst,pLast,Dn) ;
              UpDuration := NoteDuration(pLow) ;
              DnDuration := NoteDuration(pHigh);
              Changes := 0 ;
              Done := (pHigh=Nil) or (pLow=Nil) or (pHigh.El.No.Pitch<=pLow.El.No.Pitch) ;
              While not Done do // Change a note stem; Each iteration decreases the pitch difference of overlap.
              Begin
                  If (UpSize>DnSize) or ((UpSize=DnSize) and Odd(Changes)) then
                       ChangeLow(pLow)
                  else ChangeHigh(pHigh) ;
                  Inc(Changes) ;
                  Done := (pHigh=Nil) or (pLow=Nil) or (pHigh.El.No.Pitch<=pLow.El.No.Pitch)
              End ;
              If Changes>0 then // NB not all stem direction changes have been monitored.
                  Warning(8, 'The stem direction had to be changed on '
                       +IntToStr(Changes)+' notes in the chord at element '+
                  IntToStr(pFirst.El.ElementNo)+' of NW stave '+IntToStr(pFirst.El.StaveNo)) ;
          End ; { RemoveOverlap }

          Function AllTheSameLength : Boolean ;
          Var OK : Boolean ;
              Len, Dur : Integer ;
              pE : pElementList ;
          Begin
              OK  := True ;
              pE  := pFirst ;
              Len := 0 ;
              While pE <> Nil do
              Begin
                  If (pE.El.ElType = Note) then
                  Begin
                      Dur := NoteDuration(pE) ;
                      If Len = 0 then Len := Dur else
                      If Len<>Dur then OK := False ;
                  End ;
                  If pE=pLast then pE := Nil else pE := pE.Next
              End ;
              AllTheSameLength := OK
          End ; { AllTheSameLength }

          Function AllInDirection(St:StemType) : Boolean ;
          Var pE : pElementList ;
              pEstem : StemType ;
              AllTheSame : Boolean ;
          Begin
              pE  := pFirst ;
              AllTheSame := True ;
              While pE <> Nil do
              Begin
                  If (pE.El.ElType = Note) then
                  Begin
                      pEstem := pE.El.No.Stem ;
                      If (pEstem in [Up,Dn]) and (pEstem<>St) then AllTheSame := False
                  End ;
                  If pE=pLast then pE := Nil else pE := pE.Next
              End ;
              AllInDirection := AllTheSame
          End ; { AllInDirection }

          Procedure FixAllStems(St:StemType) ;
          // Sets all stems to same direction
          Var pE : pElementList ;
          Begin
              pE  := pFirst ;
              While pE <> Nil do
              Begin
                  If (pE.El.ElType = Note) then pE.El.No.Stem := St ;
                  If pE=pLast then pE := Nil else pE := pE.Next
              End ;
          End ; { FixAllStems }

          Procedure AssignDifferentStems ;
          // Makes longest notes all stem Dn and others stem Up (or vice versa)
          Var pE : pElementList ;
              ShortestStem,
              LongestStem     : StemType ;
              LongestHasStem  : Boolean ;
              LongestLen, Dur : Integer ;
          Begin
              // Find longest note with stem Up or Dn:
              LongestHasStem := False ;
              LongestStem    := Unset ;
              LongestLen     := 0 ;
              pE := pFirst ;
              While pE <> Nil do
              Begin
                  If (pE.El.ElType = Note) then
                  Begin
                      Dur := NoteDuration(pE) ;
                      If LongestLen = 0 then
                      Begin
                          LongestLen     := Dur ;
                          LongestStem    := pE.El.No.Stem ;
                          LongestHasStem := LongestStem in [Up,Dn]
                      End
                      else
                      If (LongestLen = Dur) and not LongestHasStem then
                      Begin
                          LongestStem    := pE.El.No.Stem ;
                          LongestHasStem := LongestStem in [Up,Dn]
                      End
                      else
                      If (LongestLen > Dur) then
                      Begin
                          LongestLen     := Dur ;
                          LongestHasStem := pE.El.No.Stem in [Up,Dn] ;
                          If LongestHasStem then LongestStem := pE.El.No.Stem ;
                      End
                  End ;
                  If pE=pLast then pE := Nil else pE := pE.Next
              End ;
              // Assign stem of longest note:
              If not LongestHasStem then
              Begin
                  If LongestStem = Dn then LongestStem := Up else LongestStem := Dn
              End ;
              If LongestStem = Dn then ShortestStem := Up else ShortestStem := Dn ;
              // Assign all stems:
              pE := pFirst ;
              While pE <> Nil do
              Begin
                  If (pE.El.ElType = Note) then
                      If NoteDuration(pE) = LongestLen then pE.El.No.Stem := LongestStem
                      else pE.El.No.Stem := ShortestStem ;
                  If pE=pLast then pE := Nil else pE := pE.Next
              End ;
          End ; { AssignDifferentStems }

          Procedure AssignRemainingStems ;
          // pLow should be higher than pHigh and
          // pLow..pHigh contain all unassigned stems
          Var pE : pElementList ;
              DnDuration, Unsets : Integer ;
          Begin
              Unsets := PartSize(pFirst,pLast,Unset) ;
              DnDuration := NoteDuration(pHigh);
              pE := pFirst ;
              If Unsets > 0 then
              While pE <> Nil do
              Begin
                  If (pE.El.ElType = Note) and not (pE.El.No.Stem in [Up,Dn]) then
                  Begin
                      If pE.El.No.Pitch <= pHigh.El.No.Pitch then pE.El.No.Stem := Dn else
                      If pE.El.No.Pitch >= pLow.El.No.Pitch  then pE.El.No.Stem := Up else
                      If NoteDuration(pE) = DnDuration then
                         Begin pHigh := pE ; pE.El.No.Stem := Dn End else
                         Begin pLow  := pE ; pE.El.No.Stem := Up End
                  End ;
                  If pE=pLast then pE := Nil else pE := pE.Next
              End ;
          End ; { AssignRemainingStems }

Begin { FixOverlappingChordElements }
    pFirst := pChord.Next ;
    FixStemless ;          // Should be called also for chords with rests
                           // (Doesn't fix all stems.)
    FixCrotchets(Up);      // Should be called also for chords with rests
    FixCrotchets(Dn);      // Should be called also for chords with rests
    Unsets := PartSize(pFirst,pLast,Unset) ;
    If Unsets>0 then       // Try to assign some stem directions:
    If AllTheSameLength then
    Begin
       If AllInDirection(Up) then FixAllStems(Up) else
       If AllInDirection(Dn) then FixAllStems(Dn)
    End
    else AssignDifferentStems ;
    pLow   := HighestOrLowest(Up) ; // Excludes [NoStem,Auto,Unset]
    pHigh  := HighestOrLowest(Dn) ; // ditto
    ReduceNoUnsets(pHigh,pLow) ;    // Assigns stems except between pHigh and pLow
    RemoveOverlap ;
    ReduceNoUnsets(pHigh,pLow) ;    // Assigns stems except between pHigh and pLow
    AssignRemainingStems ;          // Assigns stems between pHigh and pLow
End ; { FixOverlappingChordElements }


Procedure GetChordElements(Var pFirst, pLast: pElementList) ;
{ First note of the chord being constructed is the first note or rest at or after
  the input value pFirst.
  pFirst and pLast are returned as the first and last notes/rests in the chord.
  It is assumed that elements from different chords don't overlap in the NW file.
  Generally, notes which start at the same time have the same horizontal coords,
  irrespective of stem direction or if a grace note. (The horiz positions can be 1 apart.)
  The exception is notes with reversed heads or grace and full notes on same or adjacent pitches.
  Also, overlapping stem up and stem down chords have stem dn notes moved to right.
  For stem down, the reversed head notes come first in the chord (in list and in horiz value).
  For stem up, the reverse head note come afterwards.
  Reversed head notes must belong to a chord.

  Stem up and stem dn chords overlap iff bottom note of stem up is below top note of stem dn.
  Overlapping grace note chords are 5 apart: stem up is 5 less than stem dn.
  Overlapping full note chords are 8 to 11 apart (11 in MP245).
  MP118 contains a grace note 6 pixels after full note, but not in same chord.
  MP118 contains chord with grace note and full note 4 apart, same stem direction.
  MP162 contains two full notes, same stem direction 12 apart but not in same chord.

  NW files can contain chords with reversed heads & upward stems at start of chords.
  They are incorrectly stored or incorrectly printed by NWC, so the test is one of proximity:
  notes of different chords can be as close as 16 pixels apart, those in the same chord up to 13 apart
  (for full sizenotes - grace notes can be closer).
  Note that adjacent notes with opposite stem directions may not include a reversed head
  but can be up to 11 pixels apart.
  Two close semiquavers not in a chord are 12 pixels apart in MP162,

  Assumes that grace notes can be ordinary notes of a chord in NW files.

  Code assumes one, two or three possible positions (of note heads) in chord,
  and two of stems.
     If there's an overlap of stem up and stem dn chords,
     then exactly two positions for heads, one for the stems.
     If there's no overlap between stem up and stm dn then one main position,
     for heads and two for stems
        plus one more head position if reverse head stem up note,
        plus one more head position if reverse head stem dn note.
  // Limit of horiz extent over which chord elements are sought.
  // Heads can be 13 apart, chords 3 heads wide.
}
          Type Decision = (No,Yes,MayBe) ;

          Function InSameChord(pE1,pE2:pElementList) : Decision ;
          // Assumes pE1 and pE2 are notes or rests.
          // Uses Horiz Positions, stem direction & grace note properties to see if pE1 and pE2 are in same chord.
          // "Maybe" is a decision which neither confirms nor denies membership in the same chord.
          // Jan 17: Scale the constants for when notes are closely spaced on stave?
          Const SameChordLimit   = 14 ; // Ordinary notes in same chord can be 13 apart if
                                        // heads on different sides of common stem position.
                                        // Do not increase: see MP162 comment above.
                GraceNoteLimit   =  7 ; // Grace notes in same chord can be 6 apart if
                                        // heads on different sides of common stem position.
                                        // Grace notes in different chords can be only 7 apart.
                SameDirnLimit    =  5 ; // Full note and grace note with same stem can be 4 apart in same chord.
                                        // Notes of same type can be up to 1 apart.
          Var BothGrace     : Boolean ;
              HP1,HP2,Limit : Integer ;
              pTmp          : pElementList ;
              St1, St2      : StemType ;
          Begin
              BothGrace := pE1.El.No.Grace and pE2.El.No.Grace ;
              If BothGrace then Limit := GraceNoteLimit else Limit := SameChordLimit ;
              If pE1.El.HorizPosn > pE2.El.HorizPosn then
              Begin pTmp := pE2; pE2 := pE1; pE1 := pTmp End ;
              // pE1 is now to the left of pE2
              HP1 := pE1.El.HorizPosn ;  St1 := pE1.El.No.Stem ;
              HP2 := pE2.El.HorizPosn ;  St2 := pE2.El.No.Stem ;
              If HP2-HP1 < SameDirnLimit then InSameChord := Yes else
              If (St1 = Dn) and (St2 in [Up,Dn,Auto]) then InSameChord := No else
              If (St1 = Up) and (St2 = Up) then InSameChord := No else
              If HP2-HP1 < Limit then
                  // Need to attach nostem cases to some chord:
                  If ((St1=NoStem)or(St2=NoStem)) and (St1<>St2) then InSameChord := Yes else
                  If (St1=Up) and (St2=Dn) and (pE1.El.No.Pitch=pE2.El.No.Pitch+1)
                      then InSameChord := Yes
                  else InSameChord := MayBe // Unsure if any cases of this can arise
              else
              If (St1 in [Up,Dn,Auto]) then InSameChord := No else
              If (St2 in [Up,Dn,Auto]) then InSameChord := No else
              // Remaining cases involve stems in [NoStem,Auto,Unset]:
              If HP2-HP1 >= 2*Limit then InSameChord := No else InSameChord := MayBe
          End ; { InSameChord }

          Function RightMost(pLeft,pLast: pElementList) : pElementList ;
          // Finds rightmost note or rest that might be in same chord as pFirst.
          Var pE, pNewLast : pElementList ;
          Begin
              pE := pLeft ;
              pNewLast := pLeft ;
              While pE <> Nil do
              Begin
                  If pE.El.ElType in [Note,Rest] then
                  If InSameChord(pLeft,pE) = No then pLast := pE else pNewLast := pE ;
                  If (pE<>pLast) then pE := pE.Next else pE := Nil
              End ;
              RightMost := pNewLast
          End ; { RightMost }

          Function LeftMost(pFirst,pLast: pElementList) : pElementList ;
          // Finds leftmost note or rest in range.
          Var pF, pLeft : pElementList ;
          Begin
              pLeft := Nil ;
              pF := pFirst ;
              While pF<>Nil do
              Begin
                  If (pF.El.ElType in [note,rest]) then
                  If pLeft = Nil then pLeft := pF else
                  If pLeft.El.HorizPosn > pF.El.HorizPosn then pLeft := pF ;
                  If (pF<>pLast) then pF := pF.Next else pF := Nil
              End ;
              LeftMost := pLeft
          End ; { LeftMost }

Const ChordEnd  = [Bar,Clef,TimeSig,KeySig] ;
      Max       = 40 ; // Upper bound on #notes in search space pFirst..pLast
Var   pE, pF, pPred, pLeft : pElementList ;
      Doubt,   // True iff two notes in potential chord reported "maybe" for closeness.
      Done    : Boolean ;
      LastOK    // Index of last note so far known to be in chord containing pLeft.
              : Integer ;
      OKNotes : Array[1..Max] of pElementList ; // List of possible members of the chord.

          Procedure InitOKNotes(pLeft: pElementList) ;
          Var I : Integer ;
          Begin
              For I := 2 to Max do OKNotes[I] := Nil ;
              OKNotes[1] := pLeft ;
              LastOK := 1 ;
          End ; { InitOKNotes }

          Function IsOK(pE:pElementList) : Boolean ;
          // True iff pE in OKNotes[1..LastOK]
          Var I : Integer ;
              Found : Boolean ;
          Begin
              Found := False ;
              For I := 1 to LastOK do if OKNotes[I] = pE then Found := True ;
              IsOK := Found
          End ; { IsOK }

          Procedure AddElement(pE,pF:pElementList);
          // Adds pF to OKNotes if pE is already there.
          // OKNotes[1..LastOK] are defined ;
          Begin
              If not IsOK(pF) and IsOK(pE) then
              Begin Inc(LastOK) ; OKNotes[LastOK] := pF End ;
          End ; { AddElement }

Begin { GetChordElements }
    // 1A. Find initial limits pFirt and pLast on Chord element range:
    While (pFirst<>Nil) and not (pFirst.El.ElType in [Note,Rest]) do pFirst := pFirst.Next ;
    Done  := pFirst=nil ;
    pLast := pFirst ;
    While (pLast<>Nil) and not (pLast.El.ElType in ChordEnd) do pLast := pLast.Next ;

    // 2. Determine pLeft as leftmost member of chord & update pLast as rightmost chord member:
    If not Done then // pFirst is the first note/rest.
    Begin
        pLeft := LeftMost (pFirst,pLast) ;
        pLast := RightMost(pLeft, pLast) ;
        Done  := pLast=pFirst ;
    End ;

    // 3. Check proximity of notes in pFirst..pLast and no pair are too far apart.
    //    Set Doubt to true if unsure about any pair & retract pLast if necessary.
    // The algorithm collects chord elements into the array OKNotes.
    // It selects exactly those for which there is at least one "yes"
    // and no "no"s when compared with notes already in OKNotes.
    If not Done then
    Begin
        InitOKNotes(pLeft) ;
        Doubt := False ;
        pE := pFirst ;
        While (pE <> Nil) do
        Begin // Check pFirst..pLast for chord members:
            If pE.El.ElType in [Note,Rest] then
            Begin
                pPred := pE ; // pPred is the rightmost note in pFirst..pE known to be in same chord as pE.
                pF := pFirst ;
                While pF <> Nil do
                Begin
                    If (pF<>pE) and (pF.El.ElType in [Note,Rest]) then
                    Case InSameChord(pE,pF) of
                        Yes   : Begin pPred := pF ; AddElement(pE,pF) End ;
                        MayBe : Begin pPred := pF ; Doubt := True     End ;
                        No    : Begin pLast := pPred ; pF := pLast    End ;
                    End ;
                    If pF=pLast then pF := Nil else pF := pF.Next ;
                End ;
            End ;
            If pE=pLast then pE := Nil else pE := pE.Next ;
        End ;
        Done := pLast=pFirst ;
    End ;

    // 4. Check only notes in OKNotes are in pFirst..pLast:
    If not Done and Doubt then
    Begin
        pE    := pFirst ;
        pPred := pFirst ;
        While pE <> Nil do
        Begin
            If (pE.El.ElType in [Note,Rest]) and not IsOK(pE) then
                   Begin pLast := pPred ; pE := pLast End ;
            If pE=pLast then pE := Nil else Begin pPred := pE; pE := pE.Next End
        End ;
    End ;
End; { GetChordElements }


 Procedure ShortenNotes(pFirst, pLast : pElementList; Stem : StemType) ;
// This needs updating to allow for stems which are neither Up nor Dn.
Var pF, pShort : pElementList ;
    LenpF, MinLen : Integer ;
    Changed : Boolean ;

          Function ShortestInPart(pFirst, pLast : pElementList; Stem: StemType) : pElementList ;
          Var pF, pShort : pElementList ;
              ShortestSoFar, LenPF : Integer ;
          Begin
              ShortestSoFar := MaxDur ;
              pShort := Nil ;
              pF := pFirst ;
              While pF<>Nil do
              Begin
                  If (pF.El.ElType in [Rest,Note]) and (pF.El.No.Stem=Stem) then
                  Begin
                      LenpF := NoteDuration(pF) ;
                      If (LenpF < ShortestSoFar) and (LenpF<>0) then
                      Begin ShortestSoFar := LenpF ; pShort := pF End ;
                  End ;
                  If (pF<>pLast) then pF := pF.Next else pF := Nil
              End ;
              ShortestInPart := pShort
          End ; //  ShortestInPart
Begin //  ShortenNotes
    pShort := ShortestInPart(pFirst, pLast, Stem) ;
    MinLen := NoteDuration(pShort) ;
    Changed:= False ;
    pF := pFirst ;
    While pF<>Nil do
    Begin
        If (pF<>pShort) and (pF.El.ElType in [Rest,Note]) and (pF.El.No.Stem=Stem) then
        Begin
            LenpF := NoteDuration(pF) ;
            If LenpF > MinLen then
            With pF.El.No do
            Begin  // Set length of pF to that of pShort:
                NoteSym  := pShort.El.No.NoteSym ;
                Dotted   := pShort.El.No.Dotted ;
                DoubleDot:= pShort.El.No.DoubleDot ;
                Changed  := True ;
            End
        End ;
        If (pF<>pLast) then pF := pF.Next else pF := Nil
    End ;
    If Changed then Error(59,'Shortened notes in a chord part')
End ; // ShortenNotes



Procedure MakeChord(pChord,pLast: pElementList) ;
{ This routine processes the chord elements to create an acceptable chord
  and initialises pChord:
    1 Checks for presence of rest
    2 Finds shortest note and shortest rest
    3 Changes rests to grace notes to ensure at most one rest with shortest length
    4 If rest in chord:
       a) Fixes stems of notes to be opposite that of the rest
       b) Shortens notes to all have the same length
    5 If no rest in chord:
       a) Shortens notes of upper and lower parts to have the same length
       b) Shifts notes from one part to the other to avoid parts overlapping
    6 Determines right offset for note heads
    7 Selects note/rest for Chord element
    8 Creates and inserts Chord or ReChord element or both
    9 For ReChord case, removes rest element
  It ensures the stem up and stem down parts don't overlap, and
  rests are changed to grace notes if there are shorter notes or other rests.
  The remaining processing is done by CreateReChordElements which deals with chords with rests.
}
Var pFirst, pRest : pElementList ;
    NoteLen, RestLen, ShortestLen : Integer ;
    HasRest, RestIsShorter : Boolean ;

        Function ReduceToOneRest(pFirst,pLast : pElementList; ShortestLen: Integer) : pElementList ;
        // Change all but at most one rest to grace notes, and returns remaining rest.
        // If all rests are longer than ShortestLen, then all are changed to grace notes.
        // NB Consider removing the rests instead, but implications for timing of subsequent notes.
        // May need to work on pitch of the new grace notes.
        Var pF, pRest : pElementList ;
            Changed, SeenRest : Boolean ;
        Begin
            SeenRest := False;  Changed := False;  pRest := Nil;
            pF := pFirst;
            While pF<>Nil do
            Begin
                If pF.El.ElType = Rest then
                    If SeenRest or (NoteDuration(pF) > ShortestLen) then
                    Begin Changed := True; pF.El.ElType := Note; pF.El.No.Grace  := True End
                    else
                    Begin SeenRest := True; pRest := pF End ;
                If (pF<>pLast) then pF := pF.Next else pF := Nil
            End ;
            If Changed then Error(58,'Changed some rests in chord to grace notes with meaningless pitch') ;
            ReduceToOneRest := pRest
        End ; { ReduceToOneRest }

        Function GetRest(pFirst, pLast : pElementList) : pElementList ;
        Var pF, pRest : pElementList ;
        Begin
            pRest := Nil;  pF := pFirst ;
            While (pF<>Nil) do
            Begin
                If (pF.El.ElType = Rest) then Begin pRest := pF ; pF := pLast End ;
                If (pF<>pLast) then pF := pF.Next else pF := Nil
            End ;
            GetRest := pRest
        End ; { GetRest }

        Procedure SetInChord(pFirst, pLast : pElementList) ;
        Var pE : pElementList ;
        Begin
            pE := pFirst ;
            While pE <> Nil do
            Begin
                If pE.El.ElType in [Note,Rest,Chord] then pE.El.No.InChord := True ;
                If pE=pLast then pE := Nil else pE := pE.Next
            End
        End ; { SetInChord }

        Procedure SlurChord(pFirst,pLast : pElementList) ;
        // Copy slur values to rest of chord.
        Var pE, pSlur : pElementList ;
        Begin
            pSlur := Nil ;
            pE    := pFirst ;
            While pE<>Nil do
            Begin
                With pE.El do
                If (ElType = Note) and (No.SlurIn or No.SlurOut) then pSlur := pE ;
                If pE = pLast then pE := Nil else pE := pE.Next
            End ;
            pE := pFirst ;
            If pSlur<>Nil then
            While pE<>Nil do
            Begin
                With pE.El do
                If (pE<>pSlur) and (ElType = Note) and (No.NoteSym = pSlur.El.No.NoteSym) then
                Begin
                    No.SlurIn    := pSlur.El.No.SlurIn ;
                    No.SlurOut  := pSlur.El.No.SlurOut ;
                    No.SlurPosn := pSlur.El.No.SlurPosn
                End ;
                If pE = pLast then pE := Nil else pE := pE.Next
            End ;
        End ; { SlurChord }

        Procedure BeamChord(pChord,pLast: pElementList) ;
        Var pE : pElementList ;
            BeamInUp, BeamInDn, BeamOutUp, BeamOutDn : Boolean ;
            BeamUpPosn, BeamDnPosn : Position ;
        Begin
            BeamInUp  := False ;
            BeamInDn  := False ;
            BeamOutUp := False ;
            BeamOutDn := False ;
            pE        := pChord ;
            While pE<>Nil do
            Begin
                If (pE.El.ElType = Note) then
                With pE.El.No do
                Case Stem of
                Up : Begin
                        If BeamIn  then BeamInUp  := True ;
                        If BeamOut then BeamOutUp := True ;
                        If BeamIn or BeamOut then BeamUpPosn := BeamPosn
                     End ;
                Dn : Begin
                        If BeamIn  then BeamInDn  := True ;
                        If BeamOut then BeamOutDn := True ;
                        If BeamIn or BeamOut then BeamDnPosn := BeamPosn
                   End ;
                else
                End ;
                If pE = pLast then pE := Nil else pE := pE.Next
            End ;
            pE := pChord ;
            If BeamInUp or BeamInDn or BeamOutUp or BeamOutDn then
            While pE<>Nil do
            Begin
                If pE.El.ElType = Note then
                With pE.El.No do
                If Stem = Up then
                Begin
                    If BeamInUp  then BeamIn  := True ;
                    If BeamOutUp then BeamOut := True ;
                    If BeamInUp or BeamOutUp then BeamPosn := BeamUpPosn
                End
                else
                If Stem = Dn then
                Begin
                    If BeamInDn  then BeamIn  := True ;
                    If BeamOutDn then BeamOut := True ;
                    If BeamInDn or BeamOutDn then BeamPosn := BeamDnPosn
                End ;
                If pE = pLast then pE := Nil else pE := pE.Next
            End ;
        End ; { BeamChord }

        Procedure TripletChord(pChord,pLast: pElementList) ;
        Var pE : pElementList ;
            TripletInUp, TripletInDn, TripletOutUp, TripletOutDn : Boolean ;
        Begin
            TripletInUp  := False ;
            TripletInDn  := False ;
            TripletOutUp := False ;
            TripletOutDn := False ;
            pE := pChord ;
            While pE<>Nil do
            Begin
                If (pE.El.ElType = Note) then
                With pE.El.No do
                Case Stem of
                Up : Begin
                        If TripletIn  then TripletInUp  := True ;
                        If TripletOut then TripletOutUp := True
                     End ;
                Dn : Begin
                        If TripletIn  then TripletInDn  := True ;
                        If TripletOut then TripletOutDn := True
                   End ;
                else
                End ;
                If pE = pLast then pE := Nil else pE := pE.Next
            End ;
            pE := pChord ;
            If TripletInUp or TripletInDn or TripletOutUp or TripletOutDn then
            While pE<>Nil do
            Begin
                If pE.El.ElType = Note then
                With pE.El.No do
                If Stem = Up then
                Begin
                    If TripletInUp  then TripletIn  := True ;
                    If TripletOutUp then TripletOut := True ;
                End
                else
                If Stem = Dn then
                Begin
                    If TripletInDn  then TripletIn  := True ;
                    If TripletOutDn then TripletOut := True ;
                End ;
                If pE = pLast then pE := Nil else pE := pE.Next
            End ;
        End ; { TripletChord }

        Function Highest(pFirst,pLast: pElementList; Len: Integer) : pElementList ;
        { Finds highest *note* of duration Len.
          There are problems for chords mixing grace note with normal notes.
          If the highest note is a grace note, the whole chord is made into grace notes
          in the NWC editor. If the highest non-grace note is chosen, all the chord
          appears as normal notes in the NWC editor. }
        Var pF, pHighest : pElementList ;
        Begin
            pHighest := Nil ;
            pF := pFirst ;
            While pF<>Nil do
            Begin
                If (pF.El.ElType=Note) and (NoteDuration(pF) = Len) then
                If (pHighest=Nil) or (pF.El.No.Pitch > pHighest.El.No.Pitch) then pHighest := pF ;
                If (pF<>pLast) then pF := pF.Next else pF := Nil
            End ;
            Highest := pHighest
        End ; { Highest }

        Function Shortest(pFirst, pLast : pElementList; NoRe:StaveElement) : Integer ;
        // Looks at notes OR rests in [pFirst..pLast] to find that of shortest duration.
        Var pF : pElementList ;
            ShortestSoFar, LenPF : Integer ;
        Begin
            ShortestSoFar := MaxDur ;
            pF := pFirst ;
            While pF<>Nil do
            Begin
                If pF.El.ElType=NoRe then
                Begin
                    LenpF := NoteDuration(pF) ;
                    If (LenpF < ShortestSoFar) and (LenpF<>0) then ShortestSoFar := LenpF
                End ;
                If (pF<>pLast) then pF := pF.Next else pF := Nil
            End ;
            Shortest := ShortestSoFar
        End ; { Shortest }

Begin { MakeChord }
    pFirst  := pChord.Next ;
    pRest   := GetRest (pFirst,pLast) ;
    HasRest := pRest<>Nil ;
    If HasRest then
    Begin // Check no shorter notes. If so, rests need changing to grace notes.
        NoteLen := Shortest(pFirst,pLast,Note) ;
        RestLen := Shortest(pFirst,pLast,Rest) ;
        RestIsShorter := RestLen<NoteLen ;
        If NoteLen < RestLen then ShortestLen := NoteLen else ShortestLen :=  RestLen ;
        pRest := ReduceToOneRest(pFirst,pLast,ShortestLen) ;
        If NoteLen < RestLen then // All rests were changed to grace notes.
            Begin pRest := Nil; HasRest := False End ;
        If HasRest then FixStems(pFirst,pLast,pRest{,RestIsShorter}) ;
    End ;
    If not HasRest then FixOverlappingChordElements(pChord,pLast) ;
//  ShortenNotes(pFirst,pLast,Up)  ;
//  ShortenNotes(pFirst,pLast,Dn)  ;
    ReorderChordEls(pChord,pLast)  ; // moves non-note/non-rest elements to after chord
    AssignRightShifts(pFirst,pLast); // Determines which note heads are on right side of stem.
    RightShiftChordEl(pChord,pLast); // Updates pChord to include right shift
    SlurChord(pFirst,pLast) ;        // Copies any slur to all chord elements
    BeamChord(pChord,pLast) ;        // Copies any beam to all chord elements
    TripletChord(pChord,pLast) ;     // Copies any triplet to all chord elements
    // Initialise pChord element:
    NoteLen         := Shortest(pFirst,pLast,Note) ;
    pChord.El       := Highest(pFirst,pLast,NoteLen).El ;
    pChord.El.ElType:= Chord ;
    pChord.El.No.WithReChord := HasRest ; // Indicates need for ReChord element
    SetInChord(pChord, pLast) ;
End ; { MakeChord }

Procedure MakeChords(pS : pStave) ;
// Inserts a chord element where there is more than one note/rest at the same horizontal posn.
// A further pass over the elments of pS will create ReChord elements.
Var pE, pInsert,
    pFirst,  // The first note in the chord
    pLast,   // The last note in the chord.
    pChord : pElementList ;
    ChordSize: Integer ;

        Function NoReCount(pFirst, pLast : pElementList) : Integer ;
        Var Size    : Integer ;
            pE      : pElementList ;
        Begin
            Size := 0 ; pE := pFirst ;
            While pE <> Nil do
            Begin
                If pE.El.ElType in [Note,Rest] then Inc(Size) ;
                If pE=pLast then pE := Nil else pE := pE.Next
            End ;
            NoReCount := Size
        End ; { NoReCount }

Begin
    pE := pS.Elements ;
    If pE<>Nil then
    Repeat // Find next chord starting at or after pE:
        pFirst   := pE ;
        GetChordElements(pFirst, pLast) ;
        ChordSize := NoReCount(pFirst, pLast) ;
        If ChordSize>1 then
        Begin
            pChord := NewElement(Chord,pFirst) ;
            MakeChord(pChord,pLast) ;
            ChordSize := NoReCount(pFirst, pLast) ; // MakeChord may have changed the size.
            pChord.El.No.NoInChord := ChordSize;  // (Includes any rests)
            pS.NoElements:= pS.NoElements - ChordSize + 1 ;
            pInsert := pS.Elements ;
            If pInsert = pFirst then pS.Elements := pChord else
            Begin
               While (pInsert<>Nil) and (pInsert.Next <> pFirst) do pInsert := pInsert.Next ;
               If (pInsert=Nil) then Error(20,'Failed to insert chord element')
               else pInsert.Next := pChord
            End ;
        End ;
        If pLast = nil then pE := nil else pE := pLast.Next
    until pE=Nil ;
End ; { MakeChords }


{#############################################################################}
{#############################################################################}
{###########################    Slur Processing   ############################}
{#############################################################################}
{#############################################################################}



Function SlurAngle (p1,p3:pElementList) : Real ;
// Returns gradient from p1 to p3.
// (Actually negative gradient as VertPosn *increases* downwards.
Var  V1,V3,H1,H3 : Integer ;
Begin
    V1 := p1.El.VertPosn ;
    H1 := p1.El.HorizPosn;
    V3 := p3.El.VertPosn ;
    H3 := p3.El.HorizPosn;
    If (H1=H3) then
    Begin
      Error(8,'Gradient error for slur') ;
      SlurAngle := 0 ;
    End
    else
        SlurAngle := (V3-V1)/(H3-H1) ;
End ; { SlurAngle }

Function SlurOver (p1,p2,p3:pElementList) : Boolean ;
// Uses slur curvature to determines if slur is over or under the notes.
Var  V1,V2,V3,H1,H2,H3 : Integer ;
     Grad12,Grad23 : Real ;
     Over : Boolean ;
Begin
    V1 := p1.El.VertPosn ;
    H1 := p1.El.HorizPosn;
    V2 := p2.El.VertPosn ;
    H2 := p2.El.HorizPosn;
    V3 := p3.El.VertPosn ;
    H3 := p3.El.HorizPosn;
    If (H1>=H2) or (H2>=H3) then
    Begin
      Error(9,'Gradient error for slur') ;
      Over := True ;
    End
    else
    Begin
        Grad12 := (V2-V1)/(H2-H1) ;
        Grad23 := (V3-V2)/(H3-H2) ;
        Over   :=  Grad12 < Grad23  //  Remember VertPosn *increases* downwards.
    End ;
    SlurOver := Over ;
End ; { SlurOver }

Function NoteToSlurDist(pNote,pSlur:pElementList; NoteFirst:Boolean) : Integer ;
{ Returns weighted abs horiz distance between slur end point pSlur and note pNote.
  Assumes element types as expected. }
Const WrongSideBias = 3 ;
      HalfNoteHead  = 5 ;
Var Hdist, AbsHD : Integer ;
Begin
    Hdist := pSlur.El.HorizPosn - pNote.El.HorizPosn; // Normally +ve when NoteFirst
    If Hdist<0 then AbsHD := -Hdist else AbsHD := Hdist ;
    If AbsHD<HalfNoteHead then NoteToSlurDist := AbsHD else
    Begin
      If not NoteFirst then Hdist := -Hdist ;
      // Hdist should normally be +ve.
      If Hdist < 0 then Hdist := HalfNoteHead-WrongSideBias*(Hdist+HalfNoteHead) ;
      NoteToSlurDist  := Hdist ;
    End
End ; { NoteToSlurDist }

Function VPdiff(p1,p2 : pElementList) : Integer ;
// Returns abs diff in vert posns of p1, p2.
Var D : Integer ;
Begin
    D := (p1.El.VertPosn - p2.El.VertPosn + 256) mod 256 ;
    If D<128 then VPdiff := D else VPdiff := 256-D
End ; { VPdiff }

Function ClosestSlurNote(pS : pStave; pBound1,pSlur,pBound2 : pElementList; First : Boolean) : pElementList ;
// pSlur is first or last slur element according to value of First.
Const MaxDist = 100 ;
Var pE, pNote : pElementList ;
    BestVdiff, VDiff, BarNo, Dist, BestDist : Integer ;
    S  : String ;
Begin
    // 1 Get closest note to pSlur in pBound1..pBound2:
    pNote    := nil ;
    BestDist := MaxDist ;
    pE       := pBound1 ;
    While pE<>nil do
    Begin
        If pE.El.ElType=note then
        Begin
            Dist := NoteToSlurDist(pE,pSlur,First) ; // True for first pt in slur
            If Dist < BestDist then
            Begin pNote := pE ; BestDist := Dist End ;
        End ;
        If pE = pBound2 then pE := Nil else pE := pE.Next
    End ;

    // 2 Repeat to get chord note at min Vert distance:
    If pNote<>Nil then
    Begin
        BestVdiff := VPdiff(pSlur,pNote) ;
        pE := pBound1 ;
        While pE<>Nil do
        Begin
            If pE.El.ElType=note then
            Begin
                Dist := NoteToSlurDist(pE,pSlur,First) ; // True for first pt in slur
                If Dist < BestDist+2 then
                Begin
                    VDiff := VPdiff(pSlur,pE) ;
                    If Vdiff < BestVdiff
                    then Begin pNote := pE; BestVdiff := VDiff End ;
                End ;
            End ;
            If pE = pBound2 then pE := Nil else pE := pE.Next
        End ;
    End ;

    // 3 Give error report:
    If pNote = Nil then
    Begin
        pE := pS.Elements ;
        BarNo := 1 ;
        While (pE<>nil) and (pE<>pSlur) do
            Begin If pE.El.ElType = Bar then Inc(BarNo) ; pE := pE.Next End ;
        S := 'slur in stave ' + IntToStr(pSlur.El.StaveNo) +
             ', starting in bar ' + IntToStr(BarNo) +
             ', element '+IntToStr(pSlur.El.ElementNo) ;
        If First then Error(11,'Not found note at start of '+S)
                 else Error(11,'Not found note at end of '+S)
    End;

    ClosestSlurNote := pNote
End ; { ClosestSlurNote }

Procedure BuildSlur(pS : pStave; pBound1,pSlur1,pMid,pSlur2,pBound2 : pElementList) ;
{ Slur is pSlur1..pSlur2 via pMid and is associated with notes in pBound1..pBound2.
  Some of these pts may be equal. All are non-nil slur pts.
 ... allows for incomplete slurs and updates slurs to unknown. }
Var pNote1, pNote2, pE : pElementList ;
    Posn : Position ;
Begin
    // NB: NW1.55 doesn't distinguish between slurs above or below end notes:
    If SlurOver(pSlur1,pMid,pSlur2) then Posn := Above else Posn := Below ;

    // 1 Get notes at slur ends1:
    pNote1 := ClosestSlurNote(pS,PBound1,pSlur1,pBound2,True) ;
    pNote2 := Nil ;
    If pSlur2.El.SPt.Next=0 then
        pNote2 := ClosestSlurNote(pS,PBound1,pSlur2,pBound2,False) ;
{
MessageDlg('Built slur in stave '+IntToStr(pS.StaveNo)+', slur pts '+
    IntToStr(pSlur1.El.ElementNo)+' to '+IntToStr(pSlur2.El.ElementNo)+
    ', and notes '+
    IntToStr(pNote1.El.ElementNo)+' to '+IntToStr(pNote2.El.ElementNo),
    mtInformation,[mbOK],0) ;
}
    // 2 Set slur:
    pE := pNote1 ;
    If (pNote1<>nil) and (pNote2<>Nil) then
    While pE<>nil do
    Begin
        If pE.El.ElType in [Note,Rest,Chord,ReChord] then
        Begin
            If pE<>pNote1 then Begin pE.El.No.SlurIn  := true ; pE.El.No.SlurPosn := Posn End ;
            If pE<>pNote2 then Begin pE.El.No.SlurOut := true ; pE.El.No.SlurPosn := Posn End ;
        End ;
        If pE = pNote2 then pE := Nil else pE := pE.Next
    End
End ; { BuildSlur }


Procedure FixSlurs(pS : pStave) ;
{ see above in NoteToSlurDist for an example.  SlurPts should occur in sets of three (or more),
  marking the start, middle and end of the slur. These elements include an element
  number for the next and previous elements in the triplet, 0 for no next/prev element.
  Such slur sets can be interleaved, e.g. in upper and lower parts on a stave.
  Slurs normally start 2 pixels after note to which they belong
  and finish 2 pixels before the finishing note.
  Slurs are between chords rather than notes.
  They can start and end arbitrarily high or low above the related chords.
  Sometimes they extend over stave ends. We leave the ends present in case
  joining systems joins them up.
  Pick vertically nearest note that's within 3 of 4 pixels of the horiz posn.
  Sample 1 has slurs starting 11 or more after the chords to which they apply.
  NB: v1.75 doesn't allow for choosing if slur is above or below,
  so don't know how to control this.
  If there's a barline/key sig/time sig/clef next to a slur end pt,
  the nearest chord on the other side is chosen.
  If there's a chord on either side, the usual choice is the chord to the left of the
  slur start or right of slur finish. The slur has to be less than 1/4 inter-chord
  distance from the other possible choice to select the chord within the boundary
  (This distance is between head edges, which are 5 or 6 from the head centre)
}
Const BoundaryEls = [bar,clef,rest,keysig,timesig] ;
Var pE, pF, pMid, pBound1, pBound2, pSlur1, pSlur2 : pElementList ;
    NextSlur : Integer ;
Begin
    pE := pS.Elements ;
    pBound1 := pS.Elements ;
    While pE<> Nil do
    Begin
       // 1 Get left boundary pBound1 of next slur and its start pSlur1:
       While (pE<>nil) and (pE.El.ElType<>SlurPt) do
       Begin
           If pE.El.ElType in BoundaryEls then pBound1 := pE ;
           pE := pE.Next ;
       End ;
       pSlur1 := pE ;

       If pSlur1 <> Nil then
       Begin
          // 2 Find mid pt pMid and end pt pSlur2 of current slur:
          pMid   := pSlur1 ;
          pSlur2 := pSlur1 ;
          pF     := pSlur1 ;
          While pF<>Nil  do
          Begin
              NextSlur  := pSlur2.El.SPt.Next  ;

// MessageDlg('NextSlur = '+IntToStr(NextSlur),mtInformation,[mbOK],0) ;

              If NextSlur<>0 then
              Begin
                  pMid := pSlur2 ; // Penultimate slur pt   ## Check this is a slur in/slur out pt?
                  Repeat
                      pF := pF.Next
                  until (pF=nil) or ((pF.El.ElementNo=NextSlur)and(pF.El.ElType=SlurPt)) ;
                  If pF=Nil then Error(10,'Missing slur end element')
              End
              else pF := Nil ;
              If pF <> Nil then pSlur2 := pF
          End ;
          // 3 Find right boundary pBound2 of current slur:
          pBound2 := pSlur2 ;
          While (pBound2<>Nil) and not (pBound2.El.ElType in BoundaryEls) do
                pBound2 := pBound2.Next ;
          // 4 Create slur:
          BuildSlur(pS,pBound1,pSlur1,pMid,pSlur2,pBound2) ;
          // 5 Delete splur pts:
          pF := pSlur1 ;
          While (pF<>Nil)and(pF<>pSlur2)  do
          Begin
              NextSlur  := pF.El.SPt.Next ;
              pF.El.ElType := Unknown ;
              If NextSlur<>0 then
              Repeat
                      pF := pF.Next
              until (pF=nil) or ((pF.El.ElementNo=NextSlur)and(pF.El.ElType=SlurPt))
              else pF := Nil ;
          End ;
          If pSlur2<>Nil then pSlur2.El.ElType := Unknown ;
          pE := pE.Next
       End
    End
End; { FixSlurs }


{#############################################################################}
{#############################################################################}
{#########################      Text Processing     ##########################}
{#############################################################################}
{#############################################################################}

Procedure InitPerfTxt ;
Begin
     PerfTxt[AdLib    ] := 'adlibitum';  PerfMin[AdLib    ] := 5 ;
     PerfTxt[Animato  ] := 'animato'  ;  PerfMin[Animato  ] := 4 ;
     PerfTxt[Cantabile] := 'cantabile';  PerfMin[Cantabile] := 4 ;
     PerfTxt[ConBrio  ] := 'conbrio'  ;  PerfMin[ConBrio  ] := 7 ;
     PerfTxt[Dolce    ] := 'dolce'    ;  PerfMin[Dolce    ] := 5 ;
     PerfTxt[Espress  ] := 'espressivo'; PerfMin[Espress  ] := 6 ;
     PerfTxt[Grazioso ] := 'grazioso' ;  PerfMin[Grazioso ] := 4 ;
     PerfTxt[Legato   ] := 'legato'   ;  PerfMin[Legato   ] := 3 ;
     PerfTxt[Maestoso ] := 'maestoso' ;  PerfMin[Maestoso ] := 5 ;
     PerfTxt[Marc     ] := 'marcando' ;  PerfMin[Marc     ] := 4 ;
     PerfTxt[MenoMosso] := 'menomosso';  PerfMin[MenoMosso] := 9 ;
     PerfTxt[PocoAPoco] := 'pocoapoco';  PerfMin[PocoAPoco] := 9 ;
     PerfTxt[PiuMosso ] := 'piumosso' ;  PerfMin[PiuMosso ] := 8 ;
     PerfTxt[Semplice ] := 'semplice' ;  PerfMin[Semplice ] := 5 ;
     PerfTxt[Simile   ] := 'simile'   ;  PerfMin[Simile   ] := 3 ;
     PerfTxt[Solo     ] := 'solo'     ;  PerfMin[Solo     ] := 4 ;
     PerfTxt[Sostenuto] := 'sostenuto';  PerfMin[Sostenuto] := 4 ;
     PerfTxt[SottoVoce] := 'sottovoce';  PerfMin[SottoVoce] := 8 ;
     PerfTxt[Stacc    ] := 'staccato' ;  PerfMin[Stacc    ] := 4 ;
     PerfTxt[Subito   ] := 'subito'   ;  PerfMin[Subito   ] := 3 ;
     PerfTxt[Ten      ] := 'tenuto'   ;  PerfMin[Ten      ] := 3 ;
     PerfTxt[Tutti    ] := 'tutti'    ;  PerfMin[Tutti    ] := 5 ;
     PerfTxt[VoltaSubito]:='voltasubito';PerfMin[VoltaSubito]:=8 ;
     PerfTxt[NotFound ] := '' ;          PerfMin[NotFound ] := 1 ;

     VolVarTxt[Cresc  ] := 'crescendo'  ; VolVarMin[Cresc  ] := 5 ;
     VolVarTxt[Decresc] := 'decrescendo'; VolVarMin[Decresc] := 6 ;
     VolVarTxt[Dim    ] := 'diminuendo' ; VolVarMin[Dim    ] := 3 ;
     VolVarTxt[Rinf   ] := 'rfz'   ;      VolVarMin[Rinf   ] := 4 ;
     VolVarTxt[Sforz  ] := 'sfz'   ;      VolVarMin[Sforz  ] := 2 ;
     VolVarTxt[Unfound] := ''      ;      VolVarMin[Unfound] := 1 ;

     FlowTxt[NoFlow  ] := '' ;
     FlowTxt[Coda    ] := 'coda'   ;
     FlowTxt[Segno   ] := ''       ;
     FlowTxt[Fine    ] := 'fine'   ;
     FlowTxt[ToCoda  ] := 'tocoda' ;
     FlowTxt[DaCapo  ] := 'dacapo' ;
     FlowTxt[DCalCoda] := 'dcalcoda' ;
     FlowTxt[DCalFine] := 'dcalfine' ;
     FlowTxt[DalSegno] := 'dalsegno' ;
     FlowTxt[DSalCoda] := 'dsalcoda' ;
     FlowTxt[DSalFine] := 'dsalfine' ;
End ; { InitPerfTxt }

Function RemovePunctuation(I : Integer) : String ;
// Removes spaces and periods from Strings[I]
Const Period = '.' ;
Var S : String  ;
    P : Integer ;
Begin
    S :=  AnsiLowerCase(Strings[I]) ;
    While Pos(Space,S)>0 do
    Begin
        P := Pos(Space,S) ;
        S := Copy(S,0,P-1)+Copy(S,P+1,Length(S)) ;
    End ;
    While Pos(Period,S)>0 do
    Begin
        P := Pos(Period,S) ;
        S := Copy(S,0,P-1)+Copy(S,P+1,Length(S)) ;
    End ;
    RemovePunctuation := S ;
End ; { RemovePunctuation }


Function VolVariation(pE : pElementList ) : VolVarType ;
// Marks text element as Volume Variation indication where appropriate
// Requires text to be initial substring of VolVarTxt element, with optional fullstop.
Var VV, V : VolVarType ;
    S     : String  ;
    SI, L : Integer ;
Begin
    SI := pE.El.Txt.StringIndex ;
    S  := RemovePunctuation(SI) ;
    L  := Length(S) ;
    VV := UnFound ;
    For V := Cresc to Sforz do
    If (L>=VolVarMin[V]) and (L<=Length(VolVarTxt[V])) and (Pos(S,VolVarTxt[V])=1) then VV := V ;
    VolVariation := VV ;
End ; { VolVariation }

Function PerformanceStyle(pE : pElementList) : PerfStyle ;
// Marks text element as Performance Variation indication where appropriate
Var SI,L : Integer   ;
    S    : String    ;
    P,PF : PerfStyle ;
Begin
    SI := pE.El.Txt.StringIndex ;
    S  := RemovePunctuation(SI) ;
    L  := Length(S) ;
    PF := NotFound ;
    For P := AdLib to VoltaSubito do
    If (L>=PerfMin[P]) and (L<=Length(PerfTxt[P])) and (Pos(S,PerfTxt[P])=1) then PF := P ;
    PerformanceStyle := PF
End ; { PerformanceStyle }


Function FlowDirection(pE : pElementList) : FlowType ;
// Need to combine several Test elements and perhaps a Symbol element for this.
Var   SI  : Integer ;
      S   : String  ;
      F,FL  : FlowType;
Begin
    SI := pE.El.Txt.StringIndex ;
    S  := RemovePunctuation(SI) ;
    FL := NoFlow ;
    For F := Coda to DSalFine do
    If (S<>Empty) and (S = FlowTxt[F]) then FL := F ;
    FlowDirection := FL
End ; { FlowDirection }


Procedure SpecialTexts(pS : pStave) ;
// Recognises dynamic variations, performance styles, flow commands
// and concatenates split words.
Var pE : pElementList;
    VV : VolVarType  ;
    PF : PerfStyle   ;
    FL : FlowType    ;
Begin
    pE := pS.Elements ;
    While pE <> Nil do
    Begin
        If pE.El.ElType = Text then
        With pE.El.Txt do
        Begin
            VV := VolVariation(pE) ;
            PF := PerformanceStyle(pE) ;
            FL := FlowDirection(pE) ;
            If VV<> UnFound then Begin pE.El.ElType := VolVar; pE.El.VV := VV End ;
            If PF<>NotFound then Begin pE.El.ElType := PStyle; pE.El.PF := PF End ;
            If FL<>NoFlow   then Begin pE.El.ElType := Flow;   pE.El.FL := FL End ;
        End ;
        pE := pE.Next
    End ;
End ; { SpecialTexts }

Procedure GetTitle(pS : pStave) ;
Var pE, pTitle : pElementList ;
    VP, HighestVP, TitleI : Integer ;
Begin   // NB NW VPs increase down the page.
    HighestVP := pS.VertPosn ; // This should already be 0 by AdjustVertPosn.
    pTitle    := Nil  ;
    pE := pS.Elements ;
    While pE<>nil do
    Begin
        If (pE.El.ElType = Text) and (pE.El.Txt.FontNo in [2,5]) then
        Begin
            VP := pE.El.VertPosn ;
            If VP < HighestVP then Begin HighestVP := VP; pTitle := pE End
        End;
        pE := pE.Next
    End ;
    If pTitle <> Nil then
    Begin
        TitleI := pTitle.El.Txt.StringIndex ;
        pTitle.El.ElType := Unknown ;
        Music.Title.StringIndex:= TitleI ;
    End ;
End ; { GetTitle }


Procedure JoinTexts(pS : pStave) ;
{ Joins texts at same vert posn. Assumes up to 2 spaces between words.
  Also adjusts height of text to compensate for different NWC placement.
  It may be necessary to allow for VP error of +/-1 when concatenating words.
  Lyrics allocation yet to be completed.
  May assume lyrics if a) below stave b) with hyphens c) many words
}
      Function CharWidth(FntNo : Integer) : Integer ;
      Begin
          Case FntNo of
            1 :  CharWidth :=  8 ;
            2 :  CharWidth := 11 ;
            3 :  CharWidth :=  8 ;
            4 :  CharWidth := 10 ;
            5 :  CharWidth := 30 ;
            else CharWidth := 12 ;
          End
      End ; { CharWidth }

      Function VPChange(FntNo : Integer) : Integer ;
      // Adjustment to VP for mid line of text.
      Begin
          Case FntNo of
            1 :  VPChange :=  1 ;
            2 :  VPChange :=  3 ;
            3 :  VPChange :=  1 ;
            4 :  VPChange :=  2 ;
            5 :  VPChange :=  8 ;
            else VPChange :=  2 ;
          End
      End ; { VPChange }

Var pE, pF : pElementList ;
    StrI, I, L, VP, TxtFont, CW, NextHP, Count : Integer ;
    TextStr : String ;
Begin
    pE := pS.Elements ;
    While pE<>nil do
    Begin
        If (pE.El.ElType = Text) then
        Begin // Find rest of text:
            VP      := pE.El.VertPosn ;
            TxtFont := pE.El.Txt.FontNo ;
            CW      := CharWidth(TxtFont) ;
            StrI    := pE.El.Txt.StringIndex ;
            TextStr := Strings[StrI] ;
            L       := Length(TextStr) ;
            Count   := 1 ;
            NextHP  := pE.El.HorizPosn + CW*(L+3) ;
            pF := pE.Next ;
            While pF<>nil do
            Begin
                If (pF.El.ElType = Text) and
                   ((pF.El.VertPosn = VP-1) or (pF.El.VertPosn = VP) or (pF.El.VertPosn = VP+1) )
                   and (pF.El.Txt.FontNo = TxtFont) then
                If pF.El.HorizPosn<NextHP then // Append the text word:
                Begin
                    I :=  pF.El.Txt.StringIndex ;
                    If L >= 12 then
                        TextStr := TextStr + Strings[I]
                    else
                    If pF.El.HorizPosn<NextHP-1.5*CW then
                        TextStr := TextStr+' '+ Strings[I]
                    else
                        TextStr := TextStr+'  '+ Strings[I] ;
                    Inc(Count) ;
                    L := Length(Strings[I]) ;
                    NextHP := pF.El.HorizPosn + CW*(L+3) ;
                    pF.El.ElType := Unknown ;
                End
                else pF := Nil ; // Too far to next word.
                If pF<>Nil then pF := pF.Next
            End ;
            Strings[StrI] := TextStr ;
            pE.El.VertPosn := VP + VPChange(TxtFont) ;
            If (Count>10) and (pE.El.VertPosn > pS.VertPosn) then
            Begin  // Allocate as lyric - too complete:
                // Inc(pS.Stv.NoLyrics) ;
                // pS.Stv... allow up to 8 lyric lines, Have an array of 8 indices to Strings
                // Need to join up lyrics in corresponding staves of consecutive systems.
            End
        End;
        pE := pE.Next
    End ;
End ; { JoinTexts }


{#############################################################################}
{#############################################################################}
{###############    Stave Processing except Chords & Slurs   #################}
{#############################################################################}
{#############################################################################}

Procedure FixAccents(pS : pStave) ;
{  Applies accents to notes at same horizontal position.
   Works identically to the FixAccidentals procedure below.
   This must be followed by a call to DeleteUnknowns to remove the accent elements.
}
Const StopElements = [Bar,Clef,KeySig,TimeSig,NewStave];
      HPDiff = 10 ; // Max tolerance between the horiz positions.
Var pE, pF, pAcc, pNote : pElementList ;
    Ac                  : Accents      ;
    S                   : String       ;
Begin
    pE := pS.Elements ;
    While pE<>nil do
    Begin
        If pE.El.ElType = Accent then
        Begin
            pAcc := pE ;
            Ac   := pAcc.El.Ant   ;
            // Find the next note in same horiz position as pAcc:
            pNote := pAcc.Next     ;
            While (pNote<>Nil) and not (pNote.El.ElType in StopElements) and
                  (pNote.El.ElType<>Note) do
                 pNote := pNote.Next ;
            If (pNote=nil) or (pNote.El.ElType in StopElements)
               or (pAcc.El.HorizPosn+HPDiff<pNote.El.HorizPosn)  then
            Begin // Try looking before the accent.
                pNote := Nil ;
                pF := pS.Elements ;
                While (pF<>nil) and (pF<>pAcc) do
                Begin
                    If pF.El.ElType=Note then pNote := pF
                    else
                    If (pF.El.ElType in StopElements) then pNote := nil ;
                    pF := pF.Next ;
                End ;
                If (pNote<>Nil)and (pAcc.El.HorizPosn-(HPDiff div 2)>pNote.El.HorizPosn)
                then pNote := Nil ;
            End ;
            // Update the corresponding note element:
            If pNote<>nil then
            Begin
                pNote.El.No.Accent := {Ac} Accnt ;
                pAcc.El.ElType := Unknown ;
                If Ac<>Accnt then Error(14,'Noteworthy is unable to display '+AccToStr(Ac)+
                '. It has been changed to an accent.')
            End
            else
            Begin
                S := 'accent '+AccToStr(Ac)+
                     ' in '+IntToStr(pAcc.El.StaveNo)+'th stave'
                           + ', element '+IntToStr(pAcc.El.ElementNo) ;
               Error(15,'Lost note for '+S)
            End ;
            //  pAcc will be deleted by DeleteUnknowns.
        End ;
        pE := pE.Next
    End ;
End ; { FixAccents }


Procedure FixAccidentals(pS : pStave) ;
{ Applies accidental to following note of same pitch.
  Accidentals have the same vert posn & pitch as the note to which they apply.
  That note should occur before the next barline. However, the file may contain
  the note first, then the related accidental.
  Nevertheless, the horizontal posn of the note should always be at or
  after the position of the accidental.
  This must be followed by a call to DeleteUnknowns to remove the accidental elements.
}
Const StopElements = [Bar,Clef,KeySig,TimeSig,NewStave];
      HPDiff = 10 ; // 11 is needed in MP245 (due to note on adjacent pitch between accidental and corresp note.)
Var Pitch : Integer ;
    Ac    : Accidentals ;
    S     : String ;
    pE, pAcc, pNote, pF, pLast  : pElementList ;
Begin
    pE := pS.Elements ;
    While pE<>nil do
    Begin
        If pE.El.ElType = Accident then
        Begin
            pAcc  := pE ;
            Pitch := pAcc.El.Pitch ;
            Ac    := pAcc.El.Acd   ;
            // Find the next note of the same pitch as pAcc. First try after pAcc:
            pNote := pAcc.Next     ;
            While (pNote<>Nil) and not (pNote.El.ElType in StopElements) and
                  ((pNote.El.ElType<>Note)or(pNote.El.No.Pitch<>Pitch)) do
                 pNote := pNote.Next ;
            pLast := pNote ;
            If (pNote=nil) or (pNote.El.ElType in StopElements)
               or (pAcc.El.HorizPosn+HPDiff<pNote.El.HorizPosn)  then
            Begin // Try looking before the accidental.
                pNote := Nil ;
                pF := pS.Elements ;
                While (pF<>nil) and (pF<>pAcc) do
                Begin
                    If ((pF.El.ElType=Note) and (pF.El.No.Pitch=Pitch))
                    then pNote := pF
                    else
                    If (pF.El.ElType in StopElements) then pNote := nil ;
                    pF := pF.Next ;
                End ;
                If (pNote<>Nil) and (pAcc.El.HorizPosn-HPDiff div 2 > pNote.El.HorizPosn)
                then pNote := Nil ;
            End ;

            // If necessary try again after pLast:
            If (pNote = Nil) and (pLast<>Nil) and not (pLast.El.ElType in StopElements) and
                  (pLast.El.ElType=Note) and (pLast.El.No.Pitch=Pitch)
                  and (pAcc.El.HorizPosn-HPDiff div 2 < pLast.El.HorizPosn) then
                 pNote := pLast ;

            // Update the corresponding note element:
            If pNote<>nil then pNote.El.No.Accidental := Ac
            else
            Begin
                S := 'accidental, pitch '+PitchToChar(Pitch)+
                     ' in '+IntToStr(pAcc.El.StaveNo)+'th stave'
                           + ', element '+IntToStr(pAcc.El.ElementNo) ;
               Error(16,'Lost note for '+S)
            End ;
            //  pAcc will be deleted by DeleteUnknowns.
        End ;
        pE := pE.Next
    End ;
End; { FixAccidentals }

Procedure DeleteUnknowns(pS : pStave) ;
// Deletes unknown elements and those which are no longer required such as
// accents and accidentals which have been processed already.
// Also deleted are elements which are not yet recognised by ElementType:
//        Vol,VolVar,Sharps,Flats.
// Pre-conditions: Already called:
//       FixAccidentals to process Accident elements.
//       FixAccents     to process Accent   elements.
//       FixSlurs       to process SlurPt   elements.
Var pE, pPrev : pElementList ;
    Remove : Boolean ;
Begin
    pE := pS.Elements ;
    pPrev := Nil ;
    While pE<>nil do
    Begin
        Remove := False ;
        Case pE.El.ElType of
            Unknown,Accent,SlurPt, Accident :       Remove := True ;
            VolVar : If pE.El.VV = UnFound     then Remove := True ;
            PStyle : If pE.El.PF = NotFound    then Remove := True ;
            Flow   : If pE.El.Fl = NoFlow      then Remove := True ;
            TempoVar:If pE.El.TmpVar=UnknownTV then Remove := True ;
        End ;
        If Remove then
        Begin
           If pE = pS.Elements then
               Begin pS.Elements := pE.Next; pPrev := Nil End
           else
           If pPrev <> Nil then pPrev.Next := pE.Next ;
           Dispose(pE) ;
           If pPrev <> Nil then pE := pPrev.Next else pE := pS.Elements ;
           Dec(pS.NoElements) ;
        End
        else
           Begin pPrev := pE; pE := pE.Next End
    End ;
End ; { DeleteUnknowns }


Function HasBarLines(pS : pStave) : Boolean ;
Var pE    : pElementList ;
    Found : Boolean ;
Begin
    If pS = nil then
        HasBarLines := False
    else
    Begin
        pE := pS.Elements ;
        Found := False ;
        While (pE<>nil) and not Found do
        Begin
           If pE.El.ElType = Bar then Found :=True ;
           pE := pE.Next
        End ;
        HasBarLines := Found
    End ;
End ; { HasBarLines }

Procedure AddBarLines(pSbot: pStave) ;
{ Bar lines are missing from 2nd and subsequent staves in a system.
  They are added where necessary at same horizontal posn as in top line of system.
  NB Corresponding bars in different staves of the system may have different lengths
  as the length is not taken into account.
  Barlines are added to pSbot corresponding to every barline in the preceding stave.
  It is possible that other elements should be copied as well as barlines,
  such as special endings.
  NB If the final barline type is included in the stave heading, it is not copied here.
}
Var pS, pStop : pStave ;
    pE, pEtop, pPrev, pBar : pElementList ;
Begin
    pS := Music.Staves ;
    If (pS<>nil) and (pSbot<>nil) and (pSbot<>pS) and not HasBarLines(pSbot) then
    Begin
        // Find previous stave pSTop, which should already have had it barlines added:
        pSTop := nil ;
        While (pS <> Nil) and (pS<>pSbot) do Begin pStop := pS; pS := pS.Next End ;
        // Add bar lines to pSbot, scanning pStop and pSbot bar by bar in parallel:
        If pStop<>Nil then
        Begin
            pE    := pSbot.Elements ; // element in pSbot.
            pPrev := nil ;  // element before pE in pSbot.

// Need to ensure last barline is copied.

            pEtop := pStop.Elements ;
            While pEtop<>nil do // Traverse the elements of pSTop:
            Begin
                // Find next barline:
                While (pEtop<>nil) and (pEtop.El.ElType<>bar) do pEtop := pEtop.Next ;
                If (pEtop<>nil) then
                Begin
                    // Add copy of bar line to pS at same Horiz position:
                    While (pE<>nil) and (pE.El.HorizPosn<pEtop.El.HorizPosn) do
                        Begin pPrev := pE ; pE := pE.Next End ;
                    // There may be items to skip at the bar line position before inserting the bar line.
                    pBar := NewElement(Bar,Nil) ;
                    pBar.El := pEtop.El ;
                    If pPrev = nil then // Barline to be inserted at start of stave:
                    Begin
                        pBar.Next := pSbot.Elements;
                        pSbot.Elements := pBar
                    End
                    else
                    Begin
                        pBar.Next := pPrev.Next ;
                        pPrev.Next:= pBar ;
                    End ;
                    Inc(pSbot.NoElements) ;
                    // move to next bar in pStop:
                    pEtop := pEtop.Next ;
                End
            End ;
        End
    End ;
End ; { AddBarLines }


Function IsMusicElement(E:StaveElement) : Boolean ;
Begin
    IsMusicElement := E in [Clef,Bar,Note,Rest,KeySig,TimeSig,Accident]
End ; { IsMusicElement }

Procedure FixTimeSigs(pS : pStave) ;
{ Time sigs can consist of several TimeSig elements, one for each digit, e.g. 12/16.
  They can be separated by text elements and need not have the same horizontal posn values,
  but assumes no music elements separate TimeSig elements in the pElementList list.
  Upper numbers share vert posn values with barlines and newstave elements and notes on middle line.
  Lower numbers have a larger vert posn number.
  All time sig elements in a bar should belong to the same time sig.
  (But all bar lines need to be present first to check this!)
  Invisible time sigs have both numbers set to 0. This is assumed to imply 4/4 common time.
}
Var pE, pTimeSig1, pTimeSig2 : pElementList ;
    Done : Boolean ;
Begin
   // 1. First move time sig elements so there are no non-musical elements separating them.
   pE := pS.Elements ;
   // pTimeSig1 := nil ;
   Repeat // Find next time sig element at or after pE:
       While (pE<>Nil) and (pE.El.ElType<>TimeSig) do pE := pE.Next ;
       pTimeSig1 := pE ;
       Done := pTimeSig1=Nil ;
       While not Done do // Combine with remaining time sig elements in bar:
       Begin // (pE is last processed TimeSig element)
           // Skip to next music element:
           Repeat
               pE := pE.Next
           until (pE=nil) or IsMusicElement(pE.El.ElType) ;
           Done := (pE=nil) or (pE.El.ElType<>TimeSig) ;
           // Take action if pE  is a TimeSig element:
           If not Done then
           Begin
               // Combine time sigs into one element:
               pTimeSig2 := pE ;
               If pTimeSig2.El.TS.IsTop then // Upper number
                  pTimeSig1.El.TS.Top := 10*pTimeSig1.El.TS.Top + pTimeSig2.El.TS.Top
               else
                  pTimeSig1.El.TS.Bot := 10*pTimeSig1.El.TS.Bot + pTimeSig2.El.TS.Bot ;
               // Delete pTimeSig2 element:
               pTimeSig2.El.ElType := Unknown
           End ; // else reached end of time signature, marked by music element.
       End // Finished reading all TimeSig elements of a single Time Signature.
   until pE = Nil ; // Finished reading every time sig in the stave.
End ; { FixTimeSigs }

Procedure MoveNewStaves(pS : pStave) ;
{ Moves NewStave elements to be first in element list of stave pS; do this adjustment first.}
Var pE, pEold, pNewStave : pElementList ;
Begin
   pE    := pS.Elements ;
   pEOld := Nil ;
   If (pE<>Nil) and (pE.El.ElType<>NewStave) then
   // Case where newstave is not first
   Begin
       While (pE<>Nil) and (pE.El.ElType<>NewStave) do Begin pEold := pE; pE := pE.Next End ;
       If (pE<>Nil) and (pEOld<>Nil) then
       Begin
           pNewStave      := pE ;
           pEOld.Next     := pE.Next ;
           pNewStave.Next := pS.Elements ;
           pS.Elements    := pNewStave ;
       End
       else
       If Music.NWVersionNo<200 then
           Error(22,'Not found NewStave element')
   End ;
End ; { MoveNewStaves }


Procedure AdjustVertPosn(pS : pStave) ;
// Adjusts VertPosns of all elements in stave pS relative to centre line & scales them
// so WriteText, WriteFlow, WriteTempoVar place text in correct position.
// Output position is mod 256, with 0 for middle line, 1 for space below, 255 for space above, etc.
// Jan 2017: Could be the other way round!
// Output VertPosns are therefore pitches relative to 0 on the middle line.
Const NoteSpacing = 3 ; // Scaling factor: vertical pixels between notes of adjacent pitches.
Var pE : pElementList ;
    StVPos, Posn, VP : Integer ;
Begin
   StVPos := pS.VertPosn ;

   If StVPos = 0 then
   Begin
       Warning(7,'Possible loss of correct positioning of elements such as text') ;
       // There are case where this arises naturally and is corect.
   End ;
   pE := pS.Elements ;
   While (pE<>Nil) do
   Begin
       Posn := pE.El.VertPosn - StVPos ; 
       If Posn<0 then Posn := Posn+NoteSpacing*256 ;
       VP := Trunc((Posn+1)/NoteSpacing) mod 256   ;
       If VP<128 then pE.El.VertPosn := VP else pE.El.VertPosn := VP - 256 ;
       pE := pE.Next
   End ;
   pS.VertPosn := 0 ;
End ; { AdjustVertPosn }




{#############################################################################}
{#############################################################################}
{#########################         Reading          ##########################}
{#############################################################################}
{#############################################################################}



Function NoteData(Index,Actual,Bound:Integer):String ;
// Returns string giving character number Index in Line and its unexpected
// value Actual when this is greater than the maximum expected value, namely Bound.
Begin
    If (Actual>Bound) then NoteData := ' Ch'+IntToStr(Index)+'='+IntToStr(Actual)
    else NoteData := '' ;
End ; { NoteData }

Procedure UnresolvedRestData(Line:LineRA) ;
//  Reports unexpected values for note element fields.
Var S: String ;
    Problem : Boolean ;
Begin
    Problem := // Upper bounds expected for various fields:
         (Ord(Line[ 9])>7) or (Ord(Line[11])>2) or (Ord(Line[12])>3) or
         (Ord(Line[15])>1) or (Ord(Line[21])>4) ;
    If Problem then
    Begin // Upper bounds chosen to return necessary data:
        S := ' Unresolved:' ;
        S := S+NoteData( 9,Ord(Line[ 9]),7) ;
        S := S+NoteData(11,Ord(Line[11]),2) ;
        S := S+NoteData(12,Ord(Line[12]),1) ;
        S := S+NoteData(15,Ord(Line[15]),1) ;
        S := S+NoteData(21,Ord(Line[21]),4) ;
        S := S+' in rest' ;
        Error(23,S) ;
    End ;
End ; { UnresolvedRestData }

Procedure SetNoteRestData(Var Note : NoteRec; Line:LineRA) ;
// Sets parts of El.No common to Rest and Note.
Var Ghost, RestCh : Boolean ;
Begin
    With Note do
    Begin
      NoteSym  := ByteToNoteType(Line[9]) ; // Determines length of note, excluding dots.
                  // I think there's a difference of 1 between NW and NWC files.
      Pitch    := Ord(Line[10])   ;
      If Pitch > 128 then Pitch := Pitch-256 ; // Larger value means higher on stave.
      Dotted   := Ord(Line[11])=1 ;
      DoubleDot:= Ord(Line[11])=2 ;
      Case Ord(Line[12]) mod 4 of
         0 : Stem  := Up ;
         1 : Stem  := Dn ;
         2 : Stem  := NoStem ; // Used when note head has to be reversed.
         3 : Stem  := UnSet    // Occurs inexplicably for some rests.
      End ;
      RestCh   := Ord(Line[15])<>0 ;      // Unused. May indicate rest instead of note.
      Grace    := (Ord(Line[21]) = 1) or  // Grace note, printed smaller & not counted in bar length.
                  ((Ord(Line[21]) = 2)and(Ord(Line[9]) = 4)) ; // An odd exception (quaver)!
                                            // 2 for grace note with funny stem.
      Ghost    := Ord(Line[21]) = 4 ; // Unused. Has a special "x" head.
      // OneStrokeStem when Ord(Line[21]) = 2
      // DbleStrokeStem when Ord(Line[21]) = 3
      // except:
      // Hollowhead when Ord(Line[21]) = 3 for quaver and Ord(Line[21]) = 2 for semiquaver
    End ;
End ; { SetNoteRestData }


Procedure ReadRest(Var El : Element; Line:LineRA) ;
// It should also have stem properties to associate it with upper or lower line on stave.
// Line[15] contains data.
Begin
    InitialiseNoteRec(El.No) ;
    SetNoteRestData(El.No, Line) ;
    UnresolvedRestData(Line)
End ; { ReadRest }


Procedure UnresolvedNoteData(Line:LineRA) ;
//  Reports unexpected values for note element fields.
//  Case of ch9=253 in MP115: ch9 is the pitch, and element is a natural accidental.
//          ch9=251 in MP454:
Var S: String ;
    Problem : Boolean ;
Begin
    Problem :=  // Upper bounds expected for various fields:
         (Ord(Line[ 9])>7) or (Ord(Line[11])>2) or (Ord(Line[12])>2) or
         (Ord(Line[15])>0) or (Ord(Line[16])>1) or (Ord(Line[17])>1) or
         (Ord(Line[20])>1) or (Ord(Line[21])>4) or (Ord(Line[22])>0) ;
    If Problem then
    Begin // Upper bounds chosen to return necessary data:
        S := ' Unresolved:' ;
        S := S+NoteData( 7,Ord(Line[ 7]),0) ;
        S := S+NoteData( 9,Ord(Line[ 9]),7) ;
        S := S+NoteData(11,Ord(Line[11]),2) ;
        S := S+NoteData(12,Ord(Line[12]),2) ;
        S := S+NoteData(15,Ord(Line[15]),0) ;
        S := S+NoteData(16,Ord(Line[16]),1) ;
        S := S+NoteData(17,Ord(Line[17]),1) ;
        S := S+NoteData(20,Ord(Line[20]),1) ;
        S := S+NoteData(21,Ord(Line[21]),4) ;
        S := S+NoteData(22,Ord(Line[22]),0) ;
        S := S+' in note' ;
        Error(24,S) ;
    End ;
End ; { UnresolvedNoteData }

Procedure ReadNote(Var El : Element; Line:LineRA) ;
{ Ch7 = 12 normally.
 Ch7 = 35 means pitch is altered by means of a preceding accidental.
 Ch12 is not fully resolved: takes values 0,1,2 at least. Stem?
 Ch15 is not fully resolved: is upper/lower different from stem up/dn?
 Ch20 is not fully resolved: also takes values 1. Part Head reversal?
 Cases of Ch9=251,253 not resolved.
 Ch16,17 may be version dependent. In v182, slurs & accents are separate lines in file
 Ch21 is head/stem variations: 2,3 = 1,2 stem slashes; 4=crosshead
}
Var Slashes, Ch21 : Integer ;
    Crosshead     : Boolean ;
Begin
    InitialiseNoteRec(El.No) ;
    With El.No do
    Begin
      SetNoteRestData(El.No, Line) ;
      Reverse  := Odd(Ord(Line[12])div 2) ; // This is an assumtion; 2=no stem, in fact
                                     // Head is reversed on stem (as in notes a semitone apart.)
      BeamIn   := Ord(Line[13])<>0 ; // Value is element from which beam comes.
      BeamOut  := Ord(Line[14])<>0 ; // Value is element to which beam goes.
      BeamFrom := Ord(Line[13])    ; // Element No from which beam comes.
      BeamTo   := Ord(Line[14])    ; // Element No to which beam goes.
      If Music.NWVersionNo >=200 then
      Begin
           Accent := ByteToAccents(Line[17]) ;
      End
      else
      Begin  // Check version no for this.
        SlurIn   := Ord(Line[16])<>0 ;
        SlurOut  := Ord(Line[17])<>0 ;
        SlurFrom := Ord(Line[16])    ; // Element No from which slur comes.
        SlurTo   := Ord(Line[17])    ; // Element No to which slur goes.
      End ;
      TieIn    := Ord(Line[18])<>0 ; // Value is element from which tie comes.
      TieOut   := Ord(Line[19])<>0 ; // Value is element to which tie goes.
      TieFrom  := Ord(Line[18])    ; //
      TieTo    := Ord(Line[19])    ; // Element No to which tie goes.
      TieUp    := Ord(Line[20])<>0 ; // Tie (in) is above notes.
    End ;
    Ch21 := Ord(Line[21]) ;
    If Ch21 in [2,3] then Slashes := Ch21-1 else Slashes := 0 ;
    Crosshead := Ch21=4 ;
    UnresolvedNoteData(Line)
End ; { ReadNote }


Procedure ReadFlow(Var El : Element; Line:LineRA);
// Reads symbols of FlowType
Begin
    Case Ord(Line[7]) of
    115 : El.FL := Coda  ;
    116 : El.FL := Segno ;
    // 118 : El.FL := Segno ;
    else
        El.Fl := NoFlow ;
        Error(25,'Unknown Ending') ;
    End ;
End ; { ReadEnding }


Procedure ReadClef(Var El : Element; Line:LineRA) ;
// A non-initial clef. (Initial clefs are given in stave heading.)
// e.g. MP115 where Line is:
//     #001#002#038#000#186#254#002#000#001#000... bass clef
//     #001#014#211#000#186#254#007#000#000#001... treb clef
// Ch7, Ch10, Ch12 may have significance.
Begin
    // version 182,200:
    El.Cl := ByteToClef(Line[9]) // 0=treble, etc
End ; { ReadClef }


Procedure ReadAccent(Var El : Element; Line:LineRA);
// ch7,ch8,ch9,ch10 may be non-null:
//      129,13,0,4 is a mordent.  MP302
//       64,13,*,* is a > (accent).
Var Ch7 : Integer ;
Begin
    Ch7 := Ord(Line[7]) ;
    Case Ch7 of
    64 : El.Ant := Accnt ;
    129: El.Ant := Mordent ; // like a trill.
    else
        Error(27,'Unrecognised accent') ;
        El.Ant := NoAccent
    End ;
End ; { ReadAccent }

Procedure ReadVol(Var El : Element; Line:LineRA);
// Location is given by El.VertPosn
Var Ch7 : Integer ;
Begin
    Ch7 := Ord(Line[7]) ;
    Case Ch7 of
      138 : El.V := pp ;
      139 : El.V := p  ;
      140 : El.V := mp ;
      141 : El.V := mf ;
      142 : El.V := f  ;
      143 : El.V := ff ;
   // 144 : El.V := sf ;
   // 145 : El.V := fz ;
    End
End ; { ReadVol }

Procedure ReadALine(Var El : Element; Line:LineRA);
// Ch9+256*Ch10,Ch11+256*Ch12 has to be added to the HorizPosn,VertPosn
// to get the coords of the other end of the line. (See Sample2.nw for example)
Begin
    El.ElType := Unknown ;
    If Line[7] = 157 then
       Error(61,'Line segment in NW file has no equivalent in NWC format') ;
End ; { ReadALine }

Procedure ReadArpeggio(Var El : Element; Line:LineRA);
{ Example:  Line = #005#001#010#001#126#001#025#014#005#252...
  is an arpeggiano of 5 segments, starting on E=#252  }
Begin
    El.ElType := Unknown ;
    Error(61,'Arpeggio in NW file has no equivalent in NWC format') ;
End ; { ReadArpeggio }

Procedure ReadSymbol(Var El : Element; Line:LineRA);
{ see SymToStr for list of known symbols. }
Begin
    El.Sym   := Line[7] ;
    El.Pitch := Ord(Line[12]) ;
    Warning(6,'Symbol '+IntToStr(Ord(El.Sym))+' changed to text: '+SymToStr(El.Sym)) ;
End ; { ReadSymbol }

Procedure ReadPedal(Var El : Element; Line:LineRA);
Var Ch7 : Integer ;
Begin
    Ch7 := Ord(Line[7]) ;
    If Ch7 = 134 then El.Pd := peddn else El.Pd := pedup
End ; { ReadPedal }

Procedure ReadTempoVar(Var El : Element; Line:LineRA);
Var Ch7     : Integer ;
    IsBreath,
    IsFerm  : Boolean ;
Begin
    Ch7    := Ord(Line[7]) ;
    IsFerm := (Ch7=65) or (Ch7=66) ;
    IsBreath := (Ch7=128) or (Ch7=74) ;               // see Sample2.NW
    El.Pitch := Ord(Line[12]) ;
    If not IsFerm and not IsBreath then
    Begin
       Error(28, 'Unrecognised sign '+IntToStr(Ch7)) ;
       El.TmpVar := UnknownTV
    End
    else
    If IsFerm then El.TmpVar := Ferm else
    If IsBreath then El.TmpVar := Breath ;  // i.e. a pause symbol (dot inside a semicircle).
End ; { ReadTempoVar }


Procedure ReadSlurPt(Var El : Element; Line:LineRA);
{  These points occur in groups of 3 and determine 3 pts on the slur curve.
   Ch10 is the element number of the next pt in the triple, 0 for the last (=3rd) pt.
   Ch11 is the element number of the previous pt in the triple, 0 for the first pt.
   The meaning of Ch12,Ch13,Ch14,Ch15 is unknown, but the 4 values are the same for
   all 3 pts on the slur curve, and different for different slurs.
   One or more may refer instead to vertical displacement.
   Is there a reference to the indices of the notes at each end of the slur?
}
Begin
    With El.SPt do
    Begin
        Next := Ord(Line[10]) ; // Element no of next pt in slur, 0 for none.
        Prev := Ord(Line[11]) ; // Element no of prev pt in slur, 0 for none.
        Id   := Ord(Line[12]) ; // May be wrong. Ensure this is not ht measurement.
                                // Id is an identifier number for the slur.
    End
End ; { ReadSlurPt }

Procedure ReadAccidental(Var El : Element; Line:LineRA);
//  Accidentals = (NoAcc,Natural,Flat,Sharp,DoubleSharp,DoubleFlat) ;
Var P,
    AccOrd : Integer ;    // Given by ch 10 = 1 for natural; Ch12=1 for sharp
                          // Ch10=4 turns up once - unknown value.
    S : String ;
    // If Ch11 > 0 then flat? ;
    // If Ch12 > 0 then  ? ;
Begin
    P := Ord(Line[9]) ;
    If P > 128 then P := P-256 ;
    El.Pitch := P ;
    AccOrd:= Ord(Line[10]) ;
    Case AccOrd of
    254 : El.Acd := DoubleFlat ;  // guess
    255 : El.Acd := Flat ;
    0   : El.Acd := Natural ;
    1   : El.Acd := Sharp ;
    2   : El.Acd := DoubleSharp ; // guess
    else
       El.Acd := NoAcc ;
       S := ' in '+IntToStr(Ord(Line[1]))+'th stave, element '+IntToStr(Ord(Line[2])) ;
       Error(29,'Unrecognised accidental '+IntToStr(AccOrd)+ S ) ;
    End ;
End ; { ReadAccidental }


Procedure ReadBar (Var El : Element; Line:LineRA) ;
{    BarLine = (Single,Double,
                SectionOpen,SectionClose,
                LocalRepeatOpen,LocalRepeatClose,
                MasterRepeatOpen,MasteRepeatClose) ;
     Bar7 = 27,28,39,40,56,57,67,68,80
		(80 appears once) unknown interpretation, but similar to Bar9

     Bar9  = 0 for single bar line          : #027#004#000#000.#000#000#000#000#001.
     Bar9  = 1 for Double bar line          : #028#004#001#004.#001#000#000#000#000.
     Bar9  = 3 for LocalRepeatClose bar line: #057#004#003#253.#001#000#000#000#000.

     Other non-zero chars are:
       Bar10: close to 0 mod 256, range 246 to 009; unknown interpretation   
	 Bar11 = 0 or 1; unknown interpretation
	 Bar12 = 0 or 1; unknown interpretation
	 Bar15 = 0 or 1; unknown interpretation
	 Bar16 = 0 or 1; unknown interpretation
	 Bar20 = 0 or 1; unknown interpretation
	 Bar21 = 0,1,4 ; unknown interpretation
     Some characters give the number of repetitions.
}
Var BLine: BarLine ;
    Bar9 : Integer ;
    // Bar7, Bar10, Bar11, Bar15 
         
Begin
    // Bar7  := Ord(Line[ 7]) ; // Unknown meaning; mostly dictates type of barline.
       Bar9  := Ord(Line[ 9]) ; // Bar line type.
    // Bar10 := Ord(Line[10]) ; // Unknown meaning
    // Bar11 := Ord(Line[11]) ; // Unknown meaning
    // Bar15 := Ord(Line[15]) ; // Unknown meaning
{   Case Bar7 of
    27 : BLine := Single            ;
    28 : BLine := Double            ;
    39 : BLine := SectionOpen       ;
    40 : BLine := SectionClose      ;
    56 : BLine := LocalRepeatOpen   ;
    57 : BLine := LocalRepeatClose  ;
    67 : BLine := MasterRepeatOpen  ;
    68 : BLine := MasterRepeatClose ;
    else
       // 80 appears in Carols1.NW
       BLine := Single ;
       Error(30,'unknown barline '+IntToStr(Bar7)
                           + ' in '+IntToStr(Ord(Line[1]))+'th stave'
                           + ', element '+IntToStr(Ord(Line[2]))) ;
    End ; }
    Case Bar9 of // Mostly guessed:
    0 : BLine := Single            ; // Correct
    1 : BLine := Double            ; // Correct
    2 : BLine := LocalRepeatOpen   ; // Correct MP158
    3 : BLine := LocalRepeatClose  ; // MP375
    4 : BLine := MasterRepeatOpen  ; // no instances
    5 : BLine := MasterRepeatClose ; // carols1 ?
    6 : BLine := SectionOpen       ;
    7 : BLine := SectionClose      ;
    else
       BLine := Single ;
       Error(31,'Unknown barline '+IntToStr(Bar9)
                           + ' in '+IntToStr(Ord(Line[1]))+'th stave'
                           + ', element '+IntToStr(Ord(Line[2]))) ;
    End ;
    El.BL := BLine ;
End ; { ReadBar }

Procedure ReadTimeSig(pS: pStave; Var El : Element; Line:LineRA) ;
{  Many cases. Two TimeSig elements one after the other determine time sig;
   first is top, second is bottom. The numbers are stored in separate
   elements for the moment, then combined by FixTimeSigs.

   EXCEPT: 12/8 is written as three separate numbers: 1 then 8 then 2.
           perhaps separated by other elements such as text above stave.
           i.e. there's a time sig element for each digit in a time signature
           and they should occur in left to right order, but check positions.
           VertPosn is larger for the lower no, and equals that of stave for top number.
   Time sig may be given as 0/0 meaning invisible, assumed 4/4.
   There's no obvious reason for the choise of characters used!
}
Var Ch7, Ch8, No : Integer ;
    S            : String ;
Begin
    No   := 0 ;
    Ch7  := Ord(Line[7]) ;
    Ch8  := Ord(Line[8]) ;
    With El.TS do
    Begin
        Top := 0 ;
        Bot := 0 ;
        Common := False ;
        AllaBreve := False
    End ;
    If Ch8=3 then
    Begin
       If not (Ch7 in [19,20]) then
       Begin
           S := 'Unsure of common time signature' ;
           Error(32,S+' for Ch8=3, Ch7='+IntToStr(ch7)) ;
       End ;
       If Ch7 = 20 then  El.TS.AllaBreve := True  else  El.TS.Common := True ;
       El.TS.Top := 2 ;
       El.TS.Bot := 4 ;
    End
    else
    If (Ch8=2) then
    Begin
        If (Ch7> 7) and (Ch7<18) then No := Ch7- 8
        else
        If (Ch7>66) and (Ch7<77) then No := Ch7-67
        else
        Begin
          S := 'Unknown time signature' ;
          Error(34,S+' with Ch7='+IntToStr(ch7))
        End ;
        With El.TS do
        Begin
            IsTop := pS.VertPosn >= El.VertPosn ;
            If IsTop then Top := No else Bot := No ;
        End ;
    End ;
End ; { ReadTimeSig }

Procedure ReadKeySig(Var El : Element; Line:LineRA) ;
// Assume this only occurs for non-zero no of sharps or flats.
// Not known what happens for naturals to cancel a key sig.
Var OrdCh : Integer ;
Begin
    OrdCh := Ord(Line[9]) ;
    With El.KS do
    Begin
        IsFlats := OrdCh > 128  ;
        IsSharps:= not IsFlats  ;
        If IsFlats then OrdCh := 256-OrdCh ;
        Number  := OrdCh
    End
End ; { ReadKeySig }

Procedure ReadText(pS: pStave; Var El : Element; Line:LineRA) ;
{  Stores text from Line in array "Strings".
   Where "text" contains ord values above 128, the music font is intended:
                #130 is "A" ... #136 is "G"
                #128 is a flat
   See Sample1.NW for various examples where built in fonts are:
                1 : small plain
                2 : medium plain
                3 : small italic
                4 : medium italic
                5 : very large plain   "BIG"
   Texts over 12 chars go into another txy element & have no 000 terminating them.
   Attributes "small" and "italic" might be the wrong way round.
}
Var TextLen, {TestStyle,} I, OrdCh,
    Fnt   : Integer ;
    Ch    : Byte ;
    S     : String ;
Begin
    Fnt := Ord(Line[9]) ;
    If (Fnt>5) or (Fnt=0) then Error(35,'Unknown text style') ;
    If Fnt=0 then Fnt := 1 ;
    El.Txt.FontNo := Fnt ; // Doesn't work.
    El.Txt.Small := (Fnt mod 8) in [1,3] ;
    El.Txt.Italic := Odd((Fnt-1) div 2);
    TextLen := Ord(Line[10]) ;
    If TextLen>12 then
       Error(36,'Text too long to process correctly.') ;
    S := '' ;
    For I := 11 to 10+TextLen do
    Begin
       Ch := Line[I] ;
       OrdCh := Ord(Ch) ;
       If OrdCh<128 then S := S+Chr(OrdCh)  else
       If OrdCh=128 then S := S+'b' else // flat
       If OrdCh=129 then S := S+'#' else // sharp
       If OrdCh=130 then S := S+'A' else
       If OrdCh=131 then S := S+'B' else
       If OrdCh=132 then S := S+'C' else
       If OrdCh=133 then S := S+'D' else
       If OrdCh=134 then S := S+'E' else
       If OrdCh=135 then S := S+'F' else
       If OrdCh=136 then S := S+'G' else
                         S := S+'?'
    End ;
    If NoStrings<MaxStrings then Inc(NoStrings)
    else Error(37,'Run out of space for text') ;
    El.Txt.StringIndex:= NoStrings ;
    Strings[NoStrings] := S ;
End ; { ReadText }


Procedure ReadStaffHeading(pS: pStave; Var El: Element; Line: LineRA) ;
{ Processes NewSystem and NewStave elements (not the line in the file heading).
  For unusual cases see COMEFIL2.NW, COMEFILL.NW, JILUVYOU.NW, NAJOULE.NW,
                        MP14.NW, MP115.NW, MP162.NW, MP428.NW, MP428BIS.NW

  Information here should include instrument transposition and number in list
  plus stave representation as standard, upper grand, lower grand etc.
  However, this is probably in the corresponding headline.
  The time signature is in a pair of separate elements.

  Line[1]  = stave number;
  Line[2]  = element no = #elements starting at leftmost point on stave.
  Line[3]  = 32,34,35,37,38,39,40,44    height of stave?
  Line[4]  = 0
  Line[5]+256*Line[6]
           = common vert position of stave, bar lines, key sigs and upper time sig nos.
  Line[7]  = clef?,  = 2, 3 for new system or new stave. 
  Line[8]  = 0
  Line[9]  = clef // 0 for treble, 1 for bass; matches Line[7] in that Line[7,9] = 2,1 or 3,0 in all known cases.

 # of stave in system; 0 for start system, 1 for next line in system.
  The remaining numbers are usually 0, none with a known meaning.
  Line[10,11] : probably same meaning as in a note.
  Line[12] = 0,1... but a small number of values including ascii of numbers
                       and ascii of A,B,...,G.
  Line[19] = 0..6
  Line[21] = 0,1
}
Begin
    pS.StaveNo := El.StaveNo ; // = Line[1]
    pS.Stv.Clef := ByteToClef(Line[9]) ;  // Unsure; 0=treble, 1=bass.
    pS.VertPosn := El.VertPosn  ;
End ; { ReadStaffHeading }


Function ElementType(Ch7,Ch8:Byte) : StaveElement ;
{    StaveElement = (Clef,Bar,Note,Rest,TimeSig,KeySig,Text,Vol,VolVar,
                     Sign,Tempo,TempoVar,Flow,Style,Pedal,
                     NewSystem,NewStave,Sharps,Flats,Unknown) ;
     Not all of these are recognised.       0 ->48 a->97 A->65       }
Var El : StaveElement ;
    No7,No8 : Integer ;
Begin
    No7 := Ord(Ch7) ;
    No8 := Ord(Ch8) ;

    If No8=0 then
    Begin
       If No7 < 4 then El := NewStave else
                       El := Clef
    End
    else
    If (No8 = 1) then El := KeySig  else
    If (No8 = 2) then El := TimeSig else
    If (No8 = 3) then El := TimeSig else  // 3 is Common Time symbol
    If (No8 = 4) then El := Bar     else  // may not be true
    If (No8 = 7) then El := Accident else
    If (No8 = 9) then El := Text    else
    If (No8 =11) then El := ALine   else
    If (No8 =12) then El := SlurPt  else
    If (No8 =14) then El := Arpeg   else

    If (No8 = 5) then // Note or Rest:
    Begin
        If No7 in [12,25,31,33,35,44,46,48] then El := Note else
        If No7 in [39..43,51..53] then El := Rest else
                          El := Unknown ;
    End
    else
    If (No8 = 13) then // Cases of symbols
    Begin
      If (No7 in [115,116,118]) then El := Flow     else
      If (No7 in [64,129])      then El := Accent   else  // @
      If (No7 in [65,66,128])   then El := TempoVar else  // 54?
      If (No7 in [134,135])     then El := Pedal    else
      If (No7 in [138..143])    then El := Vol      else
                                     El := Symbol
    End
    else
       El := Unknown ; // No instances in my files.
    ElementType := El ;
End ; { ElementType }

Function ChToPixels(Ch1,Ch2:Byte) : Integer ;
// Ch1, Ch2 are LineRA elements, Ch1 just before Ch2.
// Line[3], Line[4] for vertical position;
// Line[5], Line[6] for horizontal position;
Var N1,N2 : Integer ;
Begin
    N1 := Ord(Ch1) ;
    N2 := Ord(Ch2) ;
    If N2>128 then N2 := N2-256 ;
    ChToPixels := 256*N2+N1 ;
End ; { ChToPixels }

Procedure ReadElement(pS: pStave; Var El : Element; ElNo,StvNo:Integer) ;
{ Reads Line from input file. Naming its elements as Ch1,Ch2,...,Ch22:
  ElNo  = 256*Ch1+Ch2.
  Horiz position on page = 256*Ch4+Ch3 (relative to left side)
  Vert position on page  = 256*Ch6+Ch5 (relative to centre or top)
  Ch7,Ch8 = element type
  Ch9..Ch22 = element details.
}
Var Line :LineRA ;
    Len  : Integer ;
    Done : Boolean ;
    ElTy : StaveElement ;
    S    : String ;
Begin
    BlockRead(ReadFile,Line,LineLen,Len) ;
    Done := Len<>LineLen ;
    If Done then
    Begin
        If (Len<>0) then Error(38,'Unexpectedly short final line for stave element') ;
    End
    else
    Begin
        If (StvNo mod 256)<>Ord(Line[1]) then
           Error(39,'Mismatch in stave number of element '+IntToStr(StvNo)) ;
        If (ElNo mod 256)<>Ord(Line[2]) then
           Error(40,'Mismatch in element number '+IntToStr(ElNo)) ;
        El.StaveNo   := StvNo;
        El.ElementNo := ElNo ;
        El.HorizPosn := ChToPixels(Line[3],Line[4]) ;
        El.VertPosn  := ChToPixels(Line[5],Line[6]) ;
        ElTy         := ElementType(Line[7],Line[8]);
        El.ElType    := ElTy ;

        Case ElTy of
          NewStave:ReadStaffHeading(pS,El,Line);
          Clef   : ReadClef  (El,Line) ;
          Note   : ReadNote  (El,Line) ;
          Rest   : ReadRest  (El,Line) ;
          Bar    : ReadBar   (El,Line) ;
          Text   : ReadText  (pS,El,Line) ;
          TimeSig: ReadTimeSig(pS,El,Line) ;
          Flow   : ReadFlow  (El,Line) ;
          KeySig : ReadKeySig(El,Line) ;
          SlurPt : ReadSlurPt(El,Line) ;
          Accident:ReadAccidental(El,Line);
          Accent : ReadAccent(El,Line) ;
          TempoVar:ReadTempoVar(El,Line) ;
          Pedal  : ReadPedal (El,Line) ;
          Vol    : ReadVol   (El,Line) ;
          ALine  : ReadALine (El,Line) ;
          Arpeg  : ReadArpeggio(El,Line);
          Symbol : ReadSymbol(El,Line) ;
        else // occurs only in HEREIAM.NW & Gold1:
          S := 'Unexpected or unknown staff element: '+El2String(ElTy)
                +' (possibly a flow direction)' ;
          Error(41,S + ' for stave '+IntToStr(StvNo)+', element '+IntToStr(ElNo))
        End ;
    End
End ; { ReadElement }




{#############################################################################}
{#############################################################################}
{#########################   Reading of Headings   ###########################}
{#############################################################################}
{#############################################################################}


Procedure SkipElement ; // Used in ReadStave for testing sample NW files.
Var Line :LineRA ;
    Len  : Integer ;
Begin
    BlockRead(ReadFile,Line,LineLen,Len) ;
End ;

Procedure ReadStave(StaveNo:Integer) ;
// The staff heading element may not be first.
// It can be preceeded by text elements, for example.
Var I,NoEls : Integer ;
    pS : pStave ;
    pEl: pElementList ;
Begin
    // Find start of elements of stave StaveNo:
    pS := Music.Staves ;
    For I := 1 to StaveNo-1 do pS := pS.Next ;
    pEl := pS.Elements ;
    NoEls := pS.NoElements ;
    pS.Stv.NoLyrics := 0 ;

    // Read the elements of the stave:
    For I := 1 to NoEls do     // Perhaps should be NoEls-1.
    Begin
        // If (StaveNo=10) and (I in[23,25]) then  SkipElement  else
        ReadElement(pS,pEl.El,I,StaveNo) ;
        If pEl<>Nil then pEl := pEl.Next
        else Error(42,'Incorrect number of elements in file header for staff')
    End ;

    // Process elements in stave pS from NW to NWC format:
    ModificationsAfterRead(pS) ;
    MoveNewStaves (pS) ;
    AddBarLines   (pS) ;  // Occasional problems
    FixSlurs      (pS) ;  // move to after make chords?
    FixTimeSigs   (pS) ;
    BeamCheck     (pS) ;
    TieCheck      (pS) ;
    MakeChords    (pS) ;
    FixAccidentals(pS) ;
    FixAccents    (pS) ;
    AdjustVertPosn(pS) ;
    JoinTexts     (pS) ;
    SpecialTexts  (pS) ;
    If StaveNo = 1 then GetTitle(pS) ;
    DeleteUnknowns(pS) ;
End ; { ReadStave }

Procedure SystemStart(StaveNo:Integer; IsLast:Boolean; StTyp:StaveType; VP : Integer) ;
// Groups Staves into systems.
// Initialises FirstInSystem and LastInSystem.
// IsLast is true iff StaveNo is last stave in its system.
Var I   : Integer ;
    pS  : pStave ;
Begin
    pS := Music.Staves ;
    If StaveNo = 1 then
    With pS^ do
    Begin
        Stv.StType    := StTyp ;
        FirstInSystem := True  ;
        LastInSystem  := IsLast;
        VertPosn      := VP
    End
    else
    If StaveNo > 1 then
    Begin
        // Skip Staves until StaveNo-1:
        For I := 1 to StaveNo-2 do pS := pS.Next ;
        // pS.Next is now stave StaveNo:
        With pS.Next^ do
        Begin
            If pS.Stv.StType = Upper then Stv.StType := Lower else
            Stv.StType    := StTyp ;
            FirstInSystem := pS.LastInSystem ;
            LastInSystem  := IsLast;
            VertPosn      := VP
        End
    End ;
End ; { SystemStart }

Procedure StartElements(StaveNo, NE:Integer) ;
// Creates record for each element in the stave.
Var I   : Integer ;
    pS  : pStave ;
    pEl : pElementList ;
Begin
    pS := Music.Staves ;
    For I := 1 to StaveNo-1 do
        pS := pS.Next ;
    With pS^ do
    Begin
        NoElements := NE ;
        Elements := Nil ;
        For I := 1 to NE do
        Begin
            pEl := NewElement(Unknown,Elements) ;
            Elements := pEl
        End ;
    End
End ; { StartElements }

Procedure StartStaves(NS,VN:Integer) ;
// Creates head record for each stave.
Var I : Integer ;
    pS : pStave ;
Begin
    If NS=0 then
    Begin
        // MessageDlg('File '+ReadFileName+' has wrong format or no staves', mtInformation, [mbOk], 0)
        Error(67,'File has wrong format or no staves') ;
    End
    else
    With Music do
    Begin
        NoStaves    := NS ;
        NWVersionNo := VN ;
        Staves      := Nil ;
        With Title do
        Begin
            Strings[0]  := Empty ;
            StringIndex := 0     ;
            FontNo      := 3     ;
            Italic      := False ;
            Small       := False ;
        End ;
        For I := 1 to NS do
        Begin
            NewStaveRec(pS) ;
            pS.StaveNo := I ;
            pS.Next := Staves ;
            Staves := pS
        End ;
    End
End ; { StartStaves }

Procedure ParseHeadLine(Line:LineRA; LineNo:Integer) ;
{ Line is the input line to parse.
The first line (i.e. LineNo=0):
   Ch1=#staves:
   Ch4 normally 0, 2 or 4.
   Ch8 = #positions above stave; =9,10 typically.

Except for first line:
   Ch1,Ch2 = Vertical position of stave centre.
   Ch3,Ch4 = Vert posn of bottom of lowest stave element
   Ch5,Ch6 = Vert posn of top of highest stave element
   Ch7 = 0
   Ch8  = #elements in stave.
   Ch13 = 0 for last line in system
          1 for non-last line in system
   NB NowStave elements in v163, but none in v200.
}
Var AboveStave,
    NoElements,
    NoStaves,
    VPosn,
    VerNo    : Integer ;
    IsLast   : Boolean ;
    StTyp    : StaveType ;
Begin
    If (LineNo=0) then
    Begin
        NoStaves := Ord(Line[1]) ;
        // Log(IntToStr(NoStaves)+' staves') ;
        // Line[2,5,10] = 0 always
        // Line[3,6] = 1,2,3
        // Line[4]   = 0,2,4
        // Line[7]   = 3,4,5
        AboveStave := Ord(Line[8]) ;   // Usually =10, ######## To incorporate
        VerNo      := Ord(Line[9]) ;   // 163,182,200  Version number
        // If Music.VerNo<=170 then Warning(5,'This file has a very early NW version and may not be converted properly.') ;
        // Line[11]  = 1 always
        // Line[12..22] usually all 0s, sometimes a list of non-zeroes.
        StartStaves(NoStaves,VerNo)
    End
    else
    Begin
        { LineNo>0: The line has info about stave with index LineNo:
          For all versions 163 to 200:
          Ch1,Ch2=vertical pos of centre of stave (=Ch5,Ch6 of newstave/newsystem line at start of stave)
               Vertical Position is measured from page centre or page top, depending on the SW version.
          Ch3,Ch4=vert pos of lowest element of stave
          Ch5,Ch6=Vert pos of highest element of stave
          Ch7=0,1
          In 8 cases spread over 4 files, Ch7 is non-zero. (MP428,MP428bis,NAJoule,JILUVYOU)
          Ch7 =1 for first of stave pair RH/LH
          (No obvious intepretation; they includes the 3 files with unusual first line. )
          Ch8=#elements in stave.
          Except v163:
          Ch9,Ch10 : position of LH end of stave.
          Ch11,Ch12: posn of RH end of stave.
          Ch9=163 for version 163.

          Ch14 = 030 for orchestral?
          Ch9-Ch22 of all lines contain the same info (font details etc.)
          (For line 0 this covers text (headers) which precedes the first stave.)
          For v163:
          Last 14 chars. (With one exception, these are identical within a file, except for Ch13):
          MP381 lines end   #163#000.#001#006#000#000#061.#184#000#000#000#006.+0:2
          MP 29 lines end   #163#000.#001#006#058#139#061.#184#093#016#000#006.+0:2
          In MP302, MP307, MP388, MP553:
          All headlines end #163#000.#001#108#083#184#093.#248#061#098#001#023.#077#184

          MP142 lines end   #163#000.#001#117#084#193#094.#248#061#098#001#032.#078#193
          MP630 lines end   #163#000.#001#117#084#193#094.#248#061#098#001#032.#078#002
        }
        If Ord(Line[2])>128 then
           VPosn := Ord(Line[1])+256*(Ord(Line[2])-256)
        else
           VPosn := Ord(Line[1])+256*Ord(Line[2]) ;
        NoElements := Ord(Line[8]) ;
        StTyp := Standard ;
        If Ord(Line[14])=30 then StTyp := Orchestra
        else
        If Ord(Line[7])=1 then StTyp := Upper ;
        IsLast     := (Line[13]=0) or (Music.NWVersionNo<=170) ;
                                     // Not known what to do for version 163.
        // Log(IntToStr(NoElements)+' elements in stave '+IntToStr(LineNo)) ;
        StartElements(LineNo,NoElements) ;
        SystemStart(LineNo,IsLast,StTyp,VPosn)
    End
End ; { ParseHeadLine }


Procedure ReadHeading(Var Done : Boolean; Var NoStaves: Integer) ;
{ Done true if at end of file.
  NoLines = #lines in heading after opening line.
  Heading contains one line for each stave after opening line.
}
Var Line : LineRA ;
    Len, I : Integer ;
    EoH  : Boolean ;
Begin
    BlockRead(ReadFile,Line,LineLen,Len) ;
    EoH := Len<LineLen ;
    If EoH then Error(43,'Unexpected end of file in first line')
    else
    Begin
        ParseHeadLine(Line,0) ;
        NoStaves := Ord(Line[1]) ;
        For I := 1 to NoStaves do
        If not EoH then
        Begin
             BlockRead(ReadFile,Line,LineLen,Len) ;
             EoH := Len<LineLen ;
             If EoH then Error(44,'Unexpected end of file')
             else
             ParseHeadLine(Line,I) ;
        End ;
    End ;
    Done := EoH ;
End ; { ReadHeading }

Procedure ParseFooter(Line : LineRA) ;
Var S : String  ;
    I, TextLen : Integer ;
Begin
    If (Ord(Line[1])<>0) or not (Ord(Line[2]) in [1,2]) then
        Error(65,'Unexpected line at end of file')
    else
    Begin
        TextLen := Ord(Line[10]) ;
        S := '' ;
        For I := 11 to 10+TextLen do S := S+Chr(Ord(Line[I])) ;
        If TextLen<12 then S := S+Space
    End ;
    Case Ord(Line[2]) of
      1 : Head := Head+S ;
      2 : Foot := Foot+S
    End
End ; { ParseFooter }

Procedure ReadFooters ;
Var Line : LineRA ;
    Len  : Integer ;
    EoH  : Boolean ;
Begin
    Head := '' ;
    Foot := '' ;
    BlockRead(ReadFile,Line,LineLen,Len) ;
    EoH := Len<LineLen ;
    While not EoH do
    Begin
        ParseFooter(Line) ;
        BlockRead(ReadFile,Line,LineLen,Len) ;
        EoH := Len<LineLen ;
    End ;
End ;

Procedure ReadNWFile ;
{ Reads from TmpFile, writes to NewFile. }
Var Done : Boolean ;
    I, NoStaves : Integer ;
    pS   : pStave ;
Begin
    ReSet(ReadFile,1) ;
    NoStrings := 0 ;
    ReadHeading(Done,NoStaves) ;
    For I := 1 to NoStaves do ReadStave(I) ;
    ReadFooters ;
    Close(ReadFile) ;
    Log('') ;
    // Perform processing necessary at system level:
    JoinSystems ;
    InsertClefs ;
    CreateReChordElements ;
    RemoveUnnecessaryElements ;
    RemoveEndBarLine ;
    ModificationsBeforeWrite ;
    TestDeletions ;
    ReWrite(NewFile,1) ;
    WriteHeading ;
    pS := Music.Staves;
    While pS<>nil do
    Begin
        WriteStave(pS) ;
        pS := pS.Next
    End ;
    Close(NewFile) ;
    Log('') ;
    DisposeOfMusic ;
End ; { ReadNWFile }



{#############################################################################}
{#############################################################################}
{########################## Conversion management: ###########################}
{#############################################################################}
{#############################################################################}

Function TMainForm.MakeUnique(WFname : String; ListBox: TFileListBox) : String ;
{ If necessary, adds subscript n to file name to make it unique in subdirectory.
  Assumes 3 character suffix accepted by mask used in ListBox. }
Var  Suffix, Prefix,
     NewName, Name : String ;
     I,C,L,N     : Integer ;
     Found         : Boolean ;
Begin
     ListBox.Update;
     L := Length(WFname) ;
     If WFname[L-3]<>'.' then
         Begin Error(45,'Unexpected file type in'+WFname); NewName := WFname End
     else
     Begin
         Prefix := Copy(WFname,0,L-4);  // all except .nwc or .tmp
         Suffix := Copy(WFname,L-3,4);
         N := -1 ;
         C := ListBox.Items.Count - 1 ;
         Found := True ;
         While Found do
         Begin
            If N = -1 then NewName := Prefix+Suffix
            else NewName := Prefix+'_'+IntToStr(N)+Suffix ;
            Found := False ;
            For I := 0 to C do
            Begin
               Name := ListBox.Items.Strings[I] ;
               If Lowercase(Name) = Lowercase(NewName) then Found := True ;
            End ;
            N := N+1 ;
         End ;
     End ;
     MakeUnique := NewName
End ; { MakeUnique }

Function TMainForm.AddSubscript(WFname : String; ListBox: TFileListBox) : String ;
{ Changes file type from .nw to .nwc }
{ If necessary, adds subscript n to file name to make it unique in subdirectory. }
Var  NewName : String ;
     L       : Integer ;
Begin
     // If Pos(OldSuffix,WFname)>0 then NewName := WFname+'c' else NewName := WFname+'C' ;
     // Replaces .nw or .NW with .nwc
     L := Length(WFname) ;
     NewName := Copy(WFname,0,L-3)+NewSuffix ;
     AddSubscript := MakeUnique(NewName,ListBox) ;
End ; { AddSubscript }


Procedure TMainForm.ConvertFile(FNo:Integer);
// Copies original file into TmpFile,
// Calls modification of Ith HTML file displayed in WorkingFilesBox.
// Copies NewFile to original location of file after saving copy of original.
Var  WDname, NewFName, WriteFileName : String ;
Begin
     // Determine name of file to be transformed:
     WDname := WorkingDirBox.Directory ;
     WFname := WorkingFileBox.Items.Strings[FNo] ;
     ReadFileName := WDname+'\'+WFname ; // Global, available to log file.
     AssignFile(ReadFile,ReadFileName) ;

     // Assign & create new file for writing to:
     NewFName      := AddSubscript(WFname,WorkingFileBox) ;
     WriteFileName := WDname+'\'+NewFname  ;
     AssignFile(NewFile,WriteFileName) ;
     Rewrite(NewFile,1);
     CloseFile(NewFile);

     // Perform NW to NWC transformation, writing result into NewFile:
     Logged := False ;
     ReadNWFile ;

     // Update user interface modified file count:
     ModifiedFiles := ModifiedFiles+1;
     ModifiedBox.Text := IntToStr(ModifiedFiles) ;
     ModifiedBox.Update ;
     ErrorBox.Text := IntToStr(Errors) ;
     ErrorBox.Update
End; { ConvertFile }



{#############################################################################}
{#############################################################################}
{#########################  Scanning for NW files:  ##########################}
{#############################################################################}
{#############################################################################}


Function TMainForm.OKFileName(FileNo:Integer) : Boolean ;
// Tests to see if this is a renamed original file or new file or otherwise.
// Excludes certain standard (system) html files.
Var FName : String ;
    L     : Integer ;
Begin
    Fname := Lowercase(WorkingFileBox.Items.Strings[FileNo]) ;
    L := Length(FName) ;
    OKFileName := (Pos(OldSuffix, Fname)=L-2) and
                  (Pos(NewSuffix, Fname)=0) ;
End ; { OKFileName }

Procedure TMainForm.ConvertDir(WDNo:Integer) ;
// Converts .NW files to .NWC files in WDNo-th subdirectory of current WorkingDirectory.
// ProIn,ProOut are progress bar values at start and end of procedure call.
var CurrentDir, WorkingDir,
    OldName, NewName : String;
    I, Start, Size : integer;
Begin
     // Store current working dir and move to new working dir:
     CurrentDir := WorkingDirBox.Directory ;
     WorkingDir := WorkingDirBox.GetItemPath(WDNo);
     WorkingDirBox.Directory := WorkingDir;

     // Convert any NW files:
     I := 0 ;
     NewName := '';
     WorkingFileBox.Update ; // File mask is set to "*.nw*"
     MainForm.Update;
     While (I<WorkingFileBox.Items.Count) do
     Begin
         // First check update of previous file has not moved
         // the previous file to new current position:
         OldName := NewName ;
         NewName := WorkingFileBox.Items.Strings[I];
         If (OldName<>NewName) and OKFileName(I) then ConvertFile(I) ;
         ErrorBox.Text := IntToStr(Errors) ;
         ErrorBox.Update ;
         WorkingFileBox.Update ;
         I := I+1 ;
     End ;

     // Convert any subdirectories:
     Start := WorkingDirBox.ItemIndex+1 ;
     Size  := WorkingDirBox.Items.Count-1 ;
     For I := Start to Size do ConvertDir(I) ;

     // Restore previous working dir:
     WorkingDirBox.Directory := CurrentDir;
End ; { ConvertDir }

Procedure TMainForm.AssignLogFile ;
{ Chooses suffix _n of log file to make it unique in the log file directory. }
Var  LogFileDir, LFName, NewName : String ;
     I,C,N   : Integer ;
     Found   : Boolean ;
Begin
     LogFileDir := LogFileBox.Directory ;
     LogFileListBox.Directory := LogFileDir ;
     LogFileListBox.Update;
     C := LogFileListBox.Items.Count-1 ;
     N := 0 ;
     Repeat
         Found := False ;
         NewName := LogFname+IntToStr(N)+'.log' ;
         For I := 0 to C do
             If LogFileListBox.Items.Strings[I] = NewName then Found := True ;
         N := N+1 ;
     until not Found ;
     LFName := LogFileDir+'\'+NewName    ;
     AssignFile(LogFile,LFName) ;
End ; { AssignLogFile }


Procedure TMainForm.ResetAll(Sender: TObject) ;
Var E : ErrorType;
    W : WarnType ;
Begin
     For E := 1 to MaxErr  do ErrorRA[E] := 0 ;
     For W := 1 to MaxWarn do WarnRA [W] := 0 ;
     ModifiedFiles := 0 ;
     Errors := 0 ;
     ErrorBox.Text := IntToStr(Errors) ;
     ErrorBox.Update ;
     Warnings := 0 ;
     WarningBox.Text := IntToStr(Warnings) ;
     WarningBox.Update ;
     ModifiedBox.Text := IntToStr(ModifiedFiles) ;
     ModifiedBox.Update ;
     GoButton.SetFocus ;
End ; { ResetAll }

Procedure ErrorReport;
Var E : ErrorType;
    W : WarnType ;
    // I : Integer ;
Begin
    Writeln(LogFile,'*** '+IntToStr(ModifiedFiles)+' files were created. ***') ;
    Writeln(LogFile,'*** Finished Updates with '+IntToStr(Errors)+' errors and '
                            +IntToStr(Warnings)+' warnings ***') ;
    If Errors>0 then
    Begin
       Writeln(LogFile,'The following errors types (and their frequencies) were reported:') ;
       For E := 1 to MaxErr do
       If ErrorRA[E]>0 then Writeln(LogFile,IntToStr(E)+' ('+IntToStr(ErrorRA[E])+'), ') ;
       Writeln(LogFile)
    End ;
    If Warnings>0 then
    Begin
       Writeln(LogFile,'The following warning types (and their numbers) were reported:') ;
       For W := 1 to MaxWarn do
       If WarnRA[W]>0 then Writeln(LogFile,IntToStr(W)+' ('+IntToStr(WarnRA[W]),'), ') ;
       Writeln(LogFile)
    End ;
End ; { ErrorReport }

Procedure TMainForm.ConfigFile(Open: Boolean) ;
{ If open, read working dir and log dir from config file and init the dir boxes,
  else collect working dir and log dirs from boxes and write to config file. }
Var LogFileDir, SourceFileDir, ConfigFileName : String ;
    CfgFile : TextFile ;
Begin
    {$I-}
    ConfigFileName := ExeDir+'\NW2NWC.cfg' ;
    AssignFile(CfgFile, ConfigFileName) ;
    If IOResult=0 then
    If Open then
    Begin
        Reset(CfgFile) ;
        If IOResult=0 then
        Begin
            Readln(CfgFile,LogFileDir) ;
            If IOResult=0 then
            Begin
                If DirectoryExists(LogFileDir) then
                Begin
                    LogFileBox.Directory := LogFileDir ;
                    LogFileBox.Update ;
                End ;
                Readln(CfgFile,SourceFileDir) ;
                If (IOResult=0) and DirectoryExists(SourceFileDir) then
                Begin
                    SourceDirBox.Directory := SourceFileDir;
                    SourceDirBox.Update
                End
            End ;
            CloseFile(CfgFile) ;
        End
    End
    else
    Begin
        Rewrite(CfgFile) ;
        If IOResult=0 then
        Begin
            LogFileDir   := LogFileBox.Directory   ;
            SourceFileDir:= SourceDirBox.Directory ;
            Writeln(CfgFile,LogFileDir) ;
            Writeln(CfgFile,SourceFileDir) ;
            CloseFile(CfgFile) ;
        End;
    End
    {$I+}
End ; { ConfigFile }

Procedure TMainForm.GoButtonClick(Sender: TObject);
Var    Start : Integer ;
       Source: String  ;
Begin { GoButtonClick }
       InitPerfTxt ;
       ResetAll(Sender) ;
       CloseButton.SetFocus ;
       AssignLogFile ;
       ReWrite(LogFile) ;
       Source   := SourceDirBox.Directory ;
       LenSource:= Length(Source) ;
       Start    := SourceDirBox.ItemIndex ;
       ConfigFile(False) ;
       ConvertDir(Start) ;
       ErrorReport ;
       CloseFile(LogFile) ;
End; { GoButtonClick }


Procedure TMainForm.CloseButtonClick(Sender: TObject);
Begin
    close
End;

procedure TMainForm.SourceDriveBoxChange(Sender: TObject);
begin
     SourceDirBox.Drive  := SourceDriveBox.Drive;
     WorkingDirBox.Drive := SourceDriveBox.Drive;
     ResetAll(Sender) ;
end;

procedure TMainForm.SourceDirBoxChange(Sender: TObject);
begin
     SourceFileBox.Directory := SourceDirBox.Directory;
     SourceFileBox.Update;
     SourceDirBox.OpenCurrent;
     WorkingDirBox.Directory := SourceDirBox.Directory;
     ResetAll(Sender) ;
end;

procedure TMainForm.WorkingDirBoxChange(Sender: TObject);
begin
     WorkingFileBox.Directory := WorkingDirBox.Directory;
     WorkingFileBox.Update;
     WorkingDirBox.OpenCurrent;
end;

procedure TMainForm.FormActivate(Sender: TObject);
begin
     GoButton.SetFocus ;
     ConfigFile(True) ;
end;


procedure TMainForm.FormCreate(Sender: TObject);
begin
     ExeDir := GetCurrentDir ;
end;

procedure TMainForm.LogFileDriveChange(Sender: TObject);
begin
     LogFileBox.Drive  := LogFileDrive.Drive;
     // LogFileBoxChange(Sender);
end;

procedure TMainForm.LogFileBoxChange(Sender: TObject);
begin
     LogFileBox.OpenCurrent;
     LogFileListBox.Directory := LogFileBox.Directory;
     LogFileListBox.Update;
     GoButton.SetFocus ;
end;

End.




